﻿using System.Collections.Generic;
using System.Linq;
using Emgu.CV;
using System.Drawing;
using Emgu.CV.Util;
using WingAIDiagnosisService.Carotid.CarotidInterFacesData;
using WingServerCommon.Log;
using System;

namespace WingAIDiagnosisService.Carotid.MathTools
{
    public struct LineFittingError
    {
        public int SerialNumber;
        public Point OriginalPoint;
        public double DiffValue;
    }

    /// <summary>
    /// 计算切面的三个点
    /// </summary>
    public struct One2DPlaneThreeBasic2DPoints
    {
        public Vector2I ZeroPoint;
        public Vector2I XAxisPoint;
        public Vector2I YAxisPoint;
    }

    /// <summary>
    /// 二维点列的取值范围
    /// </summary>
    public struct CoordinateRange
    {
        public double MinX;
        public double MinY;
        public double MaxX;
        public double MaxY;
    }

    public class MathTools2D
    {
        /// <summary>
        /// 获得整数点向量
        /// </summary>
        /// <param name="pointStart"></param>
        /// <param name="pointEnd"></param>
        /// <returns></returns>
        public static Vector2I GetVector(Vector2I pointStart, Vector2I pointEnd)
        {
            return new Vector2I(pointEnd.X - pointStart.X, pointEnd.Y - pointStart.Y);
        }

        /// <summary>
        /// 获得double点向量
        /// </summary>
        /// <param name="pointStart"></param>
        /// <param name="pointEnd"></param>
        /// <returns></returns>
        public static Vector2 GetVector(Vector2 pointStart, Vector2 pointEnd)
        {
            return new Vector2(pointEnd.X - pointStart.X, pointEnd.Y - pointStart.Y);
        }

        private static List<Point> RemoveAbnormalData(List<Point> points)
        {
            var num = 0;
            while (num < 3)
            {
                points = RemoveAbnormalData1(points);
                num++;
            }
            points = points.OrderBy(o => o.Y).ToList();
            return points;
        }

        private static List<Point> RemoveAbnormalData1(List<Point> points)
        {
            //计算拟合直线
            var count = points.Count;
            if (count < 10)
            {
                return points;
            }
            var line = CalculateFittedLineNoRemovePoint(points);
            var pointDiff = new LineFittingError[count];

            //计算每个点的差值
            double sumDiff = 0;
            for (var i = 0; i < count; ++i)
            {
                //var newX = line[0] * points[i].X + line[2];
                double diff = GetDistaceBetweenPointAndLine(line, points[i]);
                sumDiff += diff;
                pointDiff[i] = new LineFittingError()
                {
                    SerialNumber = i,
                    OriginalPoint = points[i],
                    DiffValue = diff
                };
            }
            //计算差值的，平均值和标准差
            var averageDiff = sumDiff / points.Count;
            double sumStd = 0;
            for (var i = 0; i < count; ++i)
            {
                var diff = pointDiff[i].DiffValue;
                sumStd += (diff - averageDiff) * (diff - averageDiff);
            }
            sumStd = Math.Sqrt(sumStd / count);


            var minDiff = averageDiff - 3 * sumStd;
            var maxDiff = averageDiff + 3 * sumStd;
            //删除差值大于平均值加减3倍的标准差的点
            for (var i = 0; i < count; ++i)
            {
                var diff = pointDiff[i].DiffValue;
                if (diff < minDiff || diff > maxDiff)
                {
                    pointDiff[i].DiffValue = -100000;
                }
            }
            //去掉10%的数据
            var count1 = count / 10;
            var diffList = pointDiff.ToList().OrderByDescending(o => o.DiffValue);
            var diffList1 = diffList.Skip(count1).ToList();

            var pointList = new List<Point>();
            for (var i = 0; i < diffList1.Count; ++i)
            {
                if (diffList1[i].DiffValue != -100000)
                {
                    pointList.Add(diffList1[i].OriginalPoint);
                }
            }
            return pointList;
        }

        /// <summary>
        /// 点到直线的距离
        /// </summary>
        /// <param name="line"></param>
        /// <param name="point"></param>
        /// <returns></returns>
        public static double GetDistaceBetweenPointAndLine(double[] line, Point point)
        {
            if (line[1] == 0)
            {
                return Math.Abs(point.X - line[2] / line[0]);
            }

            var A = line[0];
            var C = line[2];
            var a = point.X;
            var b = point.Y;
            var n = (A * a + A * A * b + C) / (A * A + 1);
            var m = a - A * n + A * b;
            if (Math.Abs(m - a) < 0.00001)
            {
                return 0;
            }

            var disX = m - a;
            var disY = n - b;
            return Math.Sqrt(disX * disX + disY * disY);
        }

        /// <summary>
        /// 计算一系列点的拟合直线,Ax+By+C=0,B=0或-1
        /// </summary>
        /// <param name="points"></param>
        /// <returns></returns>
        public static double[] CalculateFittedLine(List<Point> points)
        {
            double[] result = { 0, 0, 0 };
            try
            {
                //去除异常数据
                points = RemoveAbnormalData(points);
                var count = points.Count;
                if (count < 2)
                {
                    return new double[] { 0, 0, 0 };
                }

                Matrix<float> line = new Matrix<float>(new Size(1, 4));
                CvInvoke.FitLine(new VectorOfPoint(points.ToArray()), line, Emgu.CV.CvEnum.DistType.L2, 0, 0.001, 0.001);
                float x0 = line[0, 0];
                float x1 = line[1, 0];
                float x2 = line[2, 0];
                float x3 = line[3, 0];
                float k = 0;
                if (x0 < 0.000001)
                {
                    return new double[] { 1, 0, -x2 };
                }
                else
                {
                    k = x1 / x0;
                    double b = x3 - k * x2;
                    return new double[] { k, -1, b };
                }
            }
            catch (Exception e)
            {
                Logger.WriteLineError("MathTools2D CalculateFittedLine error," + e.Message + "," + e.StackTrace);
                return result;
            }
        }

        /// <summary>
        /// 计算一系列点的拟合直线,Ax+By+C=0,B=0或-1,不去除异常点
        /// </summary>
        /// <param name="points"></param>
        /// <returns></returns>
        public static double[] CalculateFittedLineNoRemovePoint(List<Point> points)
        {
            double[] result = { 0, 0, 0 };
            try
            {
                var count = points.Count;
                if (count < 2)
                {
                    return new double[] { 0, 0, 0 };
                }

                Matrix<float> line = new Matrix<float>(new Size(1, 4));
                CvInvoke.FitLine(new VectorOfPoint(points.ToArray()), line, Emgu.CV.CvEnum.DistType.L2, 0, 0.001, 0.001);
                double x0 = line[0, 0];
                double x1 = line[1, 0];
                double x2 = line[2, 0];
                double x3 = line[3, 0];
                double k = 0;
                if (x0 < 0.000001)
                {
                    return new double[] { 1, 0, -x2 };
                }
                else
                {
                    k = x1 / x0;
                    double b = x3 - k * x2;
                    return new double[] { k, -1, b };
                }
            }
            catch (Exception e)
            {
                Logger.WriteLineError("MathTools2D CalculateFittedLineNoRemovePoint error," + e.Message + "," + e.StackTrace);
                return result;
            }
        }

        ///// <summary>
        /////  计算一系列点的拟合直线
        ///// </summary>
        ///// <param name="arrayX"></param>
        ///// <param name="arrayY"></param>
        ///// <returns></returns>
        //public static double[] CalculateFittedLine(double[] arrayX, double[] arrayY)
        //{
        //    try
        //    {
        //        double[] result = { 0, 0, 0 };
        //        double slope = 0;
        //        if (arrayX.Length == arrayY.Length)
        //        {
        //            var averX = arrayX.Average();
        //            var averY = arrayY.Average();

        //            double molecular = 0;
        //            double denominator = 0;
        //            for (int i = 0; i < arrayX.Length; i++)
        //            {
        //                molecular += (arrayX[i] - averX) * (arrayY[i] - averY);
        //                denominator += Math.Pow(arrayX[i] - averX, 2);
        //            }

        //            if (Math.Abs(denominator) < MathToolsBasis.Precision)
        //            {
        //                return new double[] { 1, 0, -averX };
        //            }
        //            else
        //            {
        //                slope = molecular / denominator;
        //            }

        //            result[0] = slope;
        //            result[1] = -1;
        //            result[2] = averY - result[0] * averX;
        //            if (double.IsNaN(result[0]))
        //            {
        //                result[0] = 0;
        //            }
        //            if (double.IsNaN(result[2]))
        //            {
        //                result[2] = 0;
        //            }
        //        }
        //        return result;
        //    }
        //    catch (Exception e)
        //    {
        //        Logger.WriteLineError("MathTools2D CalculateFittedLine error," + e.Message + "," + e.StackTrace);
        //        return new double[3] { 0, 0, 0 };
        //    }
        //}

        /// <summary>
        /// 获得一系列二维点的中点
        /// </summary>
        /// <param name="endPointList2D"></param>
        /// <returns></returns>
        public static Point GetCenterPoint(List<Vector2I> endPointList2D)
        {
            var count = endPointList2D.Count;
            double x = 0;
            double y = 0;
            for (var i = 0; i < count; ++i)
            {
                x += endPointList2D[i].X;
                y += endPointList2D[i].Y;
            }
            return new Point((int)Math.Round(x / count), (int)Math.Round(y / count));
        }

        /// <summary>
        /// 计算二维向量夹角的余弦值
        /// </summary>
        /// <param name="v1"></param>
        /// <param name="v2"></param>
        /// <returns></returns>
        public static double GetCosA2D(Vector2 v1, Vector2 v2)
        {
            if (v1.Equals(new Vector2(0, 0)))
            {
                return 1;
            }
            if (v2.Equals(new Vector2(0, 0)))
            {
                return 1;
            }
            var d1 = Math.Sqrt(Math.Abs(v1.X * v1.X + v1.Y * v1.Y));
            var d2 = Math.Sqrt(Math.Abs(v2.X * v2.X + v2.Y * v2.Y));
            if (Math.Abs(d1) < MathToolsBasis.Precision || Math.Abs(d2) < MathToolsBasis.Precision)
            {
                return 1;
            }
            return (v1.X * v2.X + v1.Y * v2.Y) / (d1 * d2);
        }

        /// <summary>
        /// 计算二维向量夹角的余弦值
        /// </summary>
        /// <param name="v1"></param>
        /// <param name="v2"></param>
        /// <returns></returns>
        public static double GetCosA2D(Vector2I v1, Vector2I v2)
        {
            if (v1.Equals(new Vector2I(0, 0)))
            {
                return 1;
            }
            if (v2.Equals(new Vector2I(0, 0)))
            {
                return 1;
            }

            var d1 = Math.Sqrt(Math.Abs(v1.X * v1.X + v1.Y * v1.Y));
            var d2 = Math.Sqrt(Math.Abs(v2.X * v2.X + v2.Y * v2.Y));
            if (Math.Abs(d1) < MathToolsBasis.Precision || Math.Abs(d2) < MathToolsBasis.Precision)
            {
                return 1;
            }
            return (v1.X * v2.X + v1.Y + v2.Y) / (d1 * d2);
        }

        /// <summary>
        /// 获得二维点旋转后的点坐标
        /// </summary>
        /// <param name="point"></param>
        /// <param name="mapMatrix">旋转矩阵</param>
        /// <returns></returns>
        public static Vector2 GetRotatedPoint(Point point, Matrix<double> mapMatrix)
        {
            try
            {
                var a1 = mapMatrix[0, 0];
                var a2 = mapMatrix[0, 1];
                var a3 = mapMatrix[1, 0];
                var a4 = mapMatrix[1, 1];
                var t1 = mapMatrix[0, 2];
                var t2 = mapMatrix[1, 2];

                var x0 = point.X;
                var y0 = point.Y;
                var xTemp = a1 * x0 + a2 * y0 + t1;
                var x = xTemp > 0 ? MathToolsBasis.GetUpperInt(xTemp) : MathToolsBasis.GetLowerInt(xTemp);
                var yTemp = a3 * x0 + a4 * y0 + t2;
                var y = yTemp > 0 ? MathToolsBasis.GetUpperInt(yTemp) : MathToolsBasis.GetLowerInt(yTemp);
                return new Vector2(x, y);
            }
            catch (Exception e)
            {
                Logger.WriteLineError("MathTools2D GetRotatedPoint error," + e.Message + "," + e.StackTrace);
                return new Vector2(point.X, point.Y);
            }
        }

        public static List<Point> ChangeVector2IToPoint(List<Vector2I> points)
        {
            var count = points.Count;
            if (count == 0)
            {
                return new List<Point>();
            }
            var pointsList = new List<Point>();
            foreach (var point in points)
            {
                pointsList.Add(new Point(point.X, point.Y));
            }
            return pointsList;
        }

        /// <summary>
        /// 获得二维点旋转后的点坐标
        /// </summary>
        /// <param name="point"></param>
        /// <param name="mapMatrix">旋转矩阵</param>
        /// <returns></returns>
        public static List<Vector2> GetRotatedPoints(List<Point> pointList, Matrix<double> mapMatrix)
        {
            try
            {
                var a1 = mapMatrix[0, 0];
                var a2 = mapMatrix[0, 1];
                var a3 = mapMatrix[1, 0];
                var a4 = mapMatrix[1, 1];
                var t1 = mapMatrix[0, 2];
                var t2 = mapMatrix[1, 2];

                var rotatedPointsList = new List<Vector2>();
                foreach (var point in pointList)
                {
                    var x0 = point.X;
                    var y0 = point.Y;
                    var xTemp = a1 * x0 + a2 * y0 + t1;
                    var x = xTemp > 0 ? MathToolsBasis.GetUpperInt(xTemp) : MathToolsBasis.GetLowerInt(xTemp);
                    var yTemp = a3 * x0 + a4 * y0 + t2;
                    var y = yTemp > 0 ? MathToolsBasis.GetUpperInt(yTemp) : MathToolsBasis.GetLowerInt(yTemp);
                    rotatedPointsList.Add(new Vector2(x, y));
                }
                return rotatedPointsList;
            }
            catch (Exception e)
            {
                Logger.WriteLineError("MathTools2D GetRotatedPoint error," + e.Message + "," + e.StackTrace);
                return new List<Vector2>();
            }
        }

        /// <summary>
        /// 获得点列的最小x，y，最大x，y，即横坐标和纵坐标的取值范围
        /// </summary>
        /// <param name="endPointList2"></param>
        /// <returns></returns>
        public static CoordinateRange GetCoordinateRange(List<Vector2> endPointList2)
        {
            double minY = int.MaxValue;
            double minX = int.MaxValue;
            double maxX = int.MinValue;
            double maxY = int.MinValue;

            for (var i = 0; i < endPointList2.Count; ++i)
            {
                if (endPointList2[i].X < minX)
                {
                    minX = endPointList2[i].X;
                }
                if (endPointList2[i].X > maxX)
                {
                    maxX = endPointList2[i].X;
                }
                if (endPointList2[i].Y < minY)
                {
                    minY = endPointList2[i].Y;
                }
                if (endPointList2[i].Y > maxY)
                {
                    maxY = endPointList2[i].Y;
                }
            }
            return new CoordinateRange
            {
                MinX = minX,
                MinY = minY,
                MaxX = maxX,
                MaxY = maxY
            };
        }
        /// <summary>
        /// 获得点列y值最大的点的索引
        /// </summary>
        /// <param name="endPointList2"></param>
        /// <returns></returns>
        public static int GetTopPointIndex(List<Vector2> endPointList2)
        {
            var count = endPointList2.Count;
            if (count == 0)
            {
                return 0;
            }
            var index = 0;
            double maxY = int.MaxValue;
            for (var i = 0; i < count; ++i)
            {
                if (endPointList2[i].Y < maxY)
                {
                    index = i;
                    maxY = endPointList2[i].Y;
                }
            }
            return index;
        }

        /// <summary>
        /// 计算多边形,各个边的直线的方程
        /// </summary>
        /// <param name="imagePointList">多边形端点的list</param>
        /// <returns></returns>
        public static List<LineEquation> GetPolygonEdgeLineEquation(List<Vector2I> imagePointList)
        {
            try
            {
                var lineList = new List<LineEquation>();
                //求直线方程,Ax+By+C=0;
                var count = imagePointList.Count;
                //多边形最少有三个边
                if (count < 2)
                {
                    return new List<LineEquation>();
                }
                for (var i = 0; i < count; ++i)
                {
                    var lineEquation = GetLineEquation(i == 0 ? imagePointList[count - 1] : imagePointList[i - 1], imagePointList[i]);
                    //直线端点的前一个点
                    Vector2I pointTemp;
                    switch (i)
                    {
                        case 1:
                            pointTemp = imagePointList[count - 1];
                            break;
                        case 0:
                            pointTemp = imagePointList[count - 2];
                            break;
                        default:
                            pointTemp = imagePointList[i - 2];
                            break;
                    }
                    //取反，为了判断方便判断图像点是不是在多边形内
                    if (lineEquation.A * pointTemp.X + lineEquation.B * pointTemp.Y + lineEquation.C < 0)
                    {
                        lineEquation = new LineEquation
                        {
                            A = -lineEquation.A,
                            B = -lineEquation.B,
                            C = -lineEquation.C
                        };
                    }
                    lineList.Add(lineEquation);
                }
                return lineList;
            }
            catch (Exception e)
            {
                Logger.WriteLineError("MathTools2D GetPolygonEdgeLineEquation error," + e.Message + "," +
                                      e.StackTrace);
                return new List<LineEquation>();
            }
        }

        /// <summary>
        /// 根据两个点求直线方程
        /// </summary>
        /// <param name="point1"></param>
        /// <param name="point2"></param>
        /// <returns></returns>
        public static LineEquation GetLineEquation(Vector2I point1, Vector2I point2)
        {
            try
            {
                if (Math.Abs(point1.X - point2.X) < MathToolsBasis.Precision && Math.Abs(point1.Y - point2.Y) < MathToolsBasis.Precision)
                {
                    return new LineEquation { A = 0, B = 0, C = 0 };
                }

                if (Math.Abs(point1.X - point2.X) < MathToolsBasis.Precision)
                {
                    return new LineEquation { A = 1, B = 0, C = -point1.X };
                }

                if (Math.Abs(point1.Y - point2.Y) < MathToolsBasis.Precision)
                {
                    return new LineEquation { A = 0, B = 1, C = -point1.Y };
                }

                var tempA = (point1.Y - point2.Y) / (float)(point1.X - point2.X);
                var tempC = point1.Y - tempA * point1.X;
                return new LineEquation { A = tempA, B = -1, C = tempC };
            }
            catch (Exception e)
            {
                Logger.WriteLineError("MathTools2D GetLineEquation error," + e.Message + "," + e.StackTrace);
                return new LineEquation { A = 0, B = 0, C = 0 };
            }
        }

        /// <summary>
        /// 判断点是否在多边形内，-1不在，0在多边形上，1在多边形内
        /// </summary>
        /// <param name="point"></param>
        /// <param name="polygonEdgeLines"></param>
        /// <returns></returns>
        public static int IsIntPolygon(Vector2I point, List<LineEquation> polygonEdgeLines)
        {
            var linesNum = polygonEdgeLines.Count;
            if (linesNum < 3)
            {
                return 0;
            }
            //判断点是否在多边形内
            for (var i = 0; i < linesNum; ++i)
            {
                var temp = polygonEdgeLines[i].A * point.X + polygonEdgeLines[i].B * point.Y + polygonEdgeLines[i].C;
                if (temp < MathToolsBasis.NegPrecision)
                {
                    return -1;
                }
                if (Math.Abs(temp) < MathToolsBasis.Precision)
                {
                    return 0;
                }
            }
            return 1;
        }

        /// <summary>
        /// 计算两点间的距离
        /// </summary>
        /// <param name="pointA"></param>
        /// <param name="pointB"></param>
        /// <returns></returns>
        public static double GetTwoPointDistance(Vector2I pointA, Vector2I pointB)
        {
            var disX = pointA.X - pointB.X;
            var disY = pointA.Y - pointB.Y;
            return Math.Sqrt(disX * disX + disY * disY);
        }

        /// <summary>
        /// 获得图像的四个端点，顺时针方向
        /// </summary>
        /// <param name="imageSize"></param>
        /// <returns></returns>
        public static List<Point> GetImageEndPoint(Size imageSize)
        {
            //旋转前图像端点
            return new List<Point>
            {
                new Point(0, 0),
                new Point(imageSize.Width - 1, 0),
                new Point(imageSize.Width - 1,  imageSize.Height- 1),
                new Point(0,  imageSize.Height - 1)
            };
        }

        /// <summary>
        /// 计算图像的旋转矩阵
        /// </summary>
        /// <param name="center">旋转中心</param>
        /// <param name="rotateAngle">大于0，逆时针旋转，小于0，顺时针旋转</param>
        /// <param name="imageSrcSize"></param>
        /// <returns></returns>
        public static Matrix<double> GetImageMapMatrix(Point center, double rotateAngle, Size imageSrcSize)
        {
            //旋转前图像端点
            var endPointListSrc = GetImageEndPoint(imageSrcSize);
            var matrixTemp = new Matrix<double>(2, 3);
            //angle是角度，不是pi值，大于0代表逆时针旋转
            //转换公式[x,y]=[a1,a2,t1][a3,a4,t2]*[x0,y0,1]
            CvInvoke.GetRotationMatrix2D(new PointF(center.X, center.Y), rotateAngle, 1, matrixTemp);
            //旋转后图像端点
            var endPointList = new List<Vector2>();
            //旋转后，坐标系Y轴方向向上 
            foreach (var onePoint in endPointListSrc)
            {
                endPointList.Add(GetRotatedPoint(onePoint, matrixTemp));
            }

            var minX = endPointList.Min(o => o.X) - 1;
            var minY = endPointList.Min(o => o.Y) - 1;
            //平移图像,否则有部分图像显示不出来
            matrixTemp[0, 2] = matrixTemp[0, 2] - minX;
            matrixTemp[1, 2] = matrixTemp[1, 2] - minY;
            return matrixTemp;
        }

    }
}
