﻿using Emgu.CV;
using Emgu.CV.Structure;
using SkiaSharp;
using System;
using System.Drawing;
using WingAIDiagnosisService.Carotid.Utilities.DetectIntima;
using WingServerCommon.Log;

namespace WingAIDiagnosisService.Carotid.Utilities
{
    class AIClipImageStorer
    {
        private const string DigitalPostfix = "{0}  mm";
        private const string Title = "1 Intima Detection";
        private const string PostMax = "Post.Max";
        private const string PostMin = "Post.Min";
        private const string PostAvg = "Post.Average";
        private const int TitleLeft = 20;
        private const int TitleTop = 20;

        /// <summary>
        /// Save image
        /// </summary>
        /// <param name="filePath">file to save</param>
        /// <param name="srcImage">Source image</param>
        /// <param name="intimaResult">intima result</param>
        /// <param name="intimaRect">intima rectangle</param>
        /// <returns>True if success</returns>
        public static bool SaveImage(string filePath, Image<Gray, byte> srcImage, CarotidIntimaMeasureResult intimaResult, IntimaRect intimaRect)
        {
            ImageTools.Adjust(srcImage.Mat);
            using (var imageBgr = srcImage.Convert<Bgra, byte>())
            {
                try
                {
                    var lowerRect = new Rectangle(intimaRect.Left, intimaRect.Top, intimaRect.Width, intimaRect.Height);
                    var dstRect = ChangeRect(imageBgr.Height, lowerRect);
                    //画下内膜绿色roi框
                    imageBgr.Draw(dstRect, new Bgra(0, 255, 0, 0), 2);
                    string maxStr = CarotidIntimaMeasureResult.UnifyPrecision(intimaResult.MaxThickness).ToString("0.00");
                    string minStr = CarotidIntimaMeasureResult.UnifyPrecision(intimaResult.MinThickness).ToString("0.00");
                    string averageStr = CarotidIntimaMeasureResult.UnifyPrecision(intimaResult.AverageThickness).ToString("0.00");
                    using (var textAreaImage = DrawTextArea(new Size(imageBgr.Width, srcImage.Height), maxStr, minStr, averageStr))
                    {
                        SaveImage(imageBgr, textAreaImage, filePath);
                        return true;
                    }
                }
                catch (Exception e)
                {
                    Logger.WriteLineError($"Draw carotid measure result error: {e}");
                }
            }

            return false;
        }

        private static Image<Bgra, byte> DrawTextArea(Size srcSize, string maxStr, string minStr, string averageStr)
        {
            SKImageInfo info = new SKImageInfo(srcSize.Width, srcSize.Height, SKColorType.Bgra8888);
            using (var skBitmap = new SKBitmap(info, info.Width * 4))
            {
                using var sKPaint = new SKPaint();
                sKPaint.Color = new SKColor(255, 255, 0, 255);
                sKPaint.TextSize = 25;
                sKPaint.Typeface = SKTypeface.FromFamilyName("Arial", (int)sKPaint.TextSize, (int)sKPaint.TextSize, SKFontStyleSlant.Upright);
                sKPaint.FakeBoldText = true;
                SKRect sizeMaxStrTemp = new SKRect();
                sKPaint.MeasureText(string.Format(DigitalPostfix, maxStr), ref sizeMaxStrTemp);

                SKRect sizeMinStrTemp = new SKRect();
                sKPaint.MeasureText(string.Format(DigitalPostfix, minStr), ref sizeMinStrTemp);

                SKRect sizeAverageStrTemp = new SKRect();
                sKPaint.MeasureText(string.Format(DigitalPostfix, averageStr), ref sizeAverageStrTemp);

                using SKCanvas canvas = new SKCanvas(skBitmap);
                //Draw title
                int marginHeight = (int)sKPaint.TextSize / 2 + (int)sKPaint.TextSize;
                int lineTop = TitleTop + marginHeight;
                canvas.DrawText(Title, TitleLeft, lineTop, sKPaint);
                int lineLeft = TitleLeft + (int)sKPaint.TextSize;
                int lineRight = lineLeft + 290;
                //draw line PostMax
                lineTop += marginHeight;
                canvas.DrawText(PostMax, lineLeft, lineTop, sKPaint);
                canvas.DrawText(string.Format(DigitalPostfix, maxStr), lineRight - sizeMaxStrTemp.Width, lineTop, sKPaint);
                //draw line PostMin
                lineTop += marginHeight;
                canvas.DrawText(PostMin, lineLeft, lineTop, sKPaint);
                canvas.DrawText(string.Format(DigitalPostfix, minStr), lineRight - sizeMinStrTemp.Width, lineTop, sKPaint);
                //draw line PostAvg
                lineTop += marginHeight;
                canvas.DrawText(PostAvg, lineLeft, lineTop, sKPaint);
                canvas.DrawText(string.Format(DigitalPostfix, averageStr), lineRight - sizeAverageStrTemp.Width, lineTop, sKPaint);
                canvas.Flush();
                var roiRect = new Rectangle(0, TitleTop, lineRight + 20, lineTop);
                using (var image = new Image<Bgra, byte>(skBitmap.Width, skBitmap.Height))
                {
                    image.Bytes = skBitmap.Bytes;
                    image.ROI = roiRect;
                    var cropedImage = new Image<Bgra, byte>(image.ROI.Width, image.ROI.Height);
                    image.CopyTo(cropedImage);
                    return cropedImage;
                }
            }
        }

        private static void SaveImage(Image<Bgra, byte> dstImage, Image<Bgra, byte> textAreaImage, string filePath)
        {
            //取出要保存图像的左下角测量信息框
            var rect = new Rectangle(new Point(10, dstImage.Height - textAreaImage.Height - 10), textAreaImage.Size);
            using (var measureInfoRect = dstImage.GetSubRect(rect))
            {
                MergeImage(textAreaImage, measureInfoRect);
                CvInvoke.Rectangle(dstImage, rect, new MCvScalar(100, 100, 100, 255));
                dstImage.Save(filePath);
            }
        }

        private static void MergeImage(Image<Bgra, byte> srcImage, Image<Bgra, byte> dstImage)
        {
            var width = srcImage.Width;
            var height = srcImage.Height;
            for (var h = 0; h < height; ++h)
            {
                for (var w = 0; w < width; ++w)
                {
                    //只获取黄色字
                    if ((srcImage[h, w].Green == 255) && (srcImage[h, w].Red == 255) && (srcImage[h, w].Blue == 0))
                    {
                        dstImage[h, w] = srcImage[h, w];
                    }
                }
            }
        }

        private static Rectangle ChangeRect(int imageHeight, Rectangle rect)
        {
            var changeValue = 10;
            var top = Math.Max(0, rect.Top - changeValue);
            var bottom = Math.Min(imageHeight - 1, rect.Bottom + changeValue);
            return new Rectangle(rect.Left, top, rect.Width, bottom - top);
        }
    }
}
