﻿using System;
using System.Collections.Generic;
using System.Linq;
using Emgu.CV;
using Emgu.CV.Structure;
using Emgu.CV.Util;
using System.Drawing;
using WingAIDiagnosisService.Carotid.MathTools;
using WingServerCommon.Log;

namespace WingAIDiagnosisService.Carotid.Utilities.DetectIntima
{
   
    
    /// <summary>
    /// 颈动脉检测roi区域
    /// </summary>
    public class IntimaRect
    {
        public IntimaRect(int left, int top, int width, int height)
        {
            Left = left;
            Top = top;
            Width = width;
            Height = height;
        }

        public int Left { get; }
        public int Top { get; }
        public int Width { get; }
        public int Height { get; }
    }

    public abstract class FilterPointResult
    {
        public int InvalidNumber;

        public IReadOnlyList<Point> Points { get; private set; }

        protected FilterPointResult(IReadOnlyList<Point> points)
        {
            Points = points;
            InvalidNumber = 0;
        }

        public void ResetPoints(IReadOnlyList<Point> points)
        {
            Points = points;
        }
    }

    public class SmoothResult : FilterPointResult
    {
        public SmoothResult(IReadOnlyList<Point> points) : base(points)
        {
        }
    }

    public class RemovePitResult : FilterPointResult
    {
        public RemovePitResult(IReadOnlyList<Point> points) : base(points)
        {
        }
    }

    /// <summary>
    /// image process change state
    /// </summary>
    public class ImageProcessChangeState
    {
        /// <summary>
        /// is brightness changed
        /// </summary>
        public bool BrightnessChanged { get; private set; }

        /// <summary>
        /// is contrast changed 
        /// </summary>
        public bool ContrastChanged { get; private set; }

        /// <summary>
        /// is sharpness changed 
        /// </summary>
        public bool SharpnessChanged { get; private set; }

        public ImageProcessChangeState(bool brightnessChanged, bool contrastChanged, bool sharpnessChanged)
        {
            BrightnessChanged = brightnessChanged;
            ContrastChanged = contrastChanged;
            SharpnessChanged = sharpnessChanged;
        }

        public void Reset(bool brightnessChanged, bool contrastChanged, bool sharpnessChanged)
        {
            BrightnessChanged = brightnessChanged;
            ContrastChanged = contrastChanged;
            SharpnessChanged = sharpnessChanged;
        }
    }

    /// <summary>
    /// carotid auto measure parameter
    /// </summary>
    public class CarotidIntimaMeasureParameter
    {
        /// <summary>
        /// measure image
        /// </summary>
        public Image<Gray,byte> Image { get; }

        /// <summary>
        /// carotid measure rect roi
        /// </summary>
        public IntimaRect SelectRoi { get; }

        /// <summary>
        /// is image process(brightness,contrast,sharpness) changed 
        /// </summary>
        public ImageProcessChangeState ImageProcessState { get; }

        /// <summary>
        /// Get physical distance per pixel
        /// </summary>
        public float PhysicalPerPixel { get; }

        public CarotidIntimaMeasureParameter(Image<Gray,byte> image, IntimaRect selectRoi, ImageProcessChangeState imageProcessState, float physicalPerPixel)
        {
            Image = image;
            SelectRoi = selectRoi;
            ImageProcessState = imageProcessState;
            PhysicalPerPixel = physicalPerPixel;
        }
    }

    /// <summary>
    /// 颈动脉检测过程错误码
    /// </summary>
    public enum CarotidIntimaMeasureErrorCode
    {
        Success = 1000,
        Failed,
    }

    /// <summary>
    /// 测量结果返回值
    /// </summary>
    public class CarotidIntimaMeasureResult
    {
        /// <summary>
        /// 错误代码
        /// </summary>
        public CarotidIntimaMeasureErrorCode ErrorCode { get; }
        /// <sumary>
        /// 内膜的上边缘点
        /// </sumary>
        public List<Point> PointUpper { get; }
        ///<sumary>
        ///内膜的上边缘点
        /// </sumary>
        public List<Point> PointLower { get; }
        ///<sumary>
        ///最大厚度
        ///</sumary>
        public float MaxThickness { get; }
        ///<sumary>
        ///最小厚度
        /// </sumary>
        public float MinThickness { get; }
        ///<sumary>
        ///平均厚度
        /// </sumary>
        public float AverageThickness { get; }
        ///<sumary>
        ///壁厚度误差，无效点除以所有点个数
        ///</sumary>
        public float SdThickness { get; }
        ///<sumary>
        ///壁有效点个数
        ///</sumary>
        public float ValidPointsNum { get; }

        /// <summary>
        /// 内膜厚度标准差
        /// </summary>
        public float ThicknessStd { get; }

        public CarotidIntimaMeasureResult(CarotidIntimaMeasureErrorCode errorCode, List<Point> pointUpper,
            List<Point> pointLower,
            float maxThickness,
            float minThickness,
            float averageThickness,
            float sdThickness,
            float validPointsNum,
            float thicknessStd)
        {
            PointUpper = pointUpper;
            PointLower = pointLower;
            MaxThickness = maxThickness;
            MinThickness = minThickness;
            AverageThickness = averageThickness;
            SdThickness = sdThickness;
            ValidPointsNum = validPointsNum;
            ErrorCode = errorCode;
            ThicknessStd = thicknessStd;
        }

        public CarotidIntimaMeasureResult(CarotidIntimaMeasureErrorCode errorCode)
        {
            ErrorCode = errorCode;
        }

        public static float UnifyPrecision(float value)
        {
            return float.Parse(Math.Round(value, 2, MidpointRounding.AwayFromZero).ToString("0.00"));
        }
    }

    /// <summary>
    /// 图像预处理返回结果
    /// </summary>
    public class ImagePreprocessingResult
    {
        public Image<Gray, byte> ImageRoiBinary;
        public Image<Gray, byte> ImageRoiCanny;
        public bool IsSuccess;

        public ImagePreprocessingResult(Image<Gray, byte> imageRoiBinary, Image<Gray, byte> imageRoiCanny, bool isSuccess) : this(isSuccess)
        {
            ImageRoiBinary = imageRoiBinary;
            ImageRoiCanny = imageRoiCanny;
        }

        public ImagePreprocessingResult(bool isSuccess)
        {
            IsSuccess = isSuccess;
        }
    }


    /// <summary>
    /// 颈动脉内膜自动测量父类
    /// </summary>
    public class CarotidIntimaMeasurement
    {
        //标准最大内膜壁厚度1.2mm
        protected const float LowerWallMaxIntimaThick = 0.12f;
        //标准最大内膜壁厚度0.1mm
        protected const float UpperWallMaxIntimaThick = 0.1f;
        // 标准最小内膜壁厚度0.3mm
        protected const float LowerWallMinIntimaThick = 0.03f;
        //标准最小内膜壁厚度0.1mm
        protected const float UpperMinIntimaThick = 0.01f;
        //内膜最小厚度,1个像素
        protected int MinIntimaThickPix = 1;
        //测量区域Roi的最小高
        protected int MinRoiHeight = 5;
        //颈动脉下内膜上边缘微调范围
        protected int LowerIntimaUpperEdgeTrimRange = 6;
        // 最大内膜厚度(像素)
        protected int MaxIntimaThickPix;
        //1cm对应的图像像素个数
        protected float PhysicalPerPixel;
        /// 二值化阈值
        protected int ImageThreshold;

        /// <summary>
        /// 求轮廓点的方差
        /// </summary>
        /// <param name="vectorPoint"></param>
        /// <returns></returns>
        protected double GetVariance(VectorOfPoint vectorPoint)
        {
            double sum = 0;
            var count = vectorPoint.Size;
            for (var i = 0; i < count; ++i)
            {
                sum += vectorPoint[i].Y;
            }
            var average = sum / count;

            sum = 0;
            for (var i = 0; i < count; ++i)
            {
                sum += (vectorPoint[i].Y - average) * (vectorPoint[i].Y - average);
            }
            var variance = sum / count;
            return variance;
        }

        /// <summary>
        /// 数据统计结果
        /// </summary>
        protected class DataStatisticsResult
        {
            /// <summary>
            /// 最大厚度
            /// </summary>
            public float MaxThick { get; }

            /// <summary>
            /// 最小厚度
            /// </summary>
            public float MinThick { get; }

            /// <summary>
            /// 平均厚度
            /// </summary>
            public float AverageThick { get; }

            /// <summary>
            /// 无效点比率
            /// </summary>
            public float SdThickness { get; }

            /// <summary>
            /// 有效点数
            /// </summary>
            public float ValidPointsNum { get; }

            /// <summary>
            /// 内膜厚度标准差
            /// </summary>
            public float ThicknessStd;

            /// <summary>
            /// 错误码
            /// </summary>
            public CarotidIntimaMeasureErrorCode ErrorCode;

            public DataStatisticsResult(CarotidIntimaMeasureErrorCode errorCode, float maxThick, float minThick, float averageThick, float sdThickness, float validPointsNum, float thicknessStd) : this(errorCode)
            {
                MaxThick = maxThick;
                MinThick = minThick;
                AverageThick = averageThick;
                SdThickness = sdThickness;
                ValidPointsNum = validPointsNum;
                ThicknessStd = thicknessStd;
            }

            public DataStatisticsResult(CarotidIntimaMeasureErrorCode errorCode)
            {
                ErrorCode = errorCode;
            }
        }

        /// <summary>
        /// 数据相关属性统计
        /// </summary>
        /// <param name="upperPoint"></param>
        /// <param name="lowerPoint"></param>
        /// <param name="invalidPitPointsNum"></param>
        /// <param name="invalidSmoothedPointNum"></param>
        /// <returns></returns>
        protected DataStatisticsResult GetDataStatistics(List<Point> upperPoint, List<Point> lowerPoint, int invalidPitPointsNum, int invalidSmoothedPointNum)
        {
            try
            {
                if (upperPoint.Count == 0)
                {
                    return new DataStatisticsResult(CarotidIntimaMeasureErrorCode.Failed);
                }
                var diffThick = new int[lowerPoint.Count];
                var sumThick = 0;
                for (var i = 0; i < diffThick.Length; ++i)
                {
                    diffThick[i] = Math.Abs(lowerPoint[i].Y - upperPoint[i].Y);
                    sumThick += diffThick[i];
                }
                var avg = sumThick / (float)diffThick.Length;
                var allPointNum = upperPoint.Count + lowerPoint.Count;
                //无效点和所有的比
                float sdThickness;
                if (allPointNum == 0)
                {
                    sdThickness = 0;
                }
                else
                {
                    sdThickness = (invalidPitPointsNum + invalidSmoothedPointNum) / (float)allPointNum;
                }

                //内膜厚度标准差
                var std = MathToolsBasis.GetStandardDeviation(diffThick);
                return new DataStatisticsResult(CarotidIntimaMeasureErrorCode.Success, diffThick.Max(), diffThick.Min(), avg, sdThickness, 2 * diffThick.Length, std);
            }
            catch (Exception e)
            {
                Logger.WriteLineError("CarotidIntimaMeasureResult GetDataStatistics error." + e.Message + "," + e.StackTrace);
                return new DataStatisticsResult(CarotidIntimaMeasureErrorCode.Failed, 0, 0, 0, 0, 0, 0);
            }
        }

        /// <summary>
        /// 转换成Rectangle
        /// </summary>
        /// <param name="selectRect"></param>
        /// <param name="imageSize"></param>
        /// <returns></returns>
        protected Rectangle GetImageRoi(IntimaRect selectRect, Size imageSize)
        {
            var left = selectRect.Left;
            var top = selectRect.Top;
            var width = selectRect.Width;
            var height = selectRect.Height;

            //防止图像roi越界
            if (left + width > imageSize.Width)
            {
                width = imageSize.Width - 1 - left;
            }
            if (top + height > imageSize.Height)
            {
                height = imageSize.Height - 1 - top;
            }
            return new Rectangle(left, top, width, height);
        }

        /// <summary>
        /// 去除杂质
        /// </summary>
        /// <param name="image"></param>
        protected void RemoveImpurity(Image<Gray, byte> image)
        {
            //去掉边缘周围的杂质----------------------------------------------------------
            //求连通域
            var contours = ImageTools.FindContours(image);
            var minWidth = image.Width / 2;
            //去掉较小的连通域
            var size = contours.Size;

            for (var i = 0; i < size; ++i)
            {
                var contour = contours[i];
                var rect = CvInvoke.BoundingRectangle(contour);
                if (rect.Width >= minWidth) continue;
                var points = contour.ToArray();
                if (rect.Height > rect.Width)
                {
                    ImageTools.ChangeImageH(image, points, rect);
                }
                else
                {
                    ImageTools.ChangeImageW(image, points, rect);
                }
            }
        }

    }
}
