﻿using System;
using Emgu.CV;
using Emgu.CV.CvEnum;
using Emgu.CV.Structure;
using System.Drawing;
using WingServerCommon.Log;

namespace WingAIDiagnosisService.Carotid.Utilities.DetectIntima
{
    public class CarotidIntimaMeasurementBothResult
    {
        /// <summary>
        /// 错误代码
        /// </summary>
        public CarotidIntimaMeasureErrorCode ErrorCode { get; }

        /// <summary>
        /// 上壁检测结果
        /// </summary>
        public CarotidIntimaMeasureResult UpperMeasureResult { get; }
        /// <summary>
        /// 下壁检测结果
        /// </summary>
        public CarotidIntimaMeasureResult LowerMeasureResult { get; }
        public CarotidIntimaMeasurementBothResult(CarotidIntimaMeasureErrorCode errorCode, CarotidIntimaMeasureResult upperMeasureResult, CarotidIntimaMeasureResult lowerMeasureResult) : this(errorCode)
        {
            UpperMeasureResult = upperMeasureResult;
            LowerMeasureResult = lowerMeasureResult;
        }
        public CarotidIntimaMeasurementBothResult(CarotidIntimaMeasureErrorCode errorCode)
        {
            ErrorCode = errorCode;
        }
    }

    /// <summary>
    /// 颈动脉内膜自动测量
    /// </summary>
    public class CarotidIntimaMeasurementBoth : CarotidIntimaMeasurement
    {
        /// <summary>
        /// 上和下颈动脉内膜壁的自动测量
        /// </summary>
        /// <param name="measureParameter"></param>
        /// <returns></returns>
        public CarotidIntimaMeasurementBothResult AutoMeasure(CarotidIntimaMeasureParameter measureParameter)
        {
            try
            {
                var rectRoi = measureParameter.SelectRoi;
                if (rectRoi.Width < 3 || rectRoi.Height < 3)
                {
                    Logger.WriteLineError("CarotidIntimaMeasurementBoth MeasureApi error." + "Selected roi is too small.");
                    return new CarotidIntimaMeasurementBothResult(CarotidIntimaMeasureErrorCode.Failed);
                }
                //防止图像roi越界
                var width = rectRoi.Width;
                var height = rectRoi.Height;
                var imageSize = measureParameter.Image.Size;
                if (rectRoi.Left + width > imageSize.Width)
                {
                    width = imageSize.Width - 1 - rectRoi.Left;
                }
                if (rectRoi.Top + height > imageSize.Height)
                {
                    height = imageSize.Height - 1 - rectRoi.Top;
                }
                var imageRect = new Rectangle(new Point(rectRoi.Left, rectRoi.Top), new Size(width, height));

                Rectangle rectVessel;
                var image = measureParameter.Image;
               
                    //找到血管位置
                rectVessel = GetBloodVesselRect(image.Copy(imageRect));
                var splitHeight = rectVessel.Top + rectVessel.Height / 2;

                //上下壁检测算法
                var rect1 = new Rectangle(new Point(rectRoi.Left, rectRoi.Top), new Size(rectRoi.Width, splitHeight));
                var rect2 = new Rectangle(new Point(rectRoi.Left, rectRoi.Top + splitHeight), new Size(rectRoi.Width, rectRoi.Height - splitHeight));

                var rectRoi1 = new IntimaRect(rect1.X, rect1.Y, rect1.Width, rect1.Height);
                var rectRoi2 = new IntimaRect(rect2.X, rect2.Y, rect2.Width, rect2.Height);

                //上内膜区域过小
                if (rectRoi1.Height < 3)
                {
                    Logger.WriteLineError("CarotidIntimaMeasurementBoth MeasureApi error." + "The height of the selection upperIntima is too small.");
                    return new CarotidIntimaMeasurementBothResult(CarotidIntimaMeasureErrorCode.Failed);
                }
                //下内膜区域过小
                if (rectRoi2.Height < 3)
                {
                    Logger.WriteLineError("CarotidIntimaMeasurementBoth MeasureApi error." + "The height of the selection lowerIntima is too small.");
                    return new CarotidIntimaMeasurementBothResult(CarotidIntimaMeasureErrorCode.Failed);
                }

                var upperMeasure = new CarotidUpperIntimaMeasurement();
                var measureParameterUpper = new CarotidIntimaMeasureParameter(measureParameter.Image, rectRoi1, measureParameter.ImageProcessState, measureParameter.PhysicalPerPixel);
                //先求上壁
                var resultUpper = upperMeasure.MeasureApi(measureParameterUpper);
                if (resultUpper.ErrorCode != CarotidIntimaMeasureErrorCode.Success)
                {
                    return new CarotidIntimaMeasurementBothResult(resultUpper.ErrorCode);
                }
                //求下壁
                var lowerMeasure = new CarotidLowerIntimaMeasurement();
                var measureParameterLower = new CarotidIntimaMeasureParameter(measureParameter.Image, rectRoi2, measureParameter.ImageProcessState, measureParameter.PhysicalPerPixel);
                var resultLower = lowerMeasure.MeasureApi(measureParameterLower);
                if (resultLower.ErrorCode != CarotidIntimaMeasureErrorCode.Success)
                {
                    return new CarotidIntimaMeasurementBothResult(resultLower.ErrorCode);
                }
                return new CarotidIntimaMeasurementBothResult(CarotidIntimaMeasureErrorCode.Success,
                                                                          resultUpper, resultLower);
            }
            catch (Exception e)
            {
                Logger.WriteLineError("CarotidIntimaMeasurementBoth MeasureApi error." + e.Message + "," + e.StackTrace);
                return new CarotidIntimaMeasurementBothResult(CarotidIntimaMeasureErrorCode.Failed);
            }
        }

        /// <summary>
        /// 获得血管区域
        /// </summary>
        /// <param name="imgGray"></param>
        /// <returns></returns>
        private Rectangle GetBloodVesselRect(CvArray<byte> imgGray)
        {
            try
            {
                //图像增强
                CvInvoke.EqualizeHist(imgGray, imgGray);
                var imageBri = new Image<Gray, byte>(imgGray.Width, imgGray.Height);
                var threshold = ImageTools.GetThreshold(imgGray);

                CvInvoke.Threshold(imgGray, imageBri, threshold, 255, ThresholdType.Binary);
                //中值滤波去噪点     
                CvInvoke.MedianBlur(imageBri, imageBri, 3);
                CvInvoke.Threshold(imageBri, imageBri, threshold, 255, ThresholdType.BinaryInv);
                return GetMaxAreaVesselRect(imageBri);
            }
            catch (Exception e)
            {
                Logger.WriteLineError("CarotidIntimaMeasurementBoth GetMaxVesselRect error." + e.Message + "," + e.StackTrace);
                return new Rectangle();
            }
        }

        /// <summary>
        /// 获得面积最大的血管位置
        /// </summary>
        /// <param name="imageBri"></param>
        /// <returns></returns>
        private Rectangle GetMaxAreaVesselRect(Image<Gray, byte> imageBri)
        {
            try
            {
                //求连通域
                var contours = ImageTools.FindContours(imageBri);
                double maxArea = 0;
                var maxIndex = 0;
                for (var i = 0; i < contours.Size; i++)
                {
                    var area = CvInvoke.ContourArea(contours[i]);
                    if (maxArea < area)
                    {
                        maxArea = area;
                        maxIndex = i;
                    }
                }
                return CvInvoke.BoundingRectangle(contours[maxIndex]);
            }
            catch (Exception e)
            {
                Logger.WriteLineError("CarotidIntimaMeasurementBoth GetMaxVesselRect error." + e.Message + "," + e.StackTrace);
                return new Rectangle();
            }
        }

    }
}