﻿using System;
using System.Collections.Generic;
using System.Linq;
using Emgu.CV;
using Emgu.CV.Structure;
using Emgu.CV.CvEnum;
using Emgu.CV.Util;
using System.Drawing;
using WingServerCommon.Log;

namespace WingAIDiagnosisService.Carotid.Utilities.DetectIntima
{
    /// <summary>
    /// 自动测量下内膜时，定义的曲线凹坑
    /// </summary>
    public struct Pit
    {
        public int PitWidth;
        public int PitHeightBegin;
        public int PitHeightEnd;
        public bool IsUpward;
        public Point[] PitEdgePoints;
    }

    /// <summary>
    /// 颈动脉下内膜检测
    /// </summary>
    public class CarotidLowerIntimaMeasurement : CarotidIntimaMeasurement
    {
        //磨平坑时修改的点个数
        private int _invalidPitPointsNum;
        //被平滑掉的点个数
        private int _invalidSmoothedPointNum;
        private CarotidIntimaMeasureParameter _measureParameter;

        private Image<Gray, byte> _imageRoiBinary;
        private Image<Gray, byte> _imageRoiCanny;

        private readonly List<Point> _upperPoint = new List<Point>();
        private readonly List<Point> _lowerPoint = new List<Point>();

        /// <summary>
        /// 下壁颈动脉自动测量对外接口
        /// </summary>
        /// <param name="image"></param>
        /// <param name="selectdedRoi"></param>     
        /// <param name="scanDepth"></param>
        /// <param name="imageHeight"></param>
        /// <returns></returns>
        public CarotidIntimaMeasureResult MeasureApi(CarotidIntimaMeasureParameter measureParameter)
        {
            if (measureParameter.Image == null)
            {
                Logger.WriteLineError("CarotidLowerIntimaMeasurement MeasureApi error.The measure image is null.");
                return new CarotidIntimaMeasureResult(CarotidIntimaMeasureErrorCode.Failed);
            }
            var imageGray = measureParameter.Image.Copy();
            try
            {
                //图像预处理
                var originalImageRectRoi = GetImageRoi(measureParameter.SelectRoi, new Size(imageGray.Width, imageGray.Height));
                //优化图像质量
                OptimizeImageQuality(imageGray, measureParameter.ImageProcessState.SharpnessChanged);
                //图像预处理
                var res = ImagePreprocessingLower(imageGray, originalImageRectRoi, 0, 0);
                if (!res.IsSuccess)
                {
                    return new CarotidIntimaMeasureResult(CarotidIntimaMeasureErrorCode.Failed);
                }

                var result = MeasureLowerWall(res.ImageRoiBinary,
                    res.ImageRoiCanny,
                    originalImageRectRoi,
                    res.ImageRoiBinary.ROI,
                    measureParameter.PhysicalPerPixel);
                return result;
            }
            catch (Exception e)
            {
                Logger.WriteLineError("CarotidLowerIntimaMeasurement MeasureApi error." + e.Message + "," + e.StackTrace);
                return new CarotidIntimaMeasureResult(CarotidIntimaMeasureErrorCode.Failed);
            }
            finally
            {
                imageGray.Dispose();
            }
        }

        /// <summary>
        /// 内膜自动测量用的图像预处理
        /// </summary>
        /// <param name="measureParameter"></param>
        /// <returns></returns>
        public Image<Gray, byte> ImagePreprocessing(CarotidIntimaMeasureParameter measureParameter)
        {
            var imageGray = measureParameter.Image.Copy();
            //优化图像质量
            OptimizeImageQuality(imageGray, measureParameter.ImageProcessState.SharpnessChanged);
            return imageGray;
        }

        /// <summary>
        /// 优化图像质量
        /// </summary>
        /// <param name="imageGray"></param>
        /// <param name="sharpnessChanged"></param>
        private void OptimizeImageQuality(Image<Gray, byte> imageGray, bool sharpnessChanged)
        {
            if (sharpnessChanged == false)
            {
                const int sharpFactor = 1;
                float[,] temp = { { 0, -sharpFactor, 0 }, { -sharpFactor, 1 + 4 * sharpFactor, -sharpFactor }, { 0, -sharpFactor, 0 } };
                var kernel = new ConvolutionKernelF(temp);
                CvInvoke.Filter2D(imageGray, imageGray, kernel, new Point(-1, -1));
            }
            //图像增强
            CvInvoke.EqualizeHist(imageGray, imageGray);
        }

        private Rectangle GetRectInImage(Size imageSize, Rectangle rect)
        {
            if (rect.Left > imageSize.Width - 1 || rect.Left < 0)
            {
                return new Rectangle();
            }
            if (rect.Top > imageSize.Height - 1 || rect.Top < 0)
            {
                return new Rectangle();
            }
            if (rect.Width <0 || rect.Height < 0)
            {
                return new Rectangle();
            }

            var right = rect.Left + rect.Width;
            if (right > imageSize.Width - 1)
            {
                right = imageSize.Width - 1;

            }
            var bootom = rect.Top + rect.Height;
            if (bootom > imageSize.Height - 1)
            {
                bootom = imageSize.Height - 1;
            }

            return new Rectangle(rect.Left, rect.Top, right - rect.Left, bootom - rect.Top);

        }

        /// <summary>
        /// 下壁颈动脉自动测量
        /// </summary>
        /// <param name="rectRoi"></param>
        /// <returns></returns>
        public CarotidIntimaMeasureResult MeasureLowerWall(Image<Gray, byte> imageRoiBinary, Image<Gray, byte> imageRoiCanny, Rectangle originalImageRectRoi, Rectangle rectRoiInput, float physicalPerPixel)
        {
            try
            {
                var rectRoi = GetRectInImage(imageRoiBinary.Size, rectRoiInput);        
                //限制最小roi
                if (rectRoi.Width < 3 || rectRoi.Height < 3)
                {
                    Logger.WriteLineError("CarotidLowerIntimaMeasurement MeasureLowerWall error." + "Selected roi is too small.");
                    return new CarotidIntimaMeasureResult(CarotidIntimaMeasureErrorCode.Failed);
                }
                _upperPoint.Clear();
                _lowerPoint.Clear();
                _invalidPitPointsNum = 0;
                _invalidSmoothedPointNum = 0;
                PhysicalPerPixel = physicalPerPixel;

                _imageRoiBinary = imageRoiBinary.Copy(rectRoi);
                _imageRoiCanny = imageRoiCanny.Copy(rectRoi);

                //PhysicalPerPixel,1 cm 有几个像素 * 内膜最大厚度（单位是像素）
                var normalThick = LowerWallMaxIntimaThick / PhysicalPerPixel;
                MaxIntimaThickPix = Math.Min((int)(1.1 * normalThick), rectRoi.Height);
                MinIntimaThickPix = (int)(LowerWallMinIntimaThick / PhysicalPerPixel);

                //下内膜检测算法                      
                var result1 = GetLowerMeasureCarotidPoint(originalImageRectRoi, ref rectRoi, MaxIntimaThickPix);
                //修改rect后再次回调
                if (result1 == 2)
                {
                    return MeasureLowerWall(_imageRoiBinary, _imageRoiCanny, originalImageRectRoi, rectRoi, PhysicalPerPixel);
                }
                if (result1 != 0)
                {
                    return new CarotidIntimaMeasureResult(CarotidIntimaMeasureErrorCode.Failed);
                }

                var result2 = GetDataStatistics(_upperPoint, _lowerPoint, _invalidPitPointsNum, _invalidSmoothedPointNum);
                if (result2.ErrorCode != CarotidIntimaMeasureErrorCode.Success)
                {
                    return new CarotidIntimaMeasureResult(result2.ErrorCode);
                }

                //每个像素代表的毫米数-------------------------------------------------------------------------------------
                double ratio = 10* PhysicalPerPixel;
                var lowerMaxThickness = (float)(result2.MaxThick * ratio);
                var lowerMinThickness = (float)(result2.MinThick * ratio);
                var lowerAverageThickness = (float)(result2.AverageThick * ratio);
                var lowerValidPointsNum = result2.ValidPointsNum;
                var lowerSdThickness = result2.SdThickness;
                var std = result2.ThicknessStd;
                return new CarotidIntimaMeasureResult(CarotidIntimaMeasureErrorCode.Success, new List<Point>(_upperPoint),
                   new List<Point>(_lowerPoint),
                    lowerMaxThickness, lowerMinThickness,
                    lowerAverageThickness, lowerSdThickness, lowerValidPointsNum, std);
            }
            catch (Exception e)
            {
                Logger.WriteLineError("CarotidLowerIntimaMeasurement MeasureLowerWall error." + e.Message + "," + e.StackTrace);
                return new CarotidIntimaMeasureResult(CarotidIntimaMeasureErrorCode.Failed);
            }
            finally
            {
                _imageRoiCanny?.Dispose();
                _imageRoiBinary?.Dispose();
            }
        }

        /// <summary>
        /// 去除掉下壁中的异常点
        /// </summary>
        /// <param name="upperPointListTemp"></param>
        /// <param name="lowerPointListTemp"></param>
        /// <param name="num"></param>
        private void RemoveInvalidUpperWallPoints(IList<Point> upperPointListTemp,
            IReadOnlyList<Point> lowerPointListTemp, out int num)
        {
            num = 0;
            var count = lowerPointListTemp.Count;
            for (var i = 0; i < count; ++i)
            {
                if (upperPointListTemp[i].Y <= Math.Max(lowerPointListTemp[i].Y - MinIntimaThickPix, 0)) continue;
                upperPointListTemp[i] = new Point(upperPointListTemp[i].X, Math.Max(lowerPointListTemp[i].Y - MinIntimaThickPix, 0));
                num++;
            }
        }

        /// <summary>
        /// 初步获得上壁边沿点
        /// </summary>
        /// <param name="imageCanny"></param>
        /// <returns></returns>
        private List<Point> GetOriginalUpperPoint(Image<Gray, byte> imageCanny)
        {
            try
            {
                var upperPointListTemp = new List<Point>();
                var dataList = new List<byte>();
                var width = imageCanny.Width;
                var height = imageCanny.Height;

                for (var w = 0; w < width; ++w)
                {
                    dataList.Clear();
                    for (var h = 0; h < height; ++h)
                    {
                        dataList.Add(imageCanny.Data[h, w, 0]);
                    }
                    var y = GetLowerIntimaUpperEdgePoints(dataList);
                    upperPointListTemp.Add(new Point(w, y));
                }
                return upperPointListTemp;
            }
            catch (Exception e)
            {
                Logger.WriteLineError("CarotidLowerIntimaMeasurement GetOriginalUpperPoint error." + e.Message + "," + e.StackTrace);
                return new List<Point>();
            }
        }

        /// <summary>
        /// 颈动脉下内膜检测
        /// </summary>
        /// <param name="_imageRoiBinary"></param>
        /// <param name="_imageRoiCanny"></param>
        /// <param name="rectRoi"></param>
        /// <param name="thick"></param>
        /// <returns></returns>
        private int GetLowerMeasureCarotidPoint(Rectangle originalImageRectRoi, ref Rectangle rectRoi, int thick)
        {
            try
            {
                var upperWall = new List<Point>();
                var lowerWall = new List<Point>();

                var upperPointListTemp = GetOriginalUpperPoint(_imageRoiCanny);
                //排除异常点
                ExcludeExceptionsOnePoint(upperPointListTemp);
                //对上壁作平滑处理
                var smoothResult = PointsSmooth.SmoothLine(upperPointListTemp);
                upperPointListTemp = smoothResult.Points.ToList();

                //缩小测量区域血管的高度
                var minY = upperPointListTemp.Min(o => o.Y);
                if (minY > 10)
                {
                    var rect = new Rectangle(rectRoi.Left, rectRoi.Top + minY - 9, rectRoi.Width, rectRoi.Height - (minY - 9));
                    rectRoi = rect;
                    return 2;
                }
                //上壁点微调
                var res = LowerIntimaUpperEdgePointTrim(_imageRoiBinary, upperPointListTemp);
                if (res != CarotidIntimaMeasureErrorCode.Success)
                {
                    return 1;
                }
                //排除单点异常
                ExcludeExceptionsOnePoint(upperPointListTemp);
                //检测上壁点中的坑
                var removePinResult = RemovePit.RemovePitInLowerIntima(upperPointListTemp);
                upperPointListTemp = removePinResult.Points.ToList();
                _invalidPitPointsNum += removePinResult.InvalidNumber;

                //对上壁作平滑处理
                smoothResult = PointsSmooth.SmoothLine(upperPointListTemp);
                upperPointListTemp = smoothResult.Points.ToList();
                _invalidSmoothedPointNum += smoothResult.InvalidNumber;

                //求下壁
                var result = GetLowerEdgePointsInLowerIntima(_imageRoiBinary, upperPointListTemp, thick);
                if (!result)
                {
                    return 1;
                }
                var lowerPointListTemp = new List<Point>(_lowerPoint);
                //检测下壁点中的坑
                removePinResult = RemovePit.RemovePitInLowerIntima(lowerPointListTemp);
                lowerPointListTemp = removePinResult.Points.ToList();
                _invalidPitPointsNum += removePinResult.InvalidNumber;

                smoothResult = PointsSmooth.SmoothLineLowerWall(lowerPointListTemp, upperPointListTemp);
                lowerPointListTemp = smoothResult.Points.ToList();
                _invalidSmoothedPointNum += smoothResult.InvalidNumber;
                //去除下壁点中的无效点
                RemoveInvalidUpperWallPoints(upperPointListTemp, lowerPointListTemp, out var smoothedNum3);
                _invalidSmoothedPointNum = _invalidSmoothedPointNum + smoothedNum3;
                //上下壁同时平滑处理
                var result1 = PointsSmooth.SmoothLowerIntimaTwoLine(upperPointListTemp, lowerPointListTemp, out var smoothedNum4);
                _invalidSmoothedPointNum = _invalidSmoothedPointNum + smoothedNum4;

                if (result1 != CarotidIntimaMeasureErrorCode.Success)
                {
                    return 1;
                }

                var count = Math.Min(upperPointListTemp.Count, lowerPointListTemp.Count);
                //把小图中的点还原到ROI中正确坐标
                for (var i = 0; i < count; ++i)
                {
                    var upperPointY = Math.Max(upperPointListTemp[i].Y, 0);
                    var lowerPointY = Math.Min(lowerPointListTemp[i].Y, rectRoi.Height - 1);
                    upperWall.Add(new Point(upperPointListTemp[i].X + rectRoi.Left, upperPointY + rectRoi.Top));
                    lowerWall.Add(new Point(lowerPointListTemp[i].X + rectRoi.Left, lowerPointY + rectRoi.Top));
                }
                //把ROI中的点还原到原始图中正确坐标
                _upperPoint.Clear();
                _lowerPoint.Clear();
                for (var i = 0; i < count; ++i)
                {
                    var upperPointY = Math.Max(upperWall[i].Y, 0);
                    var lowerPointY = Math.Min(lowerWall[i].Y, originalImageRectRoi.Height - 1);
                    _upperPoint.Add(new Point(upperPointListTemp[i].X + originalImageRectRoi.Left, upperPointY + originalImageRectRoi.Top));
                    _lowerPoint.Add(new Point(lowerPointListTemp[i].X + originalImageRectRoi.Left, lowerPointY + originalImageRectRoi.Top));
                }
                return 0;
            }
            catch (Exception e)
            {
                Logger.WriteLineError("CarotidLowerIntimaMeasurement GetLowerMeasureCarotidPoint error." + e.Message + "," + e.StackTrace);
                return 1;
            }
        }

        /// <summary>
        /// 求下内膜上边缘点
        /// </summary>
        /// <param name="dataList"></param>
        private static int GetLowerIntimaUpperEdgePoints(IReadOnlyList<byte> dataList)
        {
            try
            {
                var y = 0;
                var count = dataList.Count;
                for (var i = 0; i < count; ++i)
                {
                    if (dataList[i] <= 10) continue;
                    y = i;
                    break;
                }
                return y;
            }
            catch (Exception e)
            {
                Logger.WriteLineError("CarotidLowerIntimaMeasurement GetLowerIntimaUpperEdgePoints error." + e.Message + "," + e.StackTrace);
                return 0;
            }
        }

        /// <summary>
        /// 求下内膜时，求内膜下壁
        /// </summary>
        /// <param name="imageRoiBinary"></param>
        /// <param name="upperWall"></param>
        /// <param name="thick"></param>
        /// <returns></returns>
        private bool GetLowerEdgePointsInLowerIntima(Image<Gray, byte> imageRoiBinary, IReadOnlyList<Point> upperWall, int thick)
        {
            try
            {
                if (upperWall.Count < 3)
                {
                    return false;
                }

                if (thick <= 0)
                {
                    return false;
                }
                if (imageRoiBinary.Height < MinRoiHeight)
                {
                    return false;
                }

                var height = imageRoiBinary.Height - 1;
                thick = (MinIntimaThickPix + thick) / 2;
                var lowerWall = upperWall.Select(point => new Point(point.X, Math.Min(point.Y + thick, height))).ToList();

                //微调下边缘点
                var res = LowerIntimaLowerEdgePointTrim(imageRoiBinary, upperWall, lowerWall);
                if (res != CarotidIntimaMeasureErrorCode.Success)
                {
                    return false;
                }
                _lowerPoint.Clear();
                _lowerPoint.AddRange(lowerWall);
                return true;
            }
            catch (Exception e)
            {
                Logger.WriteLineError("CarotidLowerIntimaMeasurement GetLowerEdgePointsInLowerIntima error." + e.Message + "," + e.StackTrace);
                return false;
            }
        }

        /// <summary>
        /// 下颈动脉，上壁点微调
        /// </summary>
        /// <param name="imageRoiBinary"></param>
        /// <param name="upperWall"></param>
        /// <returns></returns>
        private CarotidIntimaMeasureErrorCode LowerIntimaUpperEdgePointTrim(Image<Gray, byte> imageRoiBinary, List<Point> upperWall)
        {
            try
            {
                var tempList = new List<Point>();
                foreach (var point in upperWall)
                {
                    var trimRange = LowerIntimaUpperEdgeTrimRange;
                    var xValue = point.X;
                    var startYValue = point.Y - trimRange;
                    if (startYValue < 0)
                    {
                        trimRange = point.Y;
                    }

                    if (trimRange == 0)
                    {
                        tempList.Add(point);
                        continue;
                    }
                    if (point.Y + trimRange > imageRoiBinary.Height - 1)
                    {
                        trimRange = imageRoiBinary.Height - 1 - point.Y;
                    }

                    if (trimRange <= 0)
                    {
                        tempList.Add(point);
                        continue;
                    }
                    startYValue = Math.Max(point.Y - trimRange, 0);
                    trimRange = 2 * trimRange + 1;

                    var yRange = new int[trimRange];
                    for (var j = 0; j < trimRange; ++j)
                    {
                        yRange[j] = startYValue + j;
                    }

                    var grayValue = new byte[trimRange];
                    for (var m = 0; m < grayValue.Length; ++m)
                    {
                        grayValue[m] = imageRoiBinary.Data[yRange[m], xValue, 0];
                    }
                    //求变化率
                    var grayDiff = new int[trimRange - 1];
                    for (var n = 0; n < grayDiff.Length; ++n)
                    {
                        grayDiff[n] = grayValue[n + 1] - grayValue[n];
                    }

                    var maxDiff = 0;
                    var maxId = 0;
                    var isFind = false;
                    for (var n = 0; n < grayDiff.Length; ++n)
                    {
                        var temp = grayDiff[n];
                        if (temp <= maxDiff) continue;
                        maxDiff = temp;
                        maxId = n;
                        isFind = true;
                    }
                    tempList.Add(isFind ? new Point(point.X, yRange[maxId]) : point);
                }
                upperWall.Clear();
                upperWall.AddRange(tempList);
                return CarotidIntimaMeasureErrorCode.Success;
            }
            catch (Exception e)
            {
                Logger.WriteLineError("CarotidLowerIntimaMeasurement LowerIntimaUpperEdgePointTrim error." +
                                      e.Message + "," + e.StackTrace);
                return CarotidIntimaMeasureErrorCode.Failed;
            }
        }

        /// <summary>
        /// 下颈动脉，下壁点微调
        /// </summary>
        /// <param name="imageRoiBinary"></param>
        /// <param name="upperWall"></param>
        /// <param name="lowerWall"></param>
        /// <returns></returns>
        private CarotidIntimaMeasureErrorCode LowerIntimaLowerEdgePointTrim(Image<Gray, byte> imageRoiBinary, IReadOnlyList<Point> upperWall, List<Point> lowerWall)
        {
            try
            {
                var trimRange = Math.Max(2, MinIntimaThickPix / 2);
                var tempList = new List<Point>();

                for (var i = 0; i < lowerWall.Count; ++i)
                {
                    var point = lowerWall[i];
                    var xValue = point.X;
                    var startYValue = point.Y - trimRange;
                    if (startYValue < upperWall[i].Y + MinIntimaThickPix)
                    {
                        startYValue = upperWall[i].Y + MinIntimaThickPix + 1;
                    }
                    if (startYValue > point.Y)
                    {
                        tempList.Add(lowerWall[i]);
                        continue;
                    }
                    trimRange = point.Y - startYValue;
                    if (trimRange == 0)
                    {
                        tempList.Add(lowerWall[i]);
                        continue;
                    }

                    if (point.Y + trimRange > Math.Min(imageRoiBinary.Height - 1, upperWall[i].Y + MaxIntimaThickPix))
                    {
                        trimRange = Math.Min(imageRoiBinary.Height - 1, upperWall[i].Y + MaxIntimaThickPix) - point.Y;
                    }
                    if (trimRange <= 0)
                    {
                        tempList.Add(lowerWall[i]);
                        continue;
                    }

                    startYValue = point.Y - trimRange;
                    trimRange = 2 * trimRange + 1;
                    var yRange = new int[trimRange];
                    for (var j = 0; j < trimRange; ++j)
                    {
                        yRange[j] = Math.Min(startYValue + j, imageRoiBinary.Height - 1);
                    }

                    var grayValue = new byte[trimRange];
                    for (var m = 0; m < grayValue.Length; ++m)
                    {
                        grayValue[m] = imageRoiBinary.Data[yRange[m], xValue, 0];
                    }

                    //求变化率
                    var grayDiff = new int[trimRange - 1];
                    for (var n = 0; n < grayDiff.Length; ++n)
                    {
                        grayDiff[n] = grayValue[n + 1] - grayValue[n];
                    }

                    var maxDiff = 0;
                    var maxId = 0;
                    var isFind = false;
                    //找到变化率最大的
                    for (var n = 0; n < grayDiff.Length; ++n)
                    {
                        var temp = grayDiff[n];
                        if (temp <= maxDiff) continue;
                        maxDiff = temp;
                        maxId = n;
                        isFind = true;
                    }

                    for (var m = 0; m < grayValue.Length; ++m)
                    {
                        if (grayValue[m] > ImageThreshold)
                        {
                            grayValue[m] = 255;
                        }
                        grayValue[m] = 0;
                    }

                    var isFind1 = false;
                    var maxId1 = 0;
                    for (var n = 1; n < grayValue.Length; ++n)
                    {
                        if (grayValue[n - 1] != 0 || grayValue[n] != 255) continue;
                        maxId1 = n;
                        isFind1 = true;
                        break;
                    }
                    if (isFind1)
                    {
                        maxId = Math.Min(Math.Min(maxId + 1, maxId1), yRange.Length - 1);
                    }
                    tempList.Add(isFind ? new Point(lowerWall[i].X, yRange[maxId]) : lowerWall[i]);
                }
                lowerWall.Clear();
                lowerWall.AddRange(tempList);
                return CarotidIntimaMeasureErrorCode.Success;
            }
            catch (Exception e)
            {
                Logger.WriteLineError("CarotidLowerIntimaMeasurement LowerIntimaLowerEdgePointTrim error." + e.Message + "," + e.StackTrace);
                return CarotidIntimaMeasureErrorCode.Failed;
            }
        }

        /// <summary>
        ///  图像初始化
        /// </summary>
        /// <param name="imageGray"></param>
        /// <param name="rectRoi"></param>
        /// <param name="subtractTheThreshold"></param>
        /// <returns></returns>
        public ImagePreprocessingResult ImagePreprocessingLower(Image<Gray, byte> imageGray, Rectangle rectRoi, int subtractTheThreshold, int num)
        {
            try
            {
                var imageRoiBinary = imageGray.Copy(rectRoi);
                //求二值化阈值---------------------------           
                var threshold = ImageTools.GetThreshold(imageRoiBinary);
                threshold -= subtractTheThreshold;
                if (threshold < 0)
                {
                    return new ImagePreprocessingResult(false);
                }
                ImageThreshold = threshold;
                var imageRoiCanny = new Image<Gray, byte>(imageRoiBinary.Width, imageRoiBinary.Height);
                //另一二值化的图像
                using (var imgBri = new Image<Gray, byte>(imageRoiBinary.Width, imageRoiBinary.Height))
                {
                    //二值化
                    CvInvoke.Threshold(imageRoiBinary, imgBri, threshold, 255, ThresholdType.Binary);
                    //中值滤波去噪点
                    CvInvoke.MedianBlur(imgBri, imgBri, 3);
                    //清除掉较小的连通域
                    RemoveImpurity(imgBri);
                    //边缘检测
                    CvInvoke.Canny(imgBri, imageRoiCanny, threshold, 255);
                    imageRoiCanny = GetImageCanny(imageRoiCanny, out var maxVectorPointList);
                    if (maxVectorPointList.Count != 0)
                        return new ImagePreprocessingResult(imageRoiBinary, imageRoiCanny, true);
                    ++num;
                    if (num > 10)
                    {
                        return new ImagePreprocessingResult(false);
                    }
                    imageRoiBinary.Dispose();
                    imageRoiCanny?.Dispose();
                    return ImagePreprocessingLower(imageGray, rectRoi, subtractTheThreshold + 10, num);
                }
            }
            catch (Exception e)
            {
                Logger.WriteLineError("CarotidLowerIntimaMeasurement ImagePreprocessingLower error." + e.Message + "," + e.StackTrace);
                return new ImagePreprocessingResult(false);
            }
        }

        /// <summary>
        /// 清除图像中杂质区域
        /// </summary>
        /// <param name="imageCanny"></param>
        /// <param name="maxVectorPointList"></param>
        /// <returns></returns>
        private Image<Gray, byte> GetImageCanny(Image<Gray, byte> imageCanny, out List<VectorOfPoint> maxVectorPointList)
        {
            try
            {
                maxVectorPointList = new List<VectorOfPoint>();
                //求连通域
                var contours = ImageTools.FindContours(imageCanny);
                var maxIdList = new List<int>();
                var maxDiffWidth = imageCanny.Width / 10;
                for (var i = 0; i < contours.Size; i++)
                {
                    var rect = CvInvoke.BoundingRectangle(contours[i]);
                    if (imageCanny.Width - rect.Width >= maxDiffWidth) continue;
                    maxVectorPointList.Add(contours[i]);
                    maxIdList.Add(i);
                }

                if (maxVectorPointList.Count == 1)
                {
                    for (var i = 0; i < contours.Size; i++)
                    {
                        if (i == maxIdList[0]) continue;
                        var temp = contours[i].ToArray();
                        var len = temp.Length;
                        for (var j = 0; j < len; ++j)
                        {
                            imageCanny.Data[temp[j].Y, temp[j].X, 0] = 0;
                        }
                    }
                }

                if (maxVectorPointList.Count <= 1) return imageCanny;
                //从最大的几个轮廓点中早到方差值最小的轨迹
                var stdList = new List<double>();
                var minId = 0;
                double minStd = 1000000;
                for (var i = 0; i < maxVectorPointList.Count; ++i)
                {
                    var std = GetVariance(maxVectorPointList[i]);
                    stdList.Add(std);
                    if (minStd <= std) continue;
                    minStd = std;
                    minId = i;
                }

                for (var i = 0; i < stdList.Count; ++i)
                {
                    if (i == minId) continue;
                    if (Math.Abs(stdList[i] - minStd) >= minStd / 2) continue;
                    //比较两个轮廓线的y值大小
                    var y1 = contours[maxIdList[i]][0].Y;
                    var y2 = contours[maxIdList[minId]][0].Y;
                    if (y1 < y2)
                    {
                        minId = i;
                    }
                }
                //清除掉其他的轮廓线
                for (var i = 0; i < contours.Size; i++)
                {
                    if (i == maxIdList[minId]) continue;
                    var temp = contours[i].ToArray();
                    var len = temp.Length;
                    for (var j = 0; j < len; ++j)
                    {
                        imageCanny.Data[temp[j].Y, temp[j].X, 0] = 0;
                    }
                }

                return imageCanny;
            }
            catch (Exception e)
            {
                Logger.WriteLineError("CarotidLowerIntimaMeasurement GetImageCanny error." + e.Message + "," + e.StackTrace);
                maxVectorPointList = new List<VectorOfPoint>();
                return imageCanny;
            }
        }

        /// <summary>
        /// 去掉异常的一个点
        /// </summary>
        /// <param name="wallPointList"></param>
        private void ExcludeExceptionsOnePoint(List<Point> wallPointList)
        {
            try
            {
                var length = wallPointList.Count;
                if (length < 3)
                {
                    return;
                }

                const int threshold = 3;
                var temp = new List<Point>();
                var firstPointY = wallPointList[0].Y;
                var secondPointY = wallPointList[1].Y;

                if (Math.Abs(firstPointY - secondPointY) > threshold)
                {
                    wallPointList[0] = new Point(wallPointList[0].X, secondPointY);
                    _invalidSmoothedPointNum++;
                }
                temp.Add(new Point(wallPointList[0].X, secondPointY));

                for (var i = 1; i < length - 1; ++i)
                {
                    var pointY = wallPointList[i].Y;
                    var beforePointY = wallPointList[i - 1].Y;
                    var nextPointY = wallPointList[i + 1].Y;
                    if (Math.Abs(pointY - beforePointY) > 2 && Math.Abs(pointY - nextPointY) > threshold)
                    {
                        temp.Add(new Point(wallPointList[i].X, (beforePointY + nextPointY) / 2));
                        _invalidSmoothedPointNum++;
                    }
                    else
                    {
                        temp.Add(new Point(wallPointList[i].X, wallPointList[i].Y));
                    }
                }

                var lastPointY = wallPointList.Last().Y;
                if (Math.Abs(lastPointY - wallPointList[length - 1].Y) > threshold)
                {
                    temp.Add(new Point(wallPointList[length - 1].X, wallPointList[length - 1].Y));
                    _invalidSmoothedPointNum++;
                }
                else
                {
                    temp.Add(wallPointList.Last());
                }

                wallPointList.Clear();
                wallPointList.AddRange(temp);
            }
            catch (Exception e)
            {
                Logger.WriteLineError("CarotidLowerIntimaMeasurement ExcludeExceptionsOnePoint error." + e.Message + "," + e.StackTrace);
            }
        }

    }
}