﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using Emgu.CV;
using Emgu.CV.Structure;
using Emgu.CV.Util;

namespace WingAIDiagnosisService.Carotid.Utilities.DetectIntima
{
    /// <summary>
    /// 自动检测结果
    /// </summary>
    public struct IntimaDetectResult
    {
        public CarotidIntimaMeasureResult UpperIntimaResult;
        public IntimaRect UpperRect;
        public CarotidIntimaMeasureResult LowerIntimaResult;
        public IntimaRect LowerRect;
    }

    /// <summary>
    /// 内膜自动测量
    /// </summary>
    public class IntimaAutoDetect
    {
        private enum IntimaType
        {
            UpperIntima = 1,
            LowerIntima = 2
        }

        private struct OneEdgeDetectResult
        {
            public readonly CarotidIntimaMeasureResult MeasureResult;
            public readonly IntimaRect Rect;

            public OneEdgeDetectResult(IntimaRect rect, CarotidIntimaMeasureResult measureResult)
            {
                Rect = rect;
                MeasureResult = measureResult;
            }
        }
        //图像高
        private int _imgHeight;
        private const int OneRectWidth = 100;

        /// <summary>
        /// 对外接口
        /// </summary>
        /// <param name="measureParameter"></param>
        /// <returns></returns>
        public IntimaDetectResult DetectIntimaApi(CarotidIntimaMeasureParameter measureParameter)
        {
            var image =measureParameter.Image;
            try
            {
                _imgHeight = image.Height;
                //识别血管位置
                var detectVessel = new DetectVessel();
                var res = detectVessel.GetVessel(image);
                //检测上内膜
                var resultUpper = DetectIntima(res.UpperLinePoints.ToArray(), measureParameter, IntimaType.UpperIntima);
                //检测下内膜
                var resultLower = DetectIntima(res.LowerLinePoints.ToArray(), measureParameter, IntimaType.LowerIntima);
                return new IntimaDetectResult
                {
                    UpperRect = resultUpper.Rect,
                    UpperIntimaResult = resultUpper.MeasureResult,
                    LowerRect = resultLower.Rect,
                    LowerIntimaResult = resultLower.MeasureResult
                };
            }
            catch (Exception e)
            {
                return new IntimaDetectResult();
            }
            finally
            {
                image.Dispose();
            }
        }

        /// <summary>
        /// 检测内膜
        /// </summary>
        /// <param name="points"></param>
        /// <param name="measureParameter"></param>
        /// <param name="detectType">1,上膜；2，下膜</param>
        /// <returns></returns>
        private OneEdgeDetectResult DetectIntima(Point[] points, CarotidIntimaMeasureParameter measureParameter, IntimaType detectType)
        {
            //屏蔽上内膜识别，跟服务器端保持一致
            if (detectType == IntimaType.UpperIntima)
            {
                return new OneEdgeDetectResult(new IntimaRect(0, 0, 0, 0),
                    new CarotidIntimaMeasureResult(CarotidIntimaMeasureErrorCode.Failed));
            }
            try
            {
                if (points.Length < 20)
                {
                    return new OneEdgeDetectResult(new IntimaRect(0, 0, 0, 0),
                        new CarotidIntimaMeasureResult(CarotidIntimaMeasureErrorCode.Failed));
                }

                var rectList = GetImageRoiList(points);
                //检测内膜
                List<CarotidIntimaMeasureResult> detectResult;
                if (detectType == IntimaType.UpperIntima)
                {
                    detectResult = DetectUpperIntima(rectList, measureParameter);
                }
                else
                {
                    detectResult = DetectLowerIntima(rectList, measureParameter);
                }

                if (detectResult.Count == 0)
                {
                    return new OneEdgeDetectResult(new IntimaRect(0, 0, 0, 0),
                        new CarotidIntimaMeasureResult(CarotidIntimaMeasureErrorCode.Failed));
                }

                //寻找最优结果
                var minStd = 100f;
                var minNum = -1;
                for (var i = 0; i < detectResult.Count; ++i)
                {
                    if (detectResult[i].ErrorCode != CarotidIntimaMeasureErrorCode.Success) continue;
                    if (minStd <= detectResult[i].ThicknessStd) continue;
                    minStd = detectResult[i].ThicknessStd;
                    minNum = i;
                }
                if (minNum == -1)
                {
                    return new OneEdgeDetectResult(new IntimaRect(0, 0, 0, 0),
                        new CarotidIntimaMeasureResult(CarotidIntimaMeasureErrorCode.Failed));
                }

                return new OneEdgeDetectResult(ChangeRect(rectList[minNum]), detectResult[minNum]);
            }
            catch (Exception e)
            {
                return new OneEdgeDetectResult(new IntimaRect(0, 0, 0, 0),
                    new CarotidIntimaMeasureResult(CarotidIntimaMeasureErrorCode.Failed));
            }
        }

        /// <summary>
        /// 把Rectangle转IntimaRect
        /// </summary>
        /// <param name="rectangle"></param>
        /// <returns></returns>
        private static IntimaRect ChangeRect(Rectangle rectangle)
        {
            return new IntimaRect(rectangle.Left, rectangle.Top, rectangle.Width, rectangle.Height);
        }

        /// <summary>
        /// 把待检测的血管边缘分成多份rect
        /// </summary>
        /// <param name="points"></param>
        /// <returns></returns>
        private List<Rectangle> GetImageRoiList(Point[] points)
        {
            var count = points.Length;
            var listPoints = new List<Point[]>();
            if (count < OneRectWidth)
            {
                listPoints.Add(points);
            }
            else
            {
                var num1 = count / OneRectWidth + 1;
                var num2 = count % OneRectWidth;
                if (num2 < 50)
                {
                    num1 -= 1;
                }

                var begin = 0;
                for (var i = 0; i < num1; ++i)
                {
                    var pointArray = points.Skip(begin).Take(OneRectWidth).ToArray();
                    listPoints.Add(pointArray);
                    begin += OneRectWidth;
                }
            }
            //生成内膜区域
            var rectList = new List<Rectangle>();
            foreach (var point in listPoints)
            {
                var contour = new VectorOfPoint(point);
                var rectTemp = CvInvoke.BoundingRectangle(contour);
                var top = Math.Max(rectTemp.Top - 20, 0);
                var bottom = Math.Min(rectTemp.Top + 20, _imgHeight - 1);
                rectList.Add(new Rectangle(new Point(rectTemp.Left, top), new Size(rectTemp.Width, bottom - top)));
            }

            return rectList;
        }

        /// <summary>
        /// 上内膜检测
        /// </summary>
        /// <param name="rectList"></param>
        /// <param name="measureParameter"></param>
        /// <returns></returns>
        private List<CarotidIntimaMeasureResult> DetectUpperIntima(List<Rectangle> rectList, CarotidIntimaMeasureParameter measureParameter)
        {
            var imageRoiBinaryArray = new Image<Gray, byte>[rectList.Count];
            var imageRoiCannyArray = new Image<Gray, byte>[rectList.Count];
            var detectResult = new List<CarotidIntimaMeasureResult>();
            var physicalPerPixel = measureParameter.PhysicalPerPixel;
            try
            {
                var upperMeasure = new CarotidUpperIntimaMeasurement();
                //优化图像质量
                using (var imageGray = CarotidUpperIntimaMeasurement.ImagePreprocessing(measureParameter))
                {
                    for (var i = 0; i < rectList.Count; ++i)
                    {
                        var result = upperMeasure.ImagePreprocessingUpper(imageGray, rectList[i], 0);
                        if (!result.IsSuccess) continue;
                        imageRoiBinaryArray[i] = result.ImageRoiBinary;
                        imageRoiCannyArray[i] = result.ImageRoiCanny;
                    }
                }

                for (var i = 0; i < imageRoiBinaryArray.Length; ++i)
                {
                    if (imageRoiBinaryArray[i] == null) continue;
                    var measure = new CarotidUpperIntimaMeasurement();
                    var result = measure.MeasureUpperWall(imageRoiBinaryArray[i], imageRoiCannyArray[i], rectList[i],
                        imageRoiBinaryArray[i].ROI, physicalPerPixel);
                    detectResult.Add(result);
                }
                return detectResult;
            }
            catch (Exception e)
            {
                return new List<CarotidIntimaMeasureResult>();
            }
            finally
            {
                foreach (var image in imageRoiBinaryArray)
                {
                    image?.Dispose();
                }
                foreach (var image in imageRoiCannyArray)
                {
                    image?.Dispose();
                }
            }
        }

        /// <summary>
        ///下内膜检测 
        /// </summary>
        /// <param name="rectList"></param>
        /// <param name="measureParameter"></param>
        /// <returns></returns>
        private List<CarotidIntimaMeasureResult> DetectLowerIntima(List<Rectangle> rectList, CarotidIntimaMeasureParameter measureParameter)
        {
            var imageRoiBinaryArray = new Image<Gray, byte>[rectList.Count];
            var imageRoiCannyArray = new Image<Gray, byte>[rectList.Count];
            var detectResult = new List<CarotidIntimaMeasureResult>();
            var physicalPerPixel = measureParameter.PhysicalPerPixel;

            try
            {
                var lowerMeasure = new CarotidLowerIntimaMeasurement();
                //优化图像质量
                using (var imageGray = lowerMeasure.ImagePreprocessing(measureParameter))
                {
                    for (var i = 0; i < rectList.Count; ++i)
                    {
                        var result = lowerMeasure.ImagePreprocessingLower(imageGray, rectList[i], 0, 0);
                        if (!result.IsSuccess) continue;
                        imageRoiBinaryArray[i] = result.ImageRoiBinary;
                        imageRoiCannyArray[i] = result.ImageRoiCanny;
                    }
                }

                var measure = new CarotidLowerIntimaMeasurement();
                for (var i = 0; i < imageRoiBinaryArray.Length; ++i)
                {
                    if (imageRoiBinaryArray[i] == null) continue;
                    var result = measure.MeasureLowerWall(imageRoiBinaryArray[i], imageRoiCannyArray[i], rectList[i],
                        imageRoiBinaryArray[i].ROI, physicalPerPixel);
                    detectResult.Add(result);
                }
                return detectResult;
            }
            catch (Exception e)
            {
                return new List<CarotidIntimaMeasureResult>();
            }
            finally
            {
                foreach (var image in imageRoiBinaryArray)
                {
                    image?.Dispose();
                }
                foreach (var image in imageRoiCannyArray)
                {
                    image?.Dispose();
                }
            }
        }

    }
}