﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using WingServerCommon.Log;

namespace WingAIDiagnosisService.Carotid.Utilities.DetectIntima
{
    public class PointsSmooth
    {
        //下内膜-----------------------------------------------------------------------------------------------------------
        /// <summary>
        /// 点平滑
        /// </summary>
        /// <param name="pointList"></param>
        /// <param name="smoothNum"></param>
        public static SmoothResult SmoothLineLowerWall(IReadOnlyList<Point> pointList, List<Point> upPointsList)
        {
            var result = new SmoothResult(pointList);
            if (result.Points.Count <= 3) return result;
            SmoothLineLowerWall1(result, upPointsList);
            SmoothLineLowerWall2(result, upPointsList);
            SmoothLineLowerWall3(result, upPointsList);
            return result;
        }

        /// <summary>
        /// 单点平滑
        /// </summary>
        /// <param name="pointList"></param>
        /// <param name="smoothNum"></param>
        private static void SmoothLineLowerWall1(FilterPointResult smoothResult, List<Point> upPointsList)
        {
            try
            {
                var pointList = smoothResult.Points;
                var tempPoints = new List<Point>(pointList);
                var count = pointList.Count;
                const int threshold = 2;
                var num = 0;
                //排除掉单点突变
                var temp1 = Math.Max((pointList[1].Y + pointList[2].Y) / 2, upPointsList[0].Y);
                if (Math.Abs(pointList[0].Y - temp1) > 2)
                {
                    tempPoints[0] = new Point(pointList[0].X, temp1);
                    num++;
                }

                for (var i = 1; i < count - 1; ++i)
                {
                    temp1 = Math.Max((pointList[i - 1].Y + pointList[i + 1].Y) / 2, upPointsList[i].Y);
                    if (Math.Abs(pointList[i].Y - temp1) > threshold)
                    {
                        tempPoints[i] = new Point(pointList[i].X, temp1);
                        num++;
                    }
                }
                smoothResult.InvalidNumber += num;
                smoothResult.ResetPoints(tempPoints);
            }
            catch (Exception e)
            {
                Logger.WriteLineError($"PointsSmooth SmoothLineLowerWall1 error {e}");
            }
        }

        /// <summary>
        /// 两个点平滑
        /// </summary>
        /// <param name="pointList"></param>
        /// <param name="smoothNum"></param>
        private static void SmoothLineLowerWall2(FilterPointResult smoothResult, IReadOnlyList<Point> upPointsList)
        {
            try
            {
                var pointList = smoothResult.Points;
                var tempPoints = new List<Point>(pointList);
                var count = pointList.Count;
                const int threshold = 2;
                var num = 0;
                for (var i = 2; i < count - 1; ++i)
                {
                    var temp1 = Math.Max((pointList[i - 2].Y + pointList[i + 1].Y) / 2, upPointsList[i].Y);
                    if (Math.Abs(pointList[i].Y - temp1) > threshold && Math.Abs(pointList[i - 1].Y - temp1) > threshold)
                    {
                        tempPoints[i - 1] = new Point(pointList[i - 1].X, temp1);
                        tempPoints[i] = new Point(pointList[i].X, temp1);
                        i += 2;
                        num += 2;
                    }
                }
                smoothResult.ResetPoints(tempPoints);
                smoothResult.InvalidNumber += num;
            }
            catch (Exception e)
            {
                Logger.WriteLineError($"PointsSmooth SmoothLineLowerWall2 error {e}");
            }
        }

        /// <summary>
        /// 三个点平滑
        /// </summary>
        /// <param name="pointList"></param>
        /// <param name="smoothNum"></param>
        private static void SmoothLineLowerWall3(FilterPointResult smoothResult, IReadOnlyList<Point> upPointsList)
        {
            try
            {
                var pointList = smoothResult.Points;
                var tempPoints = new List<Point>(pointList);
                var count = pointList.Count;
                const int threshold = 2;
                var num = 0;

                for (var i = 3; i < count - 1; ++i)
                {
                    var temp1 = Math.Max((pointList[i - 3].Y + pointList[i + 1].Y) / 2, upPointsList[i].Y);
                    if (Math.Abs(pointList[i].Y - temp1) > threshold && Math.Abs(pointList[i - 1].Y - temp1) > threshold && Math.Abs(pointList[i - 2].Y - temp1) > threshold)
                    {
                        tempPoints[i - 2] = new Point(pointList[i - 2].X, temp1);
                        tempPoints[i - 1] = new Point(pointList[i - 1].X, temp1);
                        tempPoints[i] = new Point(pointList[i].X, temp1);
                        i += 3;
                        num += 3;
                    }
                }
                smoothResult.ResetPoints(tempPoints);
                smoothResult.InvalidNumber += num;
            }
            catch (Exception e)
            {
                Logger.WriteLineError($"PointsSmooth SmoothLineLowerWall3 error {e}");
            }
        }

        /// <summary>
        /// 颈动脉下内膜测量，双壁点的平滑
        /// </summary>
        /// <param name="upperMeasureCarotidPoints"></param>
        /// <param name="lowerMeasureCarotidPoints"></param>
        public static CarotidIntimaMeasureErrorCode SmoothLowerIntimaTwoLine(List<Point> upperMeasureCarotidPoints, List<Point> lowerMeasureCarotidPoints, out int smoothedNum)
        {
            try
            {
                var count = upperMeasureCarotidPoints.Count;
                if (count < 3)
                {
                    smoothedNum = 0;
                    return CarotidIntimaMeasureErrorCode.Failed;
                }

                SmoothLowerIntimaTwoLine6(upperMeasureCarotidPoints, lowerMeasureCarotidPoints, out var smoothedNum1);
                SmoothLowerIntimaTwoLine5(upperMeasureCarotidPoints, lowerMeasureCarotidPoints, out var smoothedNum2);
                SmoothLowerIntimaTwoLine4(upperMeasureCarotidPoints, lowerMeasureCarotidPoints, out var smoothedNum3);
                SmoothLowerIntimaTwoLine3(upperMeasureCarotidPoints, lowerMeasureCarotidPoints, out var smoothedNum4);
                SmoothLowerIntimaTwoLine2(upperMeasureCarotidPoints, lowerMeasureCarotidPoints, out var smoothedNum5);
                SmoothLowerIntimaTwoLine1(upperMeasureCarotidPoints, lowerMeasureCarotidPoints, out var smoothedNum6);

                smoothedNum = smoothedNum1 + smoothedNum2 + smoothedNum3 + smoothedNum4 + smoothedNum5 + smoothedNum6;
                return CarotidIntimaMeasureErrorCode.Success;
            }
            catch (Exception e)
            {
                Logger.WriteLineError("PointsSmooth SmoothLowerIntimaTwoLine error." + e.Message + "," + e.StackTrace);
                smoothedNum = 0;
                return CarotidIntimaMeasureErrorCode.Failed;
            }
        }

        private static void SmoothLowerIntimaTwoLine1(IReadOnlyList<Point> upperMeasureCarotidPoints, List<Point> lowerMeasureCarotidPoints, out int smoothedNum)
        {
            try
            {
                var count = upperMeasureCarotidPoints.Count;
                if (count < 3)
                {
                    smoothedNum = 0;
                    return;
                }

                //排除1个点的突变
                var tempList = new List<Point> { lowerMeasureCarotidPoints[0] };
                var num = 0;
                for (var i = 1; i < count - 1; ++i)
                {
                    var thick1 = lowerMeasureCarotidPoints[i - 1].Y - upperMeasureCarotidPoints[i - 1].Y;
                    var thick2 = lowerMeasureCarotidPoints[i].Y - upperMeasureCarotidPoints[i].Y;
                    var thick3 = lowerMeasureCarotidPoints[i + 1].Y - upperMeasureCarotidPoints[i + 1].Y;
                    if (thick2 > 2 * thick1 && thick2 > 2 * thick3)
                    {
                        tempList.Add(new Point(lowerMeasureCarotidPoints[i].X, (lowerMeasureCarotidPoints[i - 1].Y + lowerMeasureCarotidPoints[i + 1].Y) / 2));
                        num++;
                        continue;
                    }
                    tempList.Add(lowerMeasureCarotidPoints[i]);
                }
                tempList.Add(lowerMeasureCarotidPoints.Last());
                lowerMeasureCarotidPoints.Clear();
                lowerMeasureCarotidPoints.AddRange(tempList);
                smoothedNum = num;
            }
            catch (Exception e)
            {
                Logger.WriteLineError("PointsSmooth SmoothLowerIntimaTwoLine1 error." + e.Message + "," + e.StackTrace);
                smoothedNum = 0;
            }
        }

        private static void SmoothLowerIntimaTwoLine2(IReadOnlyList<Point> upperMeasureCarotidPoints, List<Point> lowerMeasureCarotidPoints, out int smoothedNum)
        {
            try
            {
                var count = upperMeasureCarotidPoints.Count;
                if (count < 5)
                {
                    smoothedNum = 0;
                    return;
                }
                //排除2个点的突变
                var tempList = new List<Point> { lowerMeasureCarotidPoints[0] };
                var num = 0;
                for (var i = 1; i < count - 2; ++i)
                {
                    var thick1 = lowerMeasureCarotidPoints[i - 1].Y - upperMeasureCarotidPoints[i - 1].Y;
                    var thick2 = lowerMeasureCarotidPoints[i].Y - upperMeasureCarotidPoints[i].Y;
                    var thick3 = lowerMeasureCarotidPoints[i + 1].Y - upperMeasureCarotidPoints[i + 1].Y;
                    var thick4 = lowerMeasureCarotidPoints[i + 2].Y - upperMeasureCarotidPoints[i + 2].Y;
                    if (thick2 > 2 * thick1 && thick2 > 2 * thick4 && thick3 > 2 * thick1 && thick3 > 2 * thick4)
                    {
                        var diff = lowerMeasureCarotidPoints[i + 2].Y - lowerMeasureCarotidPoints[i - 1].Y;
                        tempList.Add(new Point(lowerMeasureCarotidPoints[i].X,
                            lowerMeasureCarotidPoints[i - 1].Y + diff / 3));
                        tempList.Add(new Point(lowerMeasureCarotidPoints[i + 1].X,
                            lowerMeasureCarotidPoints[i - 1].Y + diff * 2 / 3));
                        i++;
                        num += 2;
                        continue;
                    }
                    tempList.Add(lowerMeasureCarotidPoints[i]);
                }

                tempList.Add(lowerMeasureCarotidPoints[count - 2]);
                tempList.Add(lowerMeasureCarotidPoints[count - 1]);
                lowerMeasureCarotidPoints.Clear();
                lowerMeasureCarotidPoints.AddRange(tempList);
                smoothedNum = num;
            }
            catch (Exception e)
            {
                Logger.WriteLineError("PointsSmooth SmoothLowerIntimaTwoLine2 error." + e.Message + "," + e.StackTrace);
                smoothedNum = 0;
            }
        }

        private static void SmoothLowerIntimaTwoLine3(IReadOnlyList<Point> upperMeasureCarotidPoints, List<Point> lowerMeasureCarotidPoints, out int smoothedNum)
        {
            try
            {
                var count = upperMeasureCarotidPoints.Count;
                if (count < 6)
                {
                    smoothedNum = 0;
                    return;
                }
                //排除3个点的突变
                var tempList = new List<Point> { lowerMeasureCarotidPoints[0] };
                var num = 0;
                for (var i = 1; i < count - 3; ++i)
                {
                    var thick1 = lowerMeasureCarotidPoints[i - 1].Y - upperMeasureCarotidPoints[i - 1].Y;
                    var thick2 = lowerMeasureCarotidPoints[i].Y - upperMeasureCarotidPoints[i].Y;
                    var thick3 = lowerMeasureCarotidPoints[i + 1].Y - upperMeasureCarotidPoints[i + 1].Y;
                    var thick4 = lowerMeasureCarotidPoints[i + 2].Y - upperMeasureCarotidPoints[i + 2].Y;
                    var thick5 = lowerMeasureCarotidPoints[i + 3].Y - upperMeasureCarotidPoints[i + 3].Y;
                    if (thick2 > 2 * thick1 && thick2 > 2 * thick5)
                    {
                        if (thick3 > 2 * thick1 && thick3 > 2 * thick5)
                        {
                            if (thick4 > 2 * thick1 && thick4 > 2 * thick5)
                            {
                                var diff = lowerMeasureCarotidPoints[i + 3].Y - lowerMeasureCarotidPoints[i - 1].Y;

                                tempList.Add(new Point(lowerMeasureCarotidPoints[i].X, lowerMeasureCarotidPoints[i - 1].Y + diff / 4));
                                tempList.Add(new Point(lowerMeasureCarotidPoints[i + 1].X, lowerMeasureCarotidPoints[i - 1].Y + diff / 2));
                                tempList.Add(new Point(lowerMeasureCarotidPoints[i + 2].X, lowerMeasureCarotidPoints[i - 1].Y + diff * 3 / 4));
                                i += 2;
                                num += 3;
                                continue;
                            }
                        }
                    }
                    tempList.Add(lowerMeasureCarotidPoints[i]);
                }
                tempList.Add(lowerMeasureCarotidPoints[count - 3]);
                tempList.Add(lowerMeasureCarotidPoints[count - 2]);
                tempList.Add(lowerMeasureCarotidPoints[count - 1]);
                lowerMeasureCarotidPoints.Clear();
                lowerMeasureCarotidPoints.AddRange(tempList);
                smoothedNum = num;
            }
            catch (Exception e)
            {
                Logger.WriteLineError("PointsSmooth SmoothLowerIntimaTwoLine3 error." + e.Message + "," + e.StackTrace);
                smoothedNum = 0;
            }
        }

        private static void SmoothLowerIntimaTwoLine4(IReadOnlyList<Point> upperMeasureCarotidPoints, List<Point> lowerMeasureCarotidPoints, out int smoothedNum)
        {
            try
            {
                var count = upperMeasureCarotidPoints.Count;
                if (count < 7)
                {
                    smoothedNum = 0;
                    return;
                }
                //排除4个点的突变
                var tempList = new List<Point> { lowerMeasureCarotidPoints[0] };
                var num = 0;
                for (var i = 1; i < count - 4; ++i)
                {
                    var thick1 = lowerMeasureCarotidPoints[i - 1].Y - upperMeasureCarotidPoints[i - 1].Y;
                    var thick2 = lowerMeasureCarotidPoints[i].Y - upperMeasureCarotidPoints[i].Y;
                    var thick3 = lowerMeasureCarotidPoints[i + 1].Y - upperMeasureCarotidPoints[i + 1].Y;
                    var thick4 = lowerMeasureCarotidPoints[i + 2].Y - upperMeasureCarotidPoints[i + 2].Y;
                    var thick5 = lowerMeasureCarotidPoints[i + 3].Y - upperMeasureCarotidPoints[i + 3].Y;
                    var thick6 = lowerMeasureCarotidPoints[i + 4].Y - upperMeasureCarotidPoints[i + 4].Y;
                    if (thick2 > 2 * thick1 && thick2 > 2 * thick6)
                    {
                        if (thick3 > 2 * thick1 && thick3 > 2 * thick6)
                        {
                            if (thick4 > 2 * thick1 && thick4 > 2 * thick6)
                            {
                                if (thick5 > 2 * thick1 && thick5 > 2 * thick6)
                                {
                                    var diff = lowerMeasureCarotidPoints[i + 4].Y - lowerMeasureCarotidPoints[i - 1].Y;

                                    tempList.Add(new Point(lowerMeasureCarotidPoints[i].X,
                                        lowerMeasureCarotidPoints[i - 1].Y + diff / 5));
                                    tempList.Add(new Point(lowerMeasureCarotidPoints[i + 1].X,
                                        lowerMeasureCarotidPoints[i - 1].Y + diff * 2 / 5));
                                    tempList.Add(new Point(lowerMeasureCarotidPoints[i + 2].X,
                                        lowerMeasureCarotidPoints[i - 1].Y + diff * 3 / 5));
                                    tempList.Add(new Point(lowerMeasureCarotidPoints[i + 3].X,
                                        lowerMeasureCarotidPoints[i - 1].Y + diff * 4 / 5));
                                    i += 3;
                                    num += 4;
                                    continue;
                                }
                            }
                        }
                    }
                    tempList.Add(lowerMeasureCarotidPoints[i]);
                }

                tempList.Add(lowerMeasureCarotidPoints[count - 4]);
                tempList.Add(lowerMeasureCarotidPoints[count - 3]);
                tempList.Add(lowerMeasureCarotidPoints[count - 2]);
                tempList.Add(lowerMeasureCarotidPoints[count - 1]);
                lowerMeasureCarotidPoints.Clear();
                lowerMeasureCarotidPoints.AddRange(tempList);
                smoothedNum = num;
            }
            catch (Exception e)
            {
                Logger.WriteLineError("PointsSmooth SmoothLowerIntimaTwoLine4 error." + e.Message + "," + e.StackTrace);
                smoothedNum = 0;
            }
        }

        private static void SmoothLowerIntimaTwoLine5(IReadOnlyList<Point> upperMeasureCarotidPoints, List<Point> lowerMeasureCarotidPoints, out int smoothedNum)
        {
            try
            {
                var count = upperMeasureCarotidPoints.Count;
                if (count < 8)
                {
                    smoothedNum = 0;
                    return;
                }
                //排除5个点的突变
                var tempList = new List<Point> { lowerMeasureCarotidPoints[0] };
                var num = 0;
                for (var i = 1; i < count - 5; ++i)
                {
                    var thick1 = lowerMeasureCarotidPoints[i - 1].Y - upperMeasureCarotidPoints[i - 1].Y;
                    var thick2 = lowerMeasureCarotidPoints[i].Y - upperMeasureCarotidPoints[i].Y;
                    var thick3 = lowerMeasureCarotidPoints[i + 1].Y - upperMeasureCarotidPoints[i + 1].Y;
                    var thick4 = lowerMeasureCarotidPoints[i + 2].Y - upperMeasureCarotidPoints[i + 2].Y;
                    var thick5 = lowerMeasureCarotidPoints[i + 3].Y - upperMeasureCarotidPoints[i + 3].Y;
                    var thick6 = lowerMeasureCarotidPoints[i + 4].Y - upperMeasureCarotidPoints[i + 4].Y;
                    var thick7 = lowerMeasureCarotidPoints[i + 5].Y - upperMeasureCarotidPoints[i + 5].Y;
                    if (thick2 > 2 * thick1 && thick2 > 2 * thick7)
                    {
                        if (thick3 > 2 * thick1 && thick3 > 2 * thick7)
                        {
                            if (thick4 > 2 * thick1 && thick4 > 2 * thick7)
                            {
                                if (thick5 > 2 * thick1 && thick5 > 2 * thick7)
                                {
                                    if (thick6 > 2 * thick1 && thick6 > 2 * thick7)
                                    {
                                        var diff = lowerMeasureCarotidPoints[i + 5].Y -
                                                   lowerMeasureCarotidPoints[i - 1].Y;

                                        tempList.Add(new Point(lowerMeasureCarotidPoints[i].X,
                                            lowerMeasureCarotidPoints[i - 1].Y + diff / 6));
                                        tempList.Add(new Point(lowerMeasureCarotidPoints[i + 1].X,
                                            lowerMeasureCarotidPoints[i - 1].Y + diff / 3));
                                        tempList.Add(new Point(lowerMeasureCarotidPoints[i + 2].X,
                                            lowerMeasureCarotidPoints[i - 1].Y + diff / 2));
                                        tempList.Add(new Point(lowerMeasureCarotidPoints[i + 3].X,
                                            lowerMeasureCarotidPoints[i - 1].Y + diff * 2 / 3));
                                        tempList.Add(new Point(lowerMeasureCarotidPoints[i + 4].X,
                                            lowerMeasureCarotidPoints[i - 1].Y + diff * 5 / 6));
                                        i += 4;
                                        num += 5;
                                        continue;
                                    }
                                }
                            }
                        }
                    }
                    tempList.Add(lowerMeasureCarotidPoints[i]);
                }

                tempList.Add(lowerMeasureCarotidPoints[count - 5]);
                tempList.Add(lowerMeasureCarotidPoints[count - 4]);
                tempList.Add(lowerMeasureCarotidPoints[count - 3]);
                tempList.Add(lowerMeasureCarotidPoints[count - 2]);
                tempList.Add(lowerMeasureCarotidPoints[count - 1]);
                lowerMeasureCarotidPoints.Clear();
                lowerMeasureCarotidPoints.AddRange(tempList);
                smoothedNum = num;
            }
            catch (Exception e)
            {
                Logger.WriteLineError("PointsSmooth SmoothLowerIntimaTwoLine5 error." + e.Message + "," + e.StackTrace);
                smoothedNum = 0;
            }
        }

        private static void SmoothLowerIntimaTwoLine6(IReadOnlyList<Point> upperMeasureCarotidPoints, List<Point> lowerMeasureCarotidPoints, out int smoothedNum)
        {
            try
            {
                var count = upperMeasureCarotidPoints.Count;
                if (count < 9)
                {
                    smoothedNum = 0;
                    return;
                }

                var num = 0;
                //排除5个点的突变
                var tempList = new List<Point> { lowerMeasureCarotidPoints[0] };
                for (var i = 1; i < count - 6; ++i)
                {
                    var thick1 = lowerMeasureCarotidPoints[i - 1].Y - upperMeasureCarotidPoints[i - 1].Y;
                    var thick2 = lowerMeasureCarotidPoints[i].Y - upperMeasureCarotidPoints[i].Y;
                    var thick3 = lowerMeasureCarotidPoints[i + 1].Y - upperMeasureCarotidPoints[i + 1].Y;
                    var thick4 = lowerMeasureCarotidPoints[i + 2].Y - upperMeasureCarotidPoints[i + 2].Y;
                    var thick5 = lowerMeasureCarotidPoints[i + 3].Y - upperMeasureCarotidPoints[i + 3].Y;
                    var thick6 = lowerMeasureCarotidPoints[i + 4].Y - upperMeasureCarotidPoints[i + 4].Y;
                    var thick7 = lowerMeasureCarotidPoints[i + 5].Y - upperMeasureCarotidPoints[i + 5].Y;
                    var thick8 = lowerMeasureCarotidPoints[i + 6].Y - upperMeasureCarotidPoints[i + 6].Y;

                    if (thick2 > 2 * thick1 && thick2 > 2 * thick8)
                    {
                        if (thick3 > 2 * thick1 && thick3 > 2 * thick8)
                        {
                            if (thick4 > 2 * thick1 && thick4 > 2 * thick8)
                            {
                                if (thick5 > 2 * thick1 && thick5 > 2 * thick8)
                                {
                                    if (thick6 > 2 * thick1 && thick6 > 2 * thick8)
                                    {
                                        if (thick7 > 2 * thick1 && thick7 > 2 * thick8)
                                        {
                                            var diff = lowerMeasureCarotidPoints[i + 6].Y -
                                                       lowerMeasureCarotidPoints[i - 1].Y;

                                            tempList.Add(new Point(lowerMeasureCarotidPoints[i].X,
                                                lowerMeasureCarotidPoints[i - 1].Y + diff / 7));
                                            tempList.Add(new Point(lowerMeasureCarotidPoints[i + 1].X,
                                                lowerMeasureCarotidPoints[i - 1].Y + diff * 2 / 7));
                                            tempList.Add(new Point(lowerMeasureCarotidPoints[i + 2].X,
                                                lowerMeasureCarotidPoints[i - 1].Y + diff * 3 / 7));
                                            tempList.Add(new Point(lowerMeasureCarotidPoints[i + 3].X,
                                                lowerMeasureCarotidPoints[i - 1].Y + diff * 4 / 7));
                                            tempList.Add(new Point(lowerMeasureCarotidPoints[i + 4].X,
                                                lowerMeasureCarotidPoints[i - 1].Y + diff * 5 / 7));
                                            tempList.Add(new Point(lowerMeasureCarotidPoints[i + 5].X,
                                                lowerMeasureCarotidPoints[i - 1].Y + diff * 6 / 7));
                                            i += 5;
                                            num += 6;
                                            continue;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    tempList.Add(lowerMeasureCarotidPoints[i]);
                }

                tempList.Add(lowerMeasureCarotidPoints[count - 6]);
                tempList.Add(lowerMeasureCarotidPoints[count - 5]);
                tempList.Add(lowerMeasureCarotidPoints[count - 4]);
                tempList.Add(lowerMeasureCarotidPoints[count - 3]);
                tempList.Add(lowerMeasureCarotidPoints[count - 2]);
                tempList.Add(lowerMeasureCarotidPoints[count - 1]);
                lowerMeasureCarotidPoints.Clear();
                lowerMeasureCarotidPoints.AddRange(tempList);
                smoothedNum = num;
            }
            catch (Exception e)
            {
                Logger.WriteLineError("PointsSmooth SmoothLowerIntimaTwoLine6 error." + e.Message + "," + e.StackTrace);
                smoothedNum = 0;
            }
        }

        //上内膜----------------------------------------------------
        /// <summary>
        /// 颈动脉下内膜测量，双壁点的平滑
        /// </summary>
        /// <param name="upperMeasureCarotidPoints"></param>
        /// <param name="lowerMeasureCarotidPoints"></param>
        public static CarotidIntimaMeasureErrorCode SmoothUpperIntimaTwoLine(List<Point> upperMeasureCarotidPoints, IReadOnlyList<Point> lowerMeasureCarotidPoints, out int smoothedNum)
        {
            try
            {
                if (upperMeasureCarotidPoints==null || lowerMeasureCarotidPoints==null)
                {
                    smoothedNum = 0;
                    return CarotidIntimaMeasureErrorCode.Failed;
                }
                var count = upperMeasureCarotidPoints.Count;
                if (count < 3)
                {
                    smoothedNum = 0;
                    return CarotidIntimaMeasureErrorCode.Failed;
                }

                SmoothUpperIntimaTwoLine6(upperMeasureCarotidPoints, lowerMeasureCarotidPoints, out var smoothedNum1);
                SmoothUpperIntimaTwoLine5(upperMeasureCarotidPoints, lowerMeasureCarotidPoints, out var smoothedNum2);
                SmoothUpperIntimaTwoLine4(upperMeasureCarotidPoints, lowerMeasureCarotidPoints, out var smoothedNum3);
                SmoothUpperIntimaTwoLine3(upperMeasureCarotidPoints, lowerMeasureCarotidPoints, out var smoothedNum4);
                SmoothUpperIntimaTwoLine2(upperMeasureCarotidPoints, lowerMeasureCarotidPoints, out var smoothedNum5);
                SmoothUpperIntimaTwoLine1(upperMeasureCarotidPoints, lowerMeasureCarotidPoints, out var smoothedNum6);

                smoothedNum = smoothedNum1 + smoothedNum2 + smoothedNum3 + smoothedNum4 + smoothedNum5 + smoothedNum6;
                return CarotidIntimaMeasureErrorCode.Success;
            }
            catch (Exception e)
            {
                Logger.WriteLineError("PointsSmooth SmoothUpperIntimaTwoLine error." + e.Message + "," + e.StackTrace);
                smoothedNum = 0;
                return CarotidIntimaMeasureErrorCode.Failed;
            }
        }

        /// <summary>
        /// 平滑掉一个点的突变
        /// </summary>
        /// <param name="upperMeasureCarotidPoints"></param>
        /// <param name="lowerMeasureCarotidPoints"></param>
        /// <param name="smoothedNum"></param>
        private static void SmoothUpperIntimaTwoLine1(List<Point> upperMeasureCarotidPoints, IReadOnlyList<Point> lowerMeasureCarotidPoints, out int smoothedNum)
        {
            try
            {
                var count = upperMeasureCarotidPoints.Count;
                if (count < 3)
                {
                    smoothedNum = 0;
                    return;
                }

                //排除1个点的突变
                var tempList = new List<Point> { upperMeasureCarotidPoints[0] };
                var num = 0;
                for (var i = 1; i < count - 1; ++i)
                {
                    var thick1 = lowerMeasureCarotidPoints[i - 1].Y - upperMeasureCarotidPoints[i - 1].Y;
                    var thick2 = lowerMeasureCarotidPoints[i].Y - upperMeasureCarotidPoints[i].Y;
                    var thick3 = lowerMeasureCarotidPoints[i + 1].Y - upperMeasureCarotidPoints[i + 1].Y;
                    if (thick2 > 2 * thick1 && thick2 > 2 * thick3)
                    {
                        tempList.Add(new Point(upperMeasureCarotidPoints[i].X, (upperMeasureCarotidPoints[i - 1].Y + upperMeasureCarotidPoints[i + 1].Y) / 2));
                        num++;
                        continue;
                    }
                    tempList.Add(upperMeasureCarotidPoints[i]);
                }
                tempList.Add(upperMeasureCarotidPoints.Last());
                upperMeasureCarotidPoints.Clear();
                upperMeasureCarotidPoints.AddRange(tempList);

                smoothedNum = num;
            }
            catch (Exception e)
            {
                smoothedNum = 0;
                Logger.WriteLineError($"PointsSmooth SmoothUpperIntimaTwoLine1 error {e}");
            }
        }

        /// <summary>
        /// 平滑掉两个点的突变
        /// </summary>
        /// <param name="upperMeasureCarotidPoints"></param>
        /// <param name="lowerMeasureCarotidPoints"></param>
        /// <param name="smoothedNum"></param>
        private static void SmoothUpperIntimaTwoLine2(List<Point> upperMeasureCarotidPoints, IReadOnlyList<Point> lowerMeasureCarotidPoints, out int smoothedNum)
        {
            try
            {
                var count = upperMeasureCarotidPoints.Count;
                if (count < 5)
                {
                    smoothedNum = 0;
                    return;
                }
                //排除2个点的突变
                var tempList = new List<Point> { upperMeasureCarotidPoints[0] };
                var num = 0;
                for (var i = 1; i < count - 2; ++i)
                {
                    var thick1 = lowerMeasureCarotidPoints[i - 1].Y - upperMeasureCarotidPoints[i - 1].Y;
                    var thick2 = lowerMeasureCarotidPoints[i].Y - upperMeasureCarotidPoints[i].Y;
                    var thick3 = lowerMeasureCarotidPoints[i + 1].Y - upperMeasureCarotidPoints[i + 1].Y;
                    var thick4 = lowerMeasureCarotidPoints[i + 2].Y - upperMeasureCarotidPoints[i + 2].Y;
                    if (thick2 > 2 * thick1 && thick2 > 2 * thick4 && thick3 > 2 * thick1 && thick3 > 2 * thick4)
                    {
                        var diff = upperMeasureCarotidPoints[i + 2].Y - upperMeasureCarotidPoints[i - 1].Y;
                        tempList.Add(new Point(upperMeasureCarotidPoints[i].X,
                            upperMeasureCarotidPoints[i - 1].Y + diff / 3));
                        tempList.Add(new Point(upperMeasureCarotidPoints[i + 1].X,
                            upperMeasureCarotidPoints[i - 1].Y + diff * 2 / 3));
                        i++;
                        num += 2;
                        continue;
                    }
                    tempList.Add(upperMeasureCarotidPoints[i]);
                }

                tempList.Add(upperMeasureCarotidPoints[count - 2]);
                tempList.Add(upperMeasureCarotidPoints[count - 1]);
                upperMeasureCarotidPoints.Clear();
                upperMeasureCarotidPoints.AddRange(tempList);
                smoothedNum = num;
            }
            catch (Exception e)
            {
                smoothedNum = 0;
                Logger.WriteLineError($"PointsSmooth SmoothUpperIntimaTwoLine2 error {e}");
            }
        }

        /// <summary>
        /// 平滑掉三个点的突变
        /// </summary>
        /// <param name="upperMeasureCarotidPoints"></param>
        /// <param name="lowerMeasureCarotidPoints"></param>
        /// <param name="smoothedNum"></param>
        private static void SmoothUpperIntimaTwoLine3(List<Point> upperMeasureCarotidPoints, IReadOnlyList<Point> lowerMeasureCarotidPoints, out int smoothedNum)
        {
            try
            {
                var count = upperMeasureCarotidPoints.Count;
                if (count < 6)
                {
                    smoothedNum = 0;
                    return;
                }
                //排除3个点的突变
                var tempList = new List<Point> { upperMeasureCarotidPoints[0] };
                var num = 0;
                for (var i = 1; i < count - 3; ++i)
                {
                    var thick1 = lowerMeasureCarotidPoints[i - 1].Y - upperMeasureCarotidPoints[i - 1].Y;
                    var thick2 = lowerMeasureCarotidPoints[i].Y - upperMeasureCarotidPoints[i].Y;
                    var thick3 = lowerMeasureCarotidPoints[i + 1].Y - upperMeasureCarotidPoints[i + 1].Y;
                    var thick4 = lowerMeasureCarotidPoints[i + 2].Y - upperMeasureCarotidPoints[i + 2].Y;
                    var thick5 = lowerMeasureCarotidPoints[i + 3].Y - upperMeasureCarotidPoints[i + 3].Y;
                    if (thick2 > 2 * thick1 && thick2 > 2 * thick5)
                    {
                        if (thick3 > 2 * thick1 && thick3 > 2 * thick5)
                        {
                            if (thick4 > 2 * thick1 && thick4 > 2 * thick5)
                            {
                                var diff = upperMeasureCarotidPoints[i + 3].Y - upperMeasureCarotidPoints[i - 1].Y;

                                tempList.Add(new Point(upperMeasureCarotidPoints[i].X, upperMeasureCarotidPoints[i - 1].Y + diff / 4));
                                tempList.Add(new Point(upperMeasureCarotidPoints[i + 1].X, upperMeasureCarotidPoints[i - 1].Y + diff / 2));
                                tempList.Add(new Point(upperMeasureCarotidPoints[i + 2].X, upperMeasureCarotidPoints[i - 1].Y + diff * 3 / 4));
                                i += 2;
                                num += 3;
                                continue;
                            }
                        }
                    }
                    tempList.Add(upperMeasureCarotidPoints[i]);
                }
                tempList.Add(upperMeasureCarotidPoints[count - 3]);
                tempList.Add(upperMeasureCarotidPoints[count - 2]);
                tempList.Add(upperMeasureCarotidPoints[count - 1]);
                upperMeasureCarotidPoints.Clear();
                upperMeasureCarotidPoints.AddRange(tempList);
                smoothedNum = num;
            }
            catch (Exception e)
            {
                smoothedNum = 0;
                Logger.WriteLineError($"PointsSmooth SmoothUpperIntimaTwoLine3 error {e}");
            }
        }

        /// <summary>
        /// 平滑掉四个点的突变
        /// </summary>
        /// <param name="upperMeasureCarotidPoints"></param>
        /// <param name="lowerMeasureCarotidPoints"></param>
        /// <param name="smoothedNum"></param>
        private static void SmoothUpperIntimaTwoLine4(List<Point> upperMeasureCarotidPoints, IReadOnlyList<Point> lowerMeasureCarotidPoints, out int smoothedNum)
        {
            try
            {
                var count = upperMeasureCarotidPoints.Count;
                if (count < 7)
                {
                    smoothedNum = 0;
                    return;
                }
                //排除4个点的突变
                var tempList = new List<Point> { upperMeasureCarotidPoints[0] };
                var num = 0;
                for (var i = 1; i < count - 4; ++i)
                {
                    var thick1 = lowerMeasureCarotidPoints[i - 1].Y - upperMeasureCarotidPoints[i - 1].Y;
                    var thick2 = lowerMeasureCarotidPoints[i].Y - upperMeasureCarotidPoints[i].Y;
                    var thick3 = lowerMeasureCarotidPoints[i + 1].Y - upperMeasureCarotidPoints[i + 1].Y;
                    var thick4 = lowerMeasureCarotidPoints[i + 2].Y - upperMeasureCarotidPoints[i + 2].Y;
                    var thick5 = lowerMeasureCarotidPoints[i + 3].Y - upperMeasureCarotidPoints[i + 3].Y;
                    var thick6 = lowerMeasureCarotidPoints[i + 4].Y - upperMeasureCarotidPoints[i + 4].Y;
                    if (thick2 > 2 * thick1 && thick2 > 2 * thick6)
                    {
                        if (thick3 > 2 * thick1 && thick3 > 2 * thick6)
                        {
                            if (thick4 > 2 * thick1 && thick4 > 2 * thick6)
                            {
                                if (thick5 > 2 * thick1 && thick5 > 2 * thick6)
                                {
                                    var diff = upperMeasureCarotidPoints[i + 4].Y - upperMeasureCarotidPoints[i - 1].Y;

                                    tempList.Add(new Point(upperMeasureCarotidPoints[i].X,
                                        upperMeasureCarotidPoints[i - 1].Y + diff / 5));
                                    tempList.Add(new Point(upperMeasureCarotidPoints[i + 1].X,
                                        upperMeasureCarotidPoints[i - 1].Y + diff * 2 / 5));
                                    tempList.Add(new Point(upperMeasureCarotidPoints[i + 2].X,
                                        upperMeasureCarotidPoints[i - 1].Y + diff * 3 / 5));
                                    tempList.Add(new Point(upperMeasureCarotidPoints[i + 3].X,
                                        upperMeasureCarotidPoints[i - 1].Y + diff * 4 / 5));
                                    i += 3;
                                    num += 4;
                                    continue;
                                }
                            }
                        }
                    }
                    tempList.Add(upperMeasureCarotidPoints[i]);
                }

                tempList.Add(upperMeasureCarotidPoints[count - 4]);
                tempList.Add(upperMeasureCarotidPoints[count - 3]);
                tempList.Add(upperMeasureCarotidPoints[count - 2]);
                tempList.Add(upperMeasureCarotidPoints[count - 1]);
                upperMeasureCarotidPoints.Clear();
                upperMeasureCarotidPoints.AddRange(tempList);
                smoothedNum = num;
            }
            catch (Exception e)
            {
                smoothedNum = 0;
                Logger.WriteLineError($"PointsSmooth SmoothUpperIntimaTwoLine4 error {e}");
            }
        }

        /// <summary>
        /// 平滑掉五个点的突变
        /// </summary>
        /// <param name="upperMeasureCarotidPoints"></param>
        /// <param name="lowerMeasureCarotidPoints"></param>
        /// <param name="smoothedNum"></param>
        private static void SmoothUpperIntimaTwoLine5(List<Point> upperMeasureCarotidPoints, IReadOnlyList<Point> lowerMeasureCarotidPoints, out int smoothedNum)
        {
            try
            {
                var count = upperMeasureCarotidPoints.Count;
                if (count < 8)
                {
                    smoothedNum = 0;
                    return;
                }
                //排除5个点的突变
                var tempList = new List<Point> { upperMeasureCarotidPoints[0] };
                var num = 0;
                for (var i = 1; i < count - 5; ++i)
                {
                    var thick1 = lowerMeasureCarotidPoints[i - 1].Y - upperMeasureCarotidPoints[i - 1].Y;
                    var thick2 = lowerMeasureCarotidPoints[i].Y - upperMeasureCarotidPoints[i].Y;
                    var thick3 = lowerMeasureCarotidPoints[i + 1].Y - upperMeasureCarotidPoints[i + 1].Y;
                    var thick4 = lowerMeasureCarotidPoints[i + 2].Y - upperMeasureCarotidPoints[i + 2].Y;
                    var thick5 = lowerMeasureCarotidPoints[i + 3].Y - upperMeasureCarotidPoints[i + 3].Y;
                    var thick6 = lowerMeasureCarotidPoints[i + 4].Y - upperMeasureCarotidPoints[i + 4].Y;
                    var thick7 = lowerMeasureCarotidPoints[i + 5].Y - upperMeasureCarotidPoints[i + 5].Y;
                    if (thick2 > 2 * thick1 && thick2 > 2 * thick7)
                    {
                        if (thick3 > 2 * thick1 && thick3 > 2 * thick7)
                        {
                            if (thick4 > 2 * thick1 && thick4 > 2 * thick7)
                            {
                                if (thick5 > 2 * thick1 && thick5 > 2 * thick7)
                                {
                                    if (thick6 > 2 * thick1 && thick6 > 2 * thick7)
                                    {
                                        var diff = upperMeasureCarotidPoints[i + 5].Y -
                                                   upperMeasureCarotidPoints[i - 1].Y;

                                        tempList.Add(new Point(upperMeasureCarotidPoints[i].X,
                                            upperMeasureCarotidPoints[i - 1].Y + diff / 6));
                                        tempList.Add(new Point(upperMeasureCarotidPoints[i + 1].X,
                                            upperMeasureCarotidPoints[i - 1].Y + diff / 3));
                                        tempList.Add(new Point(upperMeasureCarotidPoints[i + 2].X,
                                            upperMeasureCarotidPoints[i - 1].Y + diff / 2));
                                        tempList.Add(new Point(upperMeasureCarotidPoints[i + 3].X,
                                            upperMeasureCarotidPoints[i - 1].Y + diff * 2 / 3));
                                        tempList.Add(new Point(upperMeasureCarotidPoints[i + 4].X,
                                            upperMeasureCarotidPoints[i - 1].Y + diff * 5 / 6));
                                        i += 4;
                                        num += 5;
                                        continue;
                                    }
                                }
                            }
                        }
                    }
                    tempList.Add(upperMeasureCarotidPoints[i]);
                }

                tempList.Add(upperMeasureCarotidPoints[count - 5]);
                tempList.Add(upperMeasureCarotidPoints[count - 4]);
                tempList.Add(upperMeasureCarotidPoints[count - 3]);
                tempList.Add(upperMeasureCarotidPoints[count - 2]);
                tempList.Add(upperMeasureCarotidPoints[count - 1]);
                upperMeasureCarotidPoints.Clear();
                upperMeasureCarotidPoints.AddRange(tempList);
                smoothedNum = num;
            }
            catch (Exception e)
            {
                smoothedNum = 0;
                Logger.WriteLineError($"PointsSmooth SmoothUpperIntimaTwoLine5 error {e}");
            }
        }

        /// <summary>
        /// 平滑掉六个点的突变
        /// </summary>
        /// <param name="upperMeasureCarotidPoints"></param>
        /// <param name="lowerMeasureCarotidPoints"></param>
        /// <param name="smoothedNum"></param>
        private static void SmoothUpperIntimaTwoLine6(List<Point> upperMeasureCarotidPoints, IReadOnlyList<Point> lowerMeasureCarotidPoints, out int smoothedNum)
        {
            try
            {
                var count = upperMeasureCarotidPoints.Count;
                if (count < 9)
                {
                    smoothedNum = 0;
                    return;
                }
                var num = 0;
                //排除5个点的突变
                var tempList = new List<Point> { upperMeasureCarotidPoints[0] };
                for (var i = 1; i < count - 6; ++i)
                {
                    var thick1 = lowerMeasureCarotidPoints[i - 1].Y - upperMeasureCarotidPoints[i - 1].Y;
                    var thick2 = lowerMeasureCarotidPoints[i].Y - upperMeasureCarotidPoints[i].Y;
                    var thick3 = lowerMeasureCarotidPoints[i + 1].Y - upperMeasureCarotidPoints[i + 1].Y;
                    var thick4 = lowerMeasureCarotidPoints[i + 2].Y - upperMeasureCarotidPoints[i + 2].Y;
                    var thick5 = lowerMeasureCarotidPoints[i + 3].Y - upperMeasureCarotidPoints[i + 3].Y;
                    var thick6 = lowerMeasureCarotidPoints[i + 4].Y - upperMeasureCarotidPoints[i + 4].Y;
                    var thick7 = lowerMeasureCarotidPoints[i + 5].Y - upperMeasureCarotidPoints[i + 5].Y;
                    var thick8 = lowerMeasureCarotidPoints[i + 6].Y - upperMeasureCarotidPoints[i + 6].Y;

                    if (thick2 > 2 * thick1 && thick2 > 2 * thick8)
                    {
                        if (thick3 > 2 * thick1 && thick3 > 2 * thick8)
                        {
                            if (thick4 > 2 * thick1 && thick4 > 2 * thick8)
                            {
                                if (thick5 > 2 * thick1 && thick5 > 2 * thick8)
                                {
                                    if (thick6 > 2 * thick1 && thick6 > 2 * thick8)
                                    {
                                        if (thick7 > 2 * thick1 && thick7 > 2 * thick8)
                                        {
                                            var diff = upperMeasureCarotidPoints[i + 6].Y -
                                                       upperMeasureCarotidPoints[i - 1].Y;

                                            tempList.Add(new Point(upperMeasureCarotidPoints[i].X,
                                                upperMeasureCarotidPoints[i - 1].Y + diff / 7));
                                            tempList.Add(new Point(upperMeasureCarotidPoints[i + 1].X,
                                                lowerMeasureCarotidPoints[i - 1].Y + diff * 2 / 7));
                                            tempList.Add(new Point(upperMeasureCarotidPoints[i + 2].X,
                                                upperMeasureCarotidPoints[i - 1].Y + diff * 3 / 7));
                                            tempList.Add(new Point(upperMeasureCarotidPoints[i + 3].X,
                                                upperMeasureCarotidPoints[i - 1].Y + diff * 4 / 7));
                                            tempList.Add(new Point(upperMeasureCarotidPoints[i + 4].X,
                                                upperMeasureCarotidPoints[i - 1].Y + diff * 5 / 7));
                                            tempList.Add(new Point(upperMeasureCarotidPoints[i + 5].X,
                                                upperMeasureCarotidPoints[i - 1].Y + diff * 6 / 7));
                                            i += 5;
                                            num += 6;
                                            continue;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    tempList.Add(upperMeasureCarotidPoints[i]);
                }

                tempList.Add(upperMeasureCarotidPoints[count - 6]);
                tempList.Add(upperMeasureCarotidPoints[count - 5]);
                tempList.Add(upperMeasureCarotidPoints[count - 4]);
                tempList.Add(upperMeasureCarotidPoints[count - 3]);
                tempList.Add(upperMeasureCarotidPoints[count - 2]);
                tempList.Add(upperMeasureCarotidPoints[count - 1]);
                upperMeasureCarotidPoints.Clear();
                upperMeasureCarotidPoints.AddRange(tempList);
                smoothedNum = num;
            }
            catch (Exception e)
            {
                smoothedNum = 0;
                Logger.WriteLineError($"PointsSmooth SmoothUpperIntimaTwoLine6 error {e}");
            }
        }

        /// <summary>
        /// 点平滑
        /// </summary>
        /// <param name="pointList"></param>
        /// <param name="smoothNum"></param>
        public static SmoothResult SmoothLineUpperWall(IReadOnlyList<Point> pointList, IReadOnlyList<Point> lowerPointListTemp)
        {
            var result = new SmoothResult(pointList);
            if (result.Points.Count <= 3) return result;
            SmoothLineUpperWall1(result, lowerPointListTemp);
            SmoothLineUpperWall2(result, lowerPointListTemp);
            SmoothLineUpperWall3(result, lowerPointListTemp);
            return result;
        }

        /// <summary>
        /// 单点平滑
        /// </summary>
        /// <param name="pointList"></param>
        /// <param name="smoothNum"></param>
        private static void SmoothLineUpperWall1(FilterPointResult smoothResult, IReadOnlyList<Point> lowerPointListTemp)
        {
            try
            {
                var pointList = smoothResult.Points;
                var tempPoints = new List<Point>(pointList);
                var count = pointList.Count;
                const int threshold = 2;
                var num = 0;
                //排除掉单点突变
                var temp1 = Math.Min((pointList[1].Y + pointList[2].Y) / 2, lowerPointListTemp[0].Y);
                if (Math.Abs(pointList[0].Y - temp1) > 2)
                {
                    tempPoints[0] = new Point(pointList[0].X, temp1);
                    num++;
                }

                for (var i = 1; i < count - 1; ++i)
                {
                    temp1 = Math.Min((pointList[i - 1].Y + pointList[i + 1].Y) / 2, lowerPointListTemp[i].Y);
                    if (Math.Abs(pointList[i].Y - temp1) <= threshold) continue;
                    tempPoints[i] = new Point(pointList[i].X, temp1);
                    num++;
                }
                smoothResult.InvalidNumber += num;
                smoothResult.ResetPoints(tempPoints);
            }
            catch (Exception e)
            {
                Logger.WriteLineError($"PointsSmooth SmoothLineUpperWall1 error {e}");
            }
        }

        /// <summary>
        /// 两个点平滑
        /// </summary>
        /// <param name="pointList"></param>
        /// <param name="smoothNum"></param>
        private static void SmoothLineUpperWall2(FilterPointResult smoothResult, IReadOnlyList<Point> lowerPointListTemp)
        {
            try
            {
                var pointList = smoothResult.Points;
                var tempPoints = new List<Point>(pointList);
                var count = pointList.Count;
                const int threshold = 2;
                var num = 0;
                for (var i = 2; i < count - 1; ++i)
                {
                    var temp1 = Math.Min((pointList[i - 2].Y + pointList[i + 1].Y) / 2, lowerPointListTemp[i].Y);
                    if (Math.Abs(pointList[i].Y - temp1) <= threshold ||
                        Math.Abs(pointList[i - 1].Y - temp1) <= threshold) continue;
                    tempPoints[i - 1] = new Point(pointList[i - 1].X, temp1);
                    tempPoints[i] = new Point(pointList[i].X, temp1);
                    i += 2;
                    num += 2;
                }
                smoothResult.ResetPoints(tempPoints);
                smoothResult.InvalidNumber += num;
            }
            catch (Exception e)
            {
                Logger.WriteLineError($"PointsSmooth SmoothLineUpperWall2 error {e}");
            }
        }

        /// <summary>
        /// 三个点平滑
        /// </summary>
        /// <param name="pointList"></param>
        /// <param name="smoothNum"></param>
        private static void SmoothLineUpperWall3(FilterPointResult smoothResult, IReadOnlyList<Point> lowerPointListTemp)
        {
            try
            {
                var pointList = smoothResult.Points;
                var tempPoints = new List<Point>(pointList);
                var count = pointList.Count;

                const int threshold = 2;
                var num = 0;
                for (var i = 3; i < count - 1; ++i)
                {
                    var temp1 = Math.Min((pointList[i - 3].Y + pointList[i + 1].Y) / 2, lowerPointListTemp[i].Y);
                    if (Math.Abs(pointList[i].Y - temp1) <= threshold ||
                        Math.Abs(pointList[i - 1].Y - temp1) <= threshold ||
                        Math.Abs(pointList[i - 2].Y - temp1) <= threshold) continue;
                    tempPoints[i - 2] = new Point(pointList[i - 2].X, temp1);
                    tempPoints[i - 1] = new Point(pointList[i - 1].X, temp1);
                    tempPoints[i] = new Point(pointList[i].X, temp1);
                    i += 3;
                    num += 3;
                }
                smoothResult.ResetPoints(tempPoints);
                smoothResult.InvalidNumber += num;
            }
            catch (Exception e)
            {
                Logger.WriteLineError($"PointsSmooth SmoothLineUpperWall3 error {e}");
            }
        }

        //上下内膜通用-------------------------------------------------------------
        /// <summary>
        /// 点平滑
        /// </summary>
        /// <param name="pointList"></param>
        /// <param name="smoothNum"></param>
        public static SmoothResult SmoothLine(IReadOnlyList<Point> pointList)
        {
            var result = new SmoothResult(pointList);
            if (result.Points.Count <= 3) return result;
            SmoothLine1(result);
            SmoothLine2(result);
            SmoothLine3(result);
            return result;
        }

        /// <summary>
        /// 单点平滑
        /// </summary>
        /// <param name="pointList"></param>
        /// <param name="smoothNum"></param>
        private static void SmoothLine1(FilterPointResult smoothResult)
        {
            try
            {
                var pointList = smoothResult.Points;
                var tempPoints = new List<Point>(pointList);
                var count = pointList.Count;
                const int threshold = 2;
                var num = 0;
                //排除掉单点突变
                var temp1 = (pointList[1].Y + pointList[2].Y) / 2;
                if (Math.Abs(pointList[0].Y - temp1) > 2)
                {
                    tempPoints[0] = new Point(pointList[0].X, temp1);
                    num++;
                }

                for (var i = 1; i < count - 1; ++i)
                {
                    temp1 = (pointList[i - 1].Y + pointList[i + 1].Y) / 2;
                    if (Math.Abs(pointList[i].Y - temp1) > threshold)
                    {
                        tempPoints[i] = new Point(pointList[i].X, temp1);
                        num++;
                    }
                }

                smoothResult.InvalidNumber += num;
                smoothResult.ResetPoints(tempPoints);
            }
            catch (Exception e)
            {
                Logger.WriteLineError($"PointsSmooth SmoothLine1 error {e}");
            }
        }

        /// <summary>
        /// 两个点平滑
        /// </summary>
        /// <param name="pointList"></param>
        /// <param name="smoothNum"></param>
        private static void SmoothLine2(FilterPointResult smoothResult)
        {
            try
            {
                var pointList = smoothResult.Points;
                var tempPoints = new List<Point>(pointList);
                var count = pointList.Count;
                const int threshold = 2;
                var num = 0;
                for (var i = 2; i < count - 1; ++i)
                {
                    var temp1 = (pointList[i - 2].Y + pointList[i + 1].Y) / 2;
                    if (Math.Abs(pointList[i].Y - temp1) > threshold && Math.Abs(pointList[i - 1].Y - temp1) > threshold)
                    {
                        tempPoints[i - 1] = new Point(pointList[i - 1].X, temp1);
                        tempPoints[i] = new Point(pointList[i].X, temp1);
                        i += 2;
                        num += 2;
                    }
                }
                smoothResult.ResetPoints(tempPoints);
                smoothResult.InvalidNumber += num;
            }
            catch (Exception e)
            {
                Logger.WriteLineError($"PointsSmooth SmoothLine2 error {e}");
            }
        }

        /// <summary>
        /// 三个点平滑
        /// </summary>
        /// <param name="pointList"></param>
        /// <param name="smoothNum"></param>
        private static void SmoothLine3(FilterPointResult smoothResult)
        {
            try
            {
                var pointList = smoothResult.Points;
                var tempPoints = new List<Point>(pointList);
                var count = pointList.Count;
                const int threshold = 2;
                var num = 0;

                for (var i = 3; i < count - 1; ++i)
                {
                    var temp1 = (pointList[i - 3].Y + pointList[i + 1].Y) / 2;
                    if (Math.Abs(pointList[i].Y - temp1) > threshold && Math.Abs(pointList[i - 1].Y - temp1) > threshold && Math.Abs(pointList[i - 2].Y - temp1) > threshold)
                    {
                        tempPoints[i - 2] = new Point(pointList[i - 2].X, temp1);
                        tempPoints[i - 1] = new Point(pointList[i - 1].X, temp1);
                        tempPoints[i] = new Point(pointList[i].X, temp1);
                        i += 3;
                        num += 3;
                    }
                }
                smoothResult.ResetPoints(tempPoints);
                smoothResult.InvalidNumber += num;
            }
            catch (Exception e)
            {
                Logger.WriteLineError($"PointsSmooth SmoothLine3 error {e}");
            }
        }
    }
}
