﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using WingServerCommon.Log;

namespace WingAIDiagnosisService.Carotid.Utilities.DetectIntima
{
    public class RemovePit
    {
        /// <summary>
        /// 下内膜，去除向上的凹坑
        /// </summary>
        /// <param name="pointsList"></param>
        /// <returns></returns>
        public static RemovePitResult RemovePitInLowerIntima(IReadOnlyList<Point> pointsList)
        {
            try
            {
                var allPitPoint = GetPitLinePoint(pointsList);
                var pitMaxPoint = GetMaxPointInPit(allPitPoint);
                var newPit = GetNewPitsBaseMaxPointInPits(pitMaxPoint, pointsList);
                //计算每个坑的属性
                var pitList = GetPitAttributes(newPit);
                //补平凹坑
                return FilletPit(pointsList, pitList);
            }
            catch (Exception e)
            {
                Logger.WriteLineError("RemovePit RemovePitInLowerIntima error." + e.Message + "," + e.StackTrace);
                return new RemovePitResult(pointsList);
            }
        }

        /// <summary>
        /// 获得在同一侧的曲线
        /// </summary>
        /// <param name="upperWall"></param>
        /// <returns></returns>
        private static IEnumerable<Point[]> GetPitLinePoint(IReadOnlyCollection<Point> points)
        {
            try
            {
                var pointList = new List<Point[]>();
                var len = points.Count;
                for (var j = 0; j < len - 5; ++j)
                {
                    for (var i = j + 5; i < len; ++i)
                    {
                        var beginIndex = j;
                        var endIndex = i;
                        var pointTemp = points.Skip(beginIndex).Take(endIndex - beginIndex).ToArray();
                        if (!IsOnTheSameSide(pointTemp)) continue;
                        //判断是朝上的坑还是朝下的坑    
                        if (IsUpperPit(pointTemp))
                        {
                            pointList.Add(points.Skip(beginIndex).Take(endIndex - beginIndex - 1).ToArray());
                        }
                    }
                }
                return pointList;
            }
            catch (Exception e)
            {
                Logger.WriteLineError("RemovePit GetPitLinePoint error." + e.Message + "," + e.StackTrace);
                return new List<Point[]>();
            }
        }

        /// <summary>
        /// 获得每个坑的坑底，Y值最大的点
        /// </summary>
        /// <param name="allPitEdgePoints"></param>
        /// <returns></returns>
        private static IEnumerable<Point> GetMaxPointInPit(IEnumerable<Point[]> allPitEdgePoints)
        {
            try
            {
                var pitMaxPointList = new List<Point>();
                foreach (var onePitEdgePoints in allPitEdgePoints)
                {
                    var maxPointY = 0;
                    var maxPointX = 0;
                    foreach (var point in onePitEdgePoints)
                    {
                        if (point.Y <= maxPointY) continue;
                        maxPointY = point.Y;
                        maxPointX = point.X;
                    }
                    pitMaxPointList.Add(new Point(maxPointX, maxPointY));
                }
                return pitMaxPointList.Distinct();
            }
            catch (Exception e)
            {
                Logger.WriteLineError("RemovePit GetMaxPointInPit error." + e.Message + "," + e.StackTrace);
                return new List<Point>();
            }
        }

        /// <summary>
        /// 根据坑低,Y值最大的点计算新的坑
        /// </summary>
        /// <param name="pitMaxPoint"></param>
        /// <param name="pointsArray"></param>
        /// <returns></returns>
        private static IEnumerable<Point[]> GetNewPitsBaseMaxPointInPits(IEnumerable<Point> pitMaxPoint, IReadOnlyList<Point> pointsArray)
        {
            try
            {
                const int yDiffRange = 3;
                //根据找到的最低点，分别向两边找坑的边
                var newPit = new List<Point[]>();
                var maxX = pointsArray.Last().X;
                foreach (var point in pitMaxPoint)
                {
                    //找左边的index
                    var beginIndex = Math.Max(point.X - 1, 0);
                    var leftMaxPoint = new Point(beginIndex, int.MaxValue);
                    while (beginIndex >= 0)
                    {
                        if (Math.Abs(point.Y - pointsArray[beginIndex].Y) < yDiffRange)
                        {
                            if (pointsArray[beginIndex].Y < leftMaxPoint.Y)
                            {
                                leftMaxPoint = pointsArray[beginIndex];
                            }
                            beginIndex--;
                            continue;
                        }

                        var pointTemp = pointsArray.Skip(beginIndex).Take(point.X - beginIndex).ToArray();
                        if (!IsOnTheSameSide(pointTemp))
                        {
                            break;
                        }

                        if (pointsArray[beginIndex].Y < leftMaxPoint.Y)
                        {
                            leftMaxPoint = pointsArray[beginIndex];
                        }
                        beginIndex--;
                    }

                    //找右边的index                
                    var endIndex = Math.Min(point.X + 1, maxX);
                    var rightMaxPoint = new Point(endIndex, int.MaxValue);
                    while (endIndex <= maxX)
                    {
                        if (Math.Abs(point.Y - pointsArray[endIndex].Y) < yDiffRange)
                        {
                            if (pointsArray[endIndex].Y < rightMaxPoint.Y)
                            {
                                rightMaxPoint = pointsArray[endIndex];
                            }
                            endIndex++;
                            continue;
                        }
                        var pointTemp = pointsArray.Skip(point.X).Take(endIndex - point.X).ToArray();
                        if (!IsOnTheSameSide(pointTemp))
                        {
                            break;
                        }

                        if (pointsArray[endIndex].Y < rightMaxPoint.Y)
                        {
                            rightMaxPoint = pointsArray[endIndex];
                        }
                        endIndex++;
                    }

                    //找到两边的最高点
                    if (leftMaxPoint.X != point.X && rightMaxPoint.X != point.X)
                    {
                        newPit.Add(pointsArray.Skip(leftMaxPoint.X).Take(rightMaxPoint.X - leftMaxPoint.X + 1).ToArray());
                    }
                }
                //去除掉重复的坑
                return RemoveSamePit(newPit);
            }
            catch (Exception e)
            {
                Logger.WriteLineError("RemovePit GetNewPitsBaseMaxPointInPits error." + e.Message + "," + e.StackTrace);
                return new List<Point[]>();
            }
        }

        /// <summary>
        /// 计算坑的属性
        /// </summary>
        /// <param name="pitEdgePoints"></param>
        /// <returns></returns>
        private static IEnumerable<Pit> GetPitAttributes(IEnumerable<Point[]> pitEdgePoints)
        {
            try
            {
                //计算每个坑的属性
                return from edgePoints in pitEdgePoints
                       let pitMaxY = edgePoints.Max(o => o.Y)
                       select new Pit
                       {
                           PitWidth = edgePoints.Last().X - edgePoints[0].X,
                           PitHeightBegin = pitMaxY - edgePoints[0].Y,
                           PitHeightEnd = pitMaxY - edgePoints.Last().Y,
                           IsUpward = IsUpperPit(edgePoints),
                           PitEdgePoints = edgePoints
                       };
            }
            catch (Exception e)
            {
                Logger.WriteLineError("RemovePit GetPitAttributes error." + e.Message + "," + e.StackTrace);
                return new List<Pit>();
            }
        }

        /// <summary>
        /// 补平凹坑
        /// </summary>
        /// <param name="pointsList"></param>
        /// <param name="pitList"></param>
        /// <param name="pointsArray"></param>
        /// <returns></returns>
        private static RemovePitResult FilletPit(IReadOnlyList<Point> pointsList, IEnumerable<Pit> pitList)
        {
            try
            {
                //高度阈值
                const int heightThreshold = 2;
                //筛选坑
                var pitList1 = pitList.Where(pit => pit.PitHeightBegin > heightThreshold && pit.PitHeightEnd > heightThreshold && pit.IsUpward).ToList();
                var changePointsNum = 0;
                var pointsArray = pointsList.ToArray();
                for (var i = 0; i < pitList1.Count; ++i)
                {
                    var points = pitList1[i].PitEdgePoints;
                    var newPoints = ImageTools.GetPointsInTwoPoint(points[0], points.Last());
                    RemoveLinePoint(newPoints);
                    var endX = points.Last().X;
                    var num = 0;

                    //将新的点替换掉坑中的点
                    for (var j = points[0].X + 1; j < endX; ++j)
                    {
                        pointsArray[j] = new Point(j, newPoints[num].Y);
                        num++;
                        changePointsNum++;
                    }
                }

                var removePinResult = new RemovePitResult(pointsList) { InvalidNumber = changePointsNum };
                removePinResult.ResetPoints(pointsArray);
                return removePinResult;
            }
            catch (Exception e)
            {
                Logger.WriteLineError("RemovePit FilletPit error." + e.Message + "," + e.StackTrace);
                return new RemovePitResult(pointsList);
            }
        }

        /// <summary>
        /// 判断点是否在同一侧
        /// </summary>
        /// <param name="points"></param>
        /// <returns></returns>
        private static bool IsOnTheSameSide(IReadOnlyList<Point> points)
        {
            try
            {
                var length = points.Count;
                if (length < 3)
                {
                    return false;
                }
                //求直线方程
                var point1 = points[0];
                var point2 = points[length - 1];

                var isSmaller = 0;
                if (point1.X == point2.X)
                {
                    for (var i = 1; i < length - 2; ++i)
                    {
                        var temp = 0;
                        if (points[i].X < point1.X)
                        {
                            temp = 1;
                        }
                        if (points[i].X > point1.X)
                        {
                            temp = 2;
                        }

                        if (isSmaller != 0)
                        {
                            if (temp != 0 && temp != isSmaller)
                            {
                                return false;
                            }
                        }
                        else
                        {
                            if (temp != 0)
                            {
                                isSmaller = temp;
                            }
                        }
                    }
                    return true;
                }
                if (point1.Y == point2.Y)
                {
                    for (var i = 1; i < length - 2; ++i)
                    {
                        var temp = 0;
                        if (points[i].Y < point1.Y)
                        {
                            temp = 1;
                        }
                        if (points[i].Y > point1.Y)
                        {
                            temp = 2;
                        }

                        if (isSmaller != 0)
                        {
                            if (temp != 0 && temp != isSmaller)
                            {
                                return false;
                            }
                        }
                        else
                        {
                            if (temp != 0)
                            {
                                isSmaller = temp;
                            }
                        }
                    }
                    return true;
                }
                var k = (point1.Y - point2.Y) / (float)(point1.X - point2.X);
                var b = point1.Y - k * point1.X;
                for (var i = 1; i < length - 2; ++i)
                {
                    var temp = 0;
                    if (k * points[i].X - points[i].Y + b > 0)
                    {
                        temp = 1;
                    }
                    if (k * points[i].X - points[i].Y + b < 0)
                    {
                        temp = 2;
                    }

                    if (isSmaller != 0)
                    {
                        if (temp != 0 && temp != isSmaller)
                        {
                            return false;
                        }
                    }
                    else
                    {
                        if (temp != 0)
                        {
                            isSmaller = temp;
                        }
                    }
                }
                return true;
            }
            catch (Exception e)
            {
                Logger.WriteLineError("RemovePit IsOnTheSameSide error." + e.Message + "," + e.StackTrace);
                return false;
            }
        }

        /// <summary>
        /// 判断坑的朝向是不是朝上的
        /// </summary>
        /// <param name="points"></param>
        /// <returns></returns>
        private static bool IsUpperPit(IReadOnlyList<Point> points)
        {
            var len = points.Count;
            var y1 = points[0].Y;
            var y2 = points[len - 1].Y;
            var yCenter = points[len / 2].Y;
            return yCenter > y1 && yCenter > y2;
        }

        /// <summary>
        /// 去除掉相同的坑
        /// </summary>
        /// <param name="pitList"></param>
        /// <returns></returns>
        private static IEnumerable<Point[]> RemoveSamePit(IEnumerable<Point[]> pitList)
        {
            try
            {
                var newPitList = new List<Point[]>();
                foreach (var pit in pitList)
                {
                    if (pit.Length < 5)
                    {
                        continue;
                    }

                    if (newPitList.Count > 0)
                    {
                        var xMin = pit[0].X;
                        var xMax = pit.Last().X - 1;
                        var isFindSame = false;
                        foreach (var point in newPitList)
                        {
                            var x2 = point[0].X;
                            var x3 = point.Last().X;
                            if (x3 <= xMax && x3 > xMin)
                            {
                                isFindSame = true;
                                break;
                            }

                            if (x2 > xMax || x2 < xMin) continue;
                            isFindSame = true;
                            break;
                        }
                        if (!isFindSame)
                        {
                            newPitList.Add(pit);
                        }
                    }
                    else
                    {
                        newPitList.Add(pit);
                    }
                }
                return newPitList;
            }
            catch (Exception e)
            {
                Logger.WriteLineError("RemovePit RemoveSamePit error." + e.Message + "," + e.StackTrace);
                return new List<Point[]>();
            }
        }

        /// <summary>
        /// 去掉直线点
        /// </summary>
        /// <param name="pointList"></param>
        /// <returns></returns>
        private static void RemoveLinePoint(IList<Point> pointList)
        {
            try
            {
                //判断补的是不是直线
                const int maxLineNum = 5;
                var len = pointList.Count;
                var lastY = pointList[0].Y;

                var beginIndex = 0;
                var endIndex = 0;
                for (var i = 1; i < len; ++i)
                {
                    if (pointList[i].Y == lastY)
                    {
                        endIndex++;
                        if (endIndex >= len - 1)
                        {
                            ChangeLinePoints(pointList, beginIndex, endIndex);
                        }
                    }
                    else
                    {
                        if (endIndex - beginIndex > maxLineNum)
                        {
                            ChangeLinePoints(pointList, beginIndex, endIndex);
                        }

                        beginIndex = i;
                        endIndex = i;
                        lastY = pointList[i].Y;
                    }
                }
            }
            catch (Exception e)
            {
                Logger.WriteLineError("RemovePit RemoveLinePoint error." + e.Message + "," + e.StackTrace);
            }
        }

        /// <summary>
        /// 将直线点改成不直的线
        /// </summary>
        /// <param name="pointList"></param>
        /// <param name="beginIndex"></param>
        /// <param name="endIndex"></param>
        private static void ChangeLinePoints(IList<Point> pointList, int beginIndex, int endIndex)
        {
            try
            {
                var addY = 0;
                var addNum = 0;
                var symbol = 1;

                var interval = (endIndex - beginIndex) / 3;
                if (interval > 6)
                {
                    interval = 6;
                }
                for (var j = beginIndex + 1; j < endIndex; ++j)
                {
                    if (addNum == interval)
                    {
                        addNum = 0;
                        switch (addY)
                        {
                            case 1:
                                symbol = -symbol;
                                addY += symbol;
                                break;
                            case -1:
                                symbol = -symbol;
                                addY += symbol;
                                break;
                            default:
                                addY += symbol;
                                break;
                        }
                    }

                    pointList[j] = new Point(pointList[j].X, pointList[j].Y + addY);
                    addNum++;
                }
            }
            catch (Exception e)
            {
                Logger.WriteLineError("RemovePit ChangeLinePoints error." + e.Message + "," + e.StackTrace);
            }
        }


    }
}
