﻿using System;
using System.Collections.Generic;
using Emgu.CV;
using Emgu.CV.Structure;
using System.Drawing;
using WingServerCommon.Log;
using WingAIDiagnosisService.Carotid.CarotidInterFacesData;
using WingAIDiagnosisService.Carotid.MathTools;
using WingAIDiagnosisService.Carotid.Utilities.DetectIntima;

namespace WingAIDiagnosisService.Carotid.Utilities
{
    public class DetectIntimaResult
    {
        /// <summary>
        /// Ant Intima Value
        /// </summary>
        public OneDetectIntimaResult AntIntima { get; set; }

        /// <summary>
        /// Post Intima Value
        /// </summary>
        public OneDetectIntimaResult PostIntima { get; set; }

        /// <summary>
        /// 内膜是否检测成功
        /// </summary>
        public bool IsSuccess { get; set; }

        public DetectIntimaResult()
        {
            AntIntima = new OneDetectIntimaResult(false);
            PostIntima = new OneDetectIntimaResult(false);
            IsSuccess = false;
        }

        public DetectIntimaResult(bool isSuccess)
        {
            IsSuccess = isSuccess;
        }

        public DetectIntimaResult(bool isSuccess, OneDetectIntimaResult antIntima, OneDetectIntimaResult postIntima)
        {
            IsSuccess = isSuccess;
            AntIntima = antIntima;
            PostIntima = postIntima;
        }
    }

    public class OneDetectIntimaResult
    {

        //内膜厚度
        public float IntimaThick { get; set; }
        //是否识别成功
        public bool IsSuccess { get; set; }

        public OneDetectIntimaResult() { }

        public OneDetectIntimaResult(bool isSuccess)
        {
            IsSuccess = isSuccess;
        }
        public OneDetectIntimaResult(bool isSuccess, float intimaThick)
        {
            IsSuccess = isSuccess;
            IntimaThick = intimaThick;
        }
    }

    //获得纵切面图像
    public class GetIntimaImage : IDisposable
    {
        private List<Vector3D> _aiPlanePoint;

        private PlaneEquation _planeEquation;

        private ModelSize _modelSize;

        public PlaneEquation GetPlaneEquation()
        {
            return _planeEquation;
        }

        public List<Vector3D> GetIntimaPlanePoint()
        {
            return _aiPlanePoint;
        }

        private static GetIntimaImage _instance;
        public static GetIntimaImage Instance
        {
            get => _instance ?? (_instance = new GetIntimaImage());
        }

        /// <summary>
        /// 获取AI切面图像
        /// </summary>
        /// <param name="_modelDataSrc"></param>
        /// <param name="_srcLengthX"></param>
        /// <param name="_srcLengthY"></param>
        /// <param name="_srcLengthZ"></param>
        /// <param name="intimaImagePath"></param>
        /// <returns></returns>
        public DetectIntimaResult GetImage(string intimaImagePath, float physicalPerPixel)
        {
            try
            {
                _modelSize = SliceHelper.Instance.GetModeSzie();
                //获得切面图
                var image = GetOriginalImage();
                if (image != null)
                {
                    //识别内膜
                    using var imageIntima = image.Copy();
                    var detectResult = GetIntimaAutoDetectResult(imageIntima, physicalPerPixel);
                    if (detectResult.LowerIntimaResult.ErrorCode == CarotidIntimaMeasureErrorCode.Success)
                    {
                        if (AIClipImageStorer.SaveImage(intimaImagePath, image, detectResult.LowerIntimaResult, detectResult.LowerRect))
                        {
                            float lowerThick = CarotidIntimaMeasureResult.UnifyPrecision(detectResult.LowerIntimaResult.AverageThickness);
                            var lowerIntimaResult = new OneDetectIntimaResult(true, lowerThick);
                            //上内膜不检测，设为false
                            return new DetectIntimaResult(true, new OneDetectIntimaResult(false, 0f), lowerIntimaResult);
                        }
                    }
                }
            
                return new DetectIntimaResult(false);
            }
            catch
            {
                return new DetectIntimaResult(false);
            }
        }

        /// <summary>
        /// 获得切面图
        /// </summary>
        /// <returns></returns>
        private Image<Gray, byte> GetOriginalImage()
        {
            //计算切面图像,平面上的三个点
            _aiPlanePoint = GetVesselSlice();
            if (_aiPlanePoint.Count < 3)
            {
                Logger.WriteLineWarn($"GetAIImage, get ai image failed!");
                return null;
            }
            _planeEquation = MathTools3D.GetPlaneEquation(_aiPlanePoint[0], _aiPlanePoint[1], _aiPlanePoint[2]);

            var image = SliceHelper.Instance.Get2DMeasureImage(_aiPlanePoint);
            return image;
        }

        /// <summary>
        /// 计算血管所在位置切面
        /// </summary>
        /// <returns></returns>
        public List<Vector3D> GetVesselSlice()
        {
            try
            {
                var pointListCenter = new List<Vector3I>();
                var arteryList = GetModelVesselAndPlaque.Instance.GetIntimaVessel();
                var count = arteryList.Count;

                for (var i = 0; i < count; i++)
                {
                    if (arteryList[i].ArteryContours.Count < 2)
                    {
                        var rect = arteryList[i].ArteryContours[0].Rect;
                        pointListCenter.Add(new Vector3I(rect.Left + rect.Width / 2, rect.Top + rect.Height / 2, arteryList[i].SerialNumber));
                    }
                }

                if (pointListCenter.Count < 3)
                {
                    return new List<Vector3D>();
                }
                //计算每个rect中点在y=0平面上的投影
                var projectionPoint = new List<Point>();
                foreach (var point in pointListCenter)
                {
                    projectionPoint.Add(new Point(point.X, point.Z));
                }

                //计算所有投影点的拟合直线
                var line = MathTools2D.CalculateFittedLine(projectionPoint);

                //计算三个过拟合直线且和y轴平行的平面上的点
                return GetLastThreePoints(line);
            }
            catch (Exception e)
            {
                Logger.WriteLineError($"GetIntimaImage, GetVesselSlice error" + e.Message + "," + e.StackTrace);
                return new List<Vector3D>();
            }

        }

        /// <summary>
        /// 计算三个过拟合直线且和y轴平行的平面上的点
        /// </summary>
        /// <param name="line"></param>
        /// <returns></returns>
        private List<Vector3D> GetLastThreePoints(IReadOnlyList<double> lineOriginal)
        {
            try
            {
                var points = new List<Vector3D>();
                if (lineOriginal[1] == 0)
                {
                    var x = -lineOriginal[2];
                    x = Math.Max(0, x);
                    x = Math.Min(x, _modelSize.ModelLengthX - 1);
                    points.Add(new Vector3D(x, 0, 0));
                    points.Add(new Vector3D(x, 0, _modelSize.ModelLengthZ - 1));
                    points.Add(new Vector3D(x, _modelSize.ModelLengthY - 1, 0));
                }
                else
                {
                    double[] line = new double[2] { lineOriginal[0], lineOriginal[2] };
                    //计算直线与z= _modelSize.ModelLengthZ-1,y=0直线的交点
                    var tempX1 = (_modelSize.ModelLengthZ - 1 - line[1]) / line[0];
                    if (tempX1 < 0)
                    {
                        var tempZ = (int)Math.Round(line[1]);
                        tempZ = Math.Max(0, tempZ);
                        tempZ = Math.Min(tempZ, _modelSize.ModelLengthZ - 1);
                        points.Add(new Vector3D(0, 0, tempZ));
                        points.Add(new Vector3D(0, _modelSize.ModelLengthY - 1, tempZ));
                    }
                    else
                    {
                        if (tempX1 < _modelSize.ModelLengthX - 1)
                        {
                            var tempX = (int)Math.Round(tempX1);
                            tempX = Math.Max(tempX, 0);
                            tempX = Math.Min(tempX, _modelSize.ModelLengthX - 1);
                            points.Add(new Vector3D(tempX, 0, _modelSize.ModelLengthZ - 1));
                            points.Add(new Vector3D(tempX, _modelSize.ModelLengthY - 1, _modelSize.ModelLengthZ - 1));
                        }
                        else
                        {
                            var tempZ = (int)Math.Round(line[0] * (_modelSize.ModelLengthX - 1) + line[1]);
                            tempZ = Math.Max(tempZ, 0);
                            tempZ = Math.Min(tempZ, _modelSize.ModelLengthZ - 1);
                            points.Add(new Vector3D(_modelSize.ModelLengthX - 1, 0, tempZ));
                            points.Add(new Vector3D(_modelSize.ModelLengthX - 1, _modelSize.ModelLengthY - 1, tempZ));
                        }
                    }

                    var tempX2 = (-line[1]) / line[0];
                    if (tempX2 < 0)
                    {
                        var tempZ = (int)Math.Round(line[1]);
                        tempZ = Math.Max(tempZ, 0);
                        tempZ = Math.Min(tempZ, _modelSize.ModelLengthZ - 1);
                        points.Add(new Vector3D(0, 0, tempZ));
                    }
                    else
                    {
                        if (tempX2 < _modelSize.ModelLengthX - 1)
                        {
                            var tempX = (int)Math.Round((-line[1]) / line[0]);
                            tempX = Math.Max(tempX, 0);
                            tempX = Math.Min(tempX, _modelSize.ModelLengthX - 1);
                            points.Add(new Vector3D(tempX, 0, 0));
                        }
                        else
                        {
                            var tempZ = (int)Math.Round((_modelSize.ModelLengthX - 1) * line[0] + line[1]);
                            tempZ = Math.Max(tempZ, 0);
                            tempZ = Math.Min(tempZ, _modelSize.ModelLengthZ - 1);
                            points.Add(new Vector3D(_modelSize.ModelLengthX - 1, 0, tempZ));
                        }
                    }
                }
                return points;
            }
            catch (Exception e)
            {
                Logger.WriteLineError($"GetIntimaImage, GetLastThreePoints error {e}");
                return new List<Vector3D>();
            }
        }

        /// <summary>
        /// 获得自动内膜结果
        /// </summary>
        /// <param name="image"></param>
        /// <returns></returns>
        private IntimaDetectResult GetIntimaAutoDetectResult(Image<Gray, byte> image, float physicalPerPixel)
        {
            try
            {
                //识别内膜
                var intimaAutoDetect = new IntimaAutoDetect();
                var imageProcessState = new ImageProcessChangeState(false, false, false);
                var carotidIntimaMeasureParameter = new CarotidIntimaMeasureParameter(image,
               new IntimaRect(0, 0, 0, 0), imageProcessState, physicalPerPixel);
                var intimaAutoDetectResult = intimaAutoDetect.DetectIntimaApi(carotidIntimaMeasureParameter);
                return intimaAutoDetectResult;
            }
            catch (Exception e)
            {
                return new IntimaDetectResult();
            }
        }

    
        /// <summary>
        /// 释放资源
        /// </summary>
        public void Dispose()
        {
            if (_aiPlanePoint != null)
            {
                _aiPlanePoint.Clear();
                _aiPlanePoint = null;
            }

        }
    }
}