﻿using System.Collections.Generic;
using Emgu.CV;
using System;
using System.Drawing;
using WingAIDiagnosisService.Carotid.MathTools;
using WingAIDiagnosisService.Carotid.CarotidInterFacesData;
using WingServerCommon.Log;
using WingAIDiagnosisService.Carotid.Utilities.CntkSeg;

namespace WingAIDiagnosisService.Carotid.Utilities
{
    /// <summary>
    /// 获得Y形血管切面
    /// </summary>
    public class GetPlaneOfVesselY
    {

        private ModelSize _modelSize;

        private PlaneEquation _planeY;

        private bool _isGetedPlaneY;

        private static GetPlaneOfVesselY _instance;
        public static GetPlaneOfVesselY Instance
        {
            get => _instance ?? (_instance = new GetPlaneOfVesselY());
        }

        public PlaneEquation GetPlaneYEquation()
        {
            if (!_isGetedPlaneY)
            {
                return new PlaneEquation();
            }
            return _planeY;
        }

        public bool GetPlane(byte[][] modelSource, ModelSize modelSize, string imageYPath)
        {
            try
            {
                _isGetedPlaneY = false;
                _modelSize = modelSize;
                //计算每个平面的动脉轮廓
                var vesselContoursCenterBegin = GetVesselContoursBegin(modelSource, out bool isFindYBegin);
                //计算每个平面的动脉轮廓
                var vesselContoursCenterEnd = GetVesselContoursEnd(modelSource);

                //计算Y图像
                var pointInPlane = GetYPlanePoint(vesselContoursCenterBegin, vesselContoursCenterEnd, isFindYBegin);

                _planeY = MathTools3D.GetPlaneEquation(pointInPlane[0], pointInPlane[1], pointInPlane[2]);
                _isGetedPlaneY = true;
                var image = SliceHelper.Instance.GetImageFromThreePoint(pointInPlane, true);
                if (image != null)
                {
                    //顺时针旋转90
                    CvInvoke.Rotate(image, image, Emgu.CV.CvEnum.RotateFlags.Rotate90Clockwise);
                    ImageTools.Adjust(image.Mat);
                    image.Save(imageYPath);
                    return true;
                }
                return false;
            }
            catch (Exception e)
            {
                Logger.WriteLineError($"GetPlaneOfVesselY GetPlane have an error{e}");
                _isGetedPlaneY = false;
                return false;
            }
        }

        /// <summary>
        /// 计算血管位置
        /// </summary>
        /// <param name="image"></param>
        /// <returns></returns>
        private List<Vector3D> GetVesselContoursBegin(byte[][] modelSource, out bool isFindYBegin)
        {
            isFindYBegin = false;
            try
            {
                int findNum = 0;
                var listPoint = new List<Vector3D>();
                var imageNum = modelSource.Length;
                //计算血管位置
                var vesselContours = new List<List<CntkDetectResult>>();
                var arteryList = GetModelVesselAndPlaque.Instance.GetYBeginVessel();

                //计算开头一段-----------------------------------------------
                var count = arteryList.Count;
                for (var i = 0; i < count; i++)
                {
                    var serialNumber = arteryList[i].SerialNumber;
                    List<CntkDetectResult> arteryDetectResult = arteryList[i].ArteryContours;
                    if (arteryDetectResult.Count == 0) continue;
                    if (arteryDetectResult.Count == 1)
                    {
                        vesselContours.Add(arteryDetectResult);
                        var countJ = arteryDetectResult.Count;
                        for (var j = 0; j < countJ; ++j)
                        {
                            if (arteryDetectResult[j].LabelType == CntkLabelType.LabelTwo) continue;
                            var featurePoints = GetContoursFeaturePoints(arteryDetectResult[j].Contour, serialNumber);
                            if (featurePoints.Count > 2)
                            {
                                findNum++;
                                listPoint.AddRange(featurePoints);
                            }
                        }
                    }
                    if (arteryDetectResult.Count > 1)
                    {
                        for (var j = 0; j < arteryDetectResult.Count; ++j)
                        {
                            var contour = arteryDetectResult[j].Contour;
                            double sumX = 0;
                            double sumY = 0;

                            //计算中心点
                            for (var i1 = 0; i1 < contour.Length; ++i1)
                            {
                                sumX += contour[i1].X;
                                sumY += contour[i1].Y;
                            }
                            findNum++;
                            listPoint.Add(new Vector3D(sumX / contour.Length, sumY / contour.Length, serialNumber));
                        }
                    }
                }

                if (findNum > 3)
                {
                    isFindYBegin = true;
                }
                return listPoint;
            }
            catch (Exception e)
            {
                Logger.WriteLineError($"GetPlaneOfVesselY GetVesselContoursBegin have an error{e}");
                return new List<Vector3D>();
            }

        }

        /// <summary>
        /// 计算血管位置
        /// </summary>
        /// <param name="image"></param>
        /// <returns></returns>
        private List<Vector3D> GetVesselContoursEnd(byte[][] modelSource)
        {
            try
            {
                var centerPointList = new List<Vector3D>();
                var imageNum = modelSource.Length;
                //计算血管位置
                var vesselContours = new List<List<CntkDetectResult>>();
                var arteryList = GetModelVesselAndPlaque.Instance.GetYEndVessel();

                //计算一段-----------------------------------------------
                var count = arteryList.Count;
                for (var i = 0; i < count; i++)
                {
                    List<CntkDetectResult> arteryDetectResult = arteryList[i].ArteryContours;
                    if (arteryDetectResult.Count == 0) continue;
                    centerPointList.Add(Get2DContoursCenterPoints(arteryDetectResult[0].Contour, arteryList[i].SerialNumber));
                }
                return centerPointList;
            }
            catch (Exception e)
            {
                Logger.WriteLineError($"GetPlaneOfVesselY GetVesselContoursEnd have an error{e}");
                return new List<Vector3D>();
            }
        }

        private List<Vector3D> GetContoursFeaturePoints(Point[] contourPoints, int z)
        {
            var pointList = new List<Vector3D>();
            //计算轮廓的中心点
            var len = contourPoints.Length;
            var sumX = 0;
            var sumY = 0;
            for (var i = 0; i < len; ++i)
            {
                sumX += contourPoints[i].X;
                sumY += contourPoints[i].Y;
            }
            var centerX = sumX / (float)len;
            var centerY = sumY / (float)len;

            //计算所有点距离中心点的最大值和最小值
            double minDistance = 100000000;
            double maxDistance = 0;
            Point maxPoint = new Point(0, 0);
            for (var i = 0; i < len; ++i)
            {
                var difX = centerX - contourPoints[i].X;
                var difY = centerY - contourPoints[i].Y;
                var dis = difX * difX + difY * difY;
                if (dis > maxDistance)
                {
                    maxDistance = dis;
                    maxPoint = contourPoints[i];
                }
                if (dis < minDistance)
                {
                    minDistance = dis;
                }
            }

            if (maxDistance > minDistance)
            {
                if (maxDistance / minDistance > 1.5)
                {
                    //算作彭大的部分
                    pointList.Add(new Vector3D(maxPoint.X, maxPoint.Y, z));
                }
            }
            //加上中心点
            pointList.Add(new Vector3D(centerX, centerY, z));

            //离上面最远点最远的点
            Point maxPoint1 = new Point(0, 0);
            maxDistance = 0;
            for (var i = 0; i < len; ++i)
            {
                var difX = maxPoint.X - contourPoints[i].X;
                var difY = maxPoint.Y - contourPoints[i].Y;
                var dis = difX * difX + difY * difY;
                if (dis > maxDistance)
                {
                    maxDistance = dis;
                    maxPoint1 = contourPoints[i];
                }
            }
            //加上离最远点最远的点
            pointList.Add(new Vector3D(maxPoint1.X, maxPoint1.Y, z));
            return pointList;
        }

        private Vector3D Get2DContoursCenterPoints(Point[] contourPoints, int z)
        {
            //计算轮廓的中心点
            var len = contourPoints.Length;
            var sumX = 0;
            var sumY = 0;
            for (var i = 0; i < len; ++i)
            {
                sumX += contourPoints[i].X;
                sumY += contourPoints[i].Y;
            }
            var centerX = sumX / (float)len;
            var centerY = sumY / (float)len;

            //加上离最远点最远的点          
            return new Vector3D(centerX, centerY, z);
        }

        private List<Vector3D> GetTwoPointInLine(List<Vector3D> pointsLsitBegin)
        {
            //所有点投影到z=0的平面上
            var count = pointsLsitBegin.Count;
            //double[] centerX = new double[count];
            //double[] centerY = new double[count];
            double xSum = 0;
            double ySum = 0;
            double zSum = 0;

            List<Point> pointCenter = new List<Point>();

            for (var i = 0; i < pointsLsitBegin.Count; ++i)
            {
                var x = pointsLsitBegin[i].X;
                var y = pointsLsitBegin[i].Y;
                xSum += x;
                ySum += y;
                pointCenter.Add(new Point((int)x, (int)y));
                zSum += pointsLsitBegin[i].Z;
            }
            //计算拟合直线方程
            var lineOriginal = MathTools2D.CalculateFittedLine(pointCenter);
            //得到Z=0平面上的两个点
            //直线中心点
            var center2D = new Vector2(xSum / count, ySum / count);
            Vector3D point1;
            Vector3D point2;
            //计算z=0平面上的两个点
            //判断直线的斜率-------------------------------------
            if (Math.Abs(lineOriginal[1]) == 0)
            {
                //x=C
                var x1 = -lineOriginal[2];
                x1 = Math.Max(0, x1);
                x1 = Math.Min(x1, _modelSize.ModelLengthX - 1);
                point1 = new Vector3D(x1, 0, 0);
                point2 = new Vector3D(x1, _modelSize.ModelLengthY - 1, 0);
            }
            else
            {
                var line = new double[2] { lineOriginal[0], lineOriginal[2] };

                if (line[0] == 0)
                {
                    //y=C
                    var y1 = line[1];
                    y1 = Math.Max(0, y1);
                    y1 = Math.Min(y1, _modelSize.ModelLengthY - 1);
                    point1 = new Vector3D(0, y1, 0);
                    point2 = new Vector3D(_modelSize.ModelLengthX - 1, y1, 0);
                }
                else
                {
                    //与四个边有交点
                    //平面中心点
                    var x2 = center2D.X;
                    x2 = Math.Max(0, x2);
                    x2 = Math.Min(x2, _modelSize.ModelLengthX - 1);
                    var y2 = center2D.Y;
                    y2 = Math.Max(0, y2);
                    y2 = Math.Min(y2, _modelSize.ModelLengthY - 1);
                    point1 = new Vector3D(x2, y2, 0);
                    //计算另一个点
                    //与x=0的交点
                    if (line[1] < 0)
                    {
                        var x3 = -line[1] / line[0];
                        x3 = Math.Max(0, x3);
                        x3 = Math.Min(x3, _modelSize.ModelLengthX - 1);
                        point2 = new Vector3D(x3, 0, 0);
                    }
                    else
                    {
                        if (line[1] >= 0 && line[1] <= _modelSize.ModelLengthY)
                        {
                            var y3 = line[1];
                            y3 = Math.Max(0, y3);
                            y3 = Math.Min(y3, _modelSize.ModelLengthY - 1);
                            point2 = new Vector3D(0, y3, 0);
                        }
                        else
                        {
                            var x4 = (_modelSize.ModelLengthY - line[1]) / line[0];
                            x4 = Math.Max(0, x4);
                            x4 = Math.Min(x4, _modelSize.ModelLengthX - 1);
                            point2 = new Vector3D(x4, _modelSize.ModelLengthY - 1, 0);
                        }
                    }
                }
            }
            return new List<Vector3D>() { point1, point2 };
        }

        private Vector3D Get3DContoursCenterPoinit(List<Vector3D> listPoint)
        {
            //计算中心点
            double sumX = 0;
            double sumY = 0;
            double sumZ = 0;
            var count = listPoint.Count;
            //计算中心点
            for (var i1 = 0; i1 < count; ++i1)
            {
                sumX += listPoint[i1].X;
                sumY += listPoint[i1].Y;
                sumZ += listPoint[i1].Z;
            }
            return new Vector3D(sumX / count, sumY / count, sumZ / count);
        }

        /// <summary>
        /// 计算Y型切面
        /// </summary>
        /// <param name="pointsLsit"></param>
        /// <returns></returns>
        private List<Vector3D> GetYPlanePoint(List<Vector3D> pointsLsitBegin, List<Vector3D> pointListEnd, bool isFindYBegin)
        {
            Vector3D point1;
            Vector3D point2;
            var centerPoint = Get3DContoursCenterPoinit(pointsLsitBegin);
            if (isFindYBegin)
            {
                var pointListInLine = GetTwoPointInLine(pointsLsitBegin);
                point1 = pointListInLine[0];
                point2 = pointListInLine[1];
            }
            else
            {
                var y = centerPoint.Y;
                y = Math.Max(0, y);
                y = Math.Min(y, _modelSize.ModelLengthY - 1);
                if (double.IsNaN(y))
                {
                    y = 0;
                }
                point1 = new Vector3D(0, y, 0);
                point2 = new Vector3D(_modelSize.ModelLengthX - 1, y, 0);
            }

            //第三个点
            Vector3D point3;
            if (pointListEnd.Count > 0)
            {
                var endCount = pointListEnd.Count;
                double xSumEnd = 0;
                double ySumEnd = 0;
                double zSumEnd = 0;
                for (var i = 0; i < pointListEnd.Count; ++i)
                {
                    xSumEnd += pointListEnd[i].X;
                    ySumEnd += pointListEnd[i].Y;
                    zSumEnd += pointListEnd[i].Z;
                }
                var y = ySumEnd / endCount;
                if (double.IsNaN(y))
                {
                    y = 0;
                }
                point3 = new Vector3D(xSumEnd / endCount, y, zSumEnd / endCount);
            }
            else
            {
                if (double.IsNaN(centerPoint.Y))
                {
                    centerPoint.Y = 0;
                }
                point3 = new Vector3D(centerPoint.X, centerPoint.Y, _modelSize.ModelLengthZ - 1);
            }

            //返回三个点
            return new List<Vector3D>() { point1, point2, point3 };
        }
    }
}