﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Runtime.InteropServices;
using Emgu.CV;
using Emgu.CV.CvEnum;
using Emgu.CV.Structure;
using WingAIDiagnosisService.Carotid.CarotidInterFacesData;
using WingAIDiagnosisService.Carotid.MathTools;
using WingServerCommon.Log;

namespace WingAIDiagnosisService.Carotid.Utilities
{
    //图像旋转参数
    public struct RotateParameter
    {
        public Matrix<double> MapMatrix;
        public Size ImageRotatedSize;
        public double Angle;
        public bool Add180;

        public RotateParameter(Matrix<double> mapMatrix)
        {
            MapMatrix = mapMatrix;
            ImageRotatedSize = new Size(0, 0);
            Angle = 0;
            Add180 = false;
        }
    }
    public class ModelMeasureSlice
    {
        //所有方向切片
        private ModelSlice _measureModelSlice;
        // 图像的端点
        private List<Vector2I> _imageEndPoint;
        //精度要求
        private readonly float _precision;
        //精度要求
        private readonly float _negPrecision;
        //切片旋转矩阵
        private CoordinateTransformation3D _coordinateTransformation3D;

        public ModelMeasureSlice(int maxSliceLength)
        {
            _precision = MathToolsBasis.Precision;
            _negPrecision = MathToolsBasis.NegPrecision;
            _measureModelSlice = new ModelSlice { SliceImageData = new byte[maxSliceLength] };
        }

        /// <summary>
        /// 设置测量图像像素值
        /// </summary>
        /// <param name="modelSlice"></param>
        /// <param name="imagePointList"></param>
        /// <returns></returns>
        public bool SetMeasureSliceData(ModelSlice modelSlice, List<Vector2I> imagePointList)
        {
            try
            {
                //防止切片图像大于测量数组大小
                var maxLength = modelSlice.SliceWidth * modelSlice.SliceHeight;
                if (maxLength > _measureModelSlice.SliceImageData.Length)
                {
                    _measureModelSlice.SliceImageData = new byte[maxLength];
                }
                _imageEndPoint = new List<Vector2I>(imagePointList);
                // 计算多边形各个边的直线方程
                var lines = MathTools2D.GetPolygonEdgeLineEquation(imagePointList);
                //获取直线内图像
                GetImageInLines(lines, modelSlice);
                return true;
            }
            catch (Exception e)
            {
                Logger.WriteLineError("ModelMeasureSlice SetMeasureSliceData have an error," + e.Message + "," + e.StackTrace);
                return false;
            }
        }

        /// <summary>
        /// 获得测量图像，主要是计算旋转角度和翻转图像
        /// </summary>
        /// <param name="coordinateTransformation3D"></param>
        /// <param name="clipCategory"></param>
        /// <param name="isFlip"></param>
        /// <returns></returns>
        public Image<Gray,byte> GetModelMeasureBitmap(CoordinateTransformation3D coordinateTransformation3D, bool isFlip)
        {
            try
            {
                _coordinateTransformation3D = coordinateTransformation3D;
                using (var slice = new Mat(_measureModelSlice.SliceHeight, _measureModelSlice.SliceWidth, DepthType.Cv8U, 1))
                {
                    Marshal.Copy(_measureModelSlice.SliceImageData, 0, slice.DataPointer, _measureModelSlice.SliceWidth * _measureModelSlice.SliceHeight);
                    var rotateParameter = GetRotateParameter(ClipCategory.Flank);
                    //旋转图像
                    using (var imageRotate = new Mat(rotateParameter.ImageRotatedSize, DepthType.Cv8U, 1))
                    {
                        //旋转图像
                        CvInvoke.WarpAffine(slice, imageRotate, rotateParameter.MapMatrix, imageRotate.Size);

                        var imageEndPoint = MathTools2D.GetRotatedPoints(MathTools2D.ChangeVector2IToPoint(_imageEndPoint), rotateParameter.MapMatrix);
                        //旋转图像180d度
                        if (rotateParameter.Add180)
                        {
                            CvInvoke.Rotate(imageRotate, imageRotate, RotateFlags.Rotate180);
                            //端点旋转180
                            imageEndPoint = GetRotated180EndPoints(imageEndPoint, imageRotate.Size);
                        }
                        //水平翻转图像
                        if (!isFlip)
                        {
                            CvInvoke.Flip(imageRotate, imageRotate, FlipType.Horizontal);
                            imageEndPoint = GetFlipedEndPoints(imageEndPoint, imageRotate.Size);
                        }

                        var roi = GetMinRoi(imageEndPoint, imageRotate.Size);
                        using (var image = imageRotate.ToImage<Gray, Byte>())
                        {
                            return image.Copy(roi);
                        }
                    }
                }
            }
            catch (Exception e)
            {
                Logger.WriteLineError("ModelMeasureSlice GetModelMeasureBitmap have an error," + e.Message + "," + e.StackTrace);
                return null;
            }
        }

        /// <summary>
        /// 计算图像的最小roi
        /// </summary>
        /// <param name="endPoints"></param>
        /// <param name="imageSize"></param>
        /// <returns></returns>
        private Rectangle GetMinRoi(List<Vector2> endPoints, Size imageSize)
        {
            var minX = int.MaxValue;
            var maxX = 0;

            var minY = int.MaxValue;
            var maxY = 0;
            for (var i = 0; i < endPoints.Count; ++i)
            {
                if (endPoints[i].X < minX)
                {
                    minX = (int)Math.Round(endPoints[i].X);
                }

                if (endPoints[i].X > maxX)
                {
                    maxX = (int)Math.Round(endPoints[i].X);
                }

                if (endPoints[i].Y < minY)
                {
                    minY = (int)Math.Round(endPoints[i].Y);
                }

                if (endPoints[i].Y > maxY)
                {
                    maxY = (int)Math.Round(endPoints[i].Y);
                }
            }

            minX = Math.Min(minX - 1, imageSize.Width);
            minX = Math.Max(0, minX);
            maxX = Math.Min(maxX + 1, imageSize.Width);
            maxX = Math.Max(0, maxX);
            minY = Math.Min(minY - 1, imageSize.Height);
            minY = Math.Max(0, minY);
            maxY = Math.Min(maxY + 1, imageSize.Height);
            maxY = Math.Max(0, maxY);
            return new Rectangle(new Point(minX, minY), new Size(maxX - minX, maxY - minY));
        }

        /// <summary>
        /// 计算旋转180度后的点坐标
        /// </summary>
        /// <param name="endPoints"></param>
        /// <param name="imageSize"></param>
        /// <returns></returns>
        private List<Vector2> GetRotated180EndPoints(List<Vector2> endPoints, Size imageSize)
        {
            var newEndPoints = new List<Vector2>();
            foreach (var points in endPoints)
            {
                var x = (int)Math.Round(points.X);
                x = Math.Max(0, x);
                x = Math.Min(imageSize.Width - 1, x);

                var y = (int)Math.Round(points.Y);
                y = Math.Max(0, y);
                y = Math.Min(imageSize.Height - 1, y);

                x = imageSize.Width - 1 - x;
                y = imageSize.Height - 1 - y;
                newEndPoints.Add(new Vector2(x, y));
            }
            return newEndPoints;
        }

        /// <summary>
        /// 计算水平翻转后的点坐标
        /// </summary>
        /// <param name="endPoints"></param>
        /// <param name="imageSize"></param>
        /// <returns></returns>
        private List<Vector2> GetFlipedEndPoints(List<Vector2> endPoints, Size imageSize)
        {
            var newEndPoints = new List<Vector2>();
            foreach (var points in endPoints)
            {
                var x = (int)Math.Round(points.X);
                x = Math.Max(0, x);
                x = Math.Min(imageSize.Width - 1, x);

                var y = (int)Math.Round(points.Y);
                y = Math.Max(0, y);
                y = Math.Min(imageSize.Height - 1, y);

                x = imageSize.Width - 1 - x;
                newEndPoints.Add(new Vector2(x, y));
            }
            return newEndPoints;
        }

        /// <summary>
        /// 获取闭合直线多边形内的图像
        /// </summary>
        /// <param name="lines"></param>
        private void GetImageInLines(List<LineEquation> lines, ModelSlice modelSlice)
        {
            try
            {
                _measureModelSlice.SliceWidth = modelSlice.SliceWidth;
                _measureModelSlice.SliceHeight = modelSlice.SliceHeight;
                var index = 0;
                for (var h = 0; h < _measureModelSlice.SliceHeight; ++h)
                {
                    for (var w = 0; w < _measureModelSlice.SliceWidth; ++w)
                    {
                        _measureModelSlice.SliceImageData[index] = modelSlice.SliceImageData[index];
                        //if (MathTools2D.IsIntPolygon(new Vector2I(w, h), lines) > 0)
                        //{
                        //    _measureModelSlice.SliceImageData[index] = modelSlice.SliceImageData[index];
                        //}
                        //else
                        //{
                        //    _measureModelSlice.SliceImageData[index] = 0;
                        //}
                        index++;
                    }
                }
            }
            catch (Exception e)
            {
                Logger.WriteLineError("ModelMeasureSlice GetImageInLines have an error," + e.Message + "," + e.StackTrace);
            }
        }

        /// <summary>
        /// 获得另一个与中心点Y坐标相等的点
        /// </summary>
        /// <param name="centerPoint2D"></param>
        /// <param name="center3Dy"></param>
        /// <returns></returns>
        private Point GetOtherPointEqualY(Point centerPoint2D)
        {
            var centerPoint3D = MathTools3D.Change2DTo3D(centerPoint2D, _coordinateTransformation3D);
            //计算切面边上的点，找到3D点的y值等于center3DY的点
            var minDisPoint = new Point(0, 0);
            double minDisY = int.MaxValue;
            //图像 两条边上的点
            var edgePointsList = new List<Point>();
            for (var i = 0; i < _measureModelSlice.SliceWidth; i++)
            {
                edgePointsList.Add(new Point(i, 0));
            }
            var maxX = _measureModelSlice.SliceWidth - 1;
            for (var i = 0; i < _measureModelSlice.SliceHeight; ++i)
            {
                edgePointsList.Add(new Point(maxX, i));
            }

            foreach (var point in edgePointsList)
            {
                if (point == centerPoint2D)
                {
                    continue;
                }
                var point3D = MathTools3D.Change2DTo3D(point, _coordinateTransformation3D);
                var dis = Math.Abs(point3D.Y - centerPoint3D.Y);
                if (dis < _precision)
                {
                    return point;
                }
                if (dis < minDisY)
                {
                    minDisY = dis;
                    minDisPoint = point;
                }
            }
            return minDisPoint;
        }

        /// <summary>
        /// 获得另一个与中心点X坐标相等的点
        /// </summary>
        /// <param name="centerPoint2D"></param>
        /// <param name="center3Dx"></param>
        /// <returns></returns>
        private Point GetOtherPointEqualX(Point centerPoint2D)
        {
            var centerPoint3D = MathTools3D.Change2DTo3D(centerPoint2D, _coordinateTransformation3D);
            //计算切面边上的点，找到3D点的x值等于center3Dx的点
            var minDisPoint = new Point(0, 0);
            double minDisX = int.MaxValue;
            //图像 两条边上的点
            var edgePointsList = new List<Point>();
            for (var i = 0; i < _measureModelSlice.SliceWidth; i++)
            {
                edgePointsList.Add(new Point(i, 0));
            }
            var maxX = _measureModelSlice.SliceWidth - 1;
            for (var i = 0; i < _measureModelSlice.SliceHeight; ++i)
            {
                edgePointsList.Add(new Point(maxX, i));
            }

            foreach (var point in edgePointsList)
            {
                if (point == centerPoint2D)
                {
                    continue;
                }
                var point3D = MathTools3D.Change2DTo3D(point, _coordinateTransformation3D);
                var dis = Math.Abs(point3D.X - centerPoint3D.X);
                if (dis < _precision)
                {
                    return point;
                }
                if (dis < minDisX)
                {
                    minDisX = dis;
                    minDisPoint = point;
                }
            }
            return minDisPoint;
        }

        /// <summary>
        /// p1靠近零点，返回的向量，已p1为起点，指向p2
        /// </summary>
        /// <param name="p1"></param>
        /// <param name="p2"></param>
        /// <returns></returns>
        private Vector2 GetVector(Point p1, Point p2)
        {
            Point pointStart;
            Point pointEnd;
            if (Math.Abs(p1.Y - p2.Y) < _precision)
            {
                if (p1.X < p2.X)
                {
                    pointStart = p1;
                    pointEnd = p2;
                }
                else
                {
                    pointStart = p2;
                    pointEnd = p1;
                }
            }
            else
            {
                if (p1.Y - p2.Y < _negPrecision)
                {
                    pointStart = p1;
                    pointEnd = p2;
                }
                else
                {
                    pointStart = p2;
                    pointEnd = p1;
                }
            }
            return new Vector2(pointEnd.X - pointStart.X, pointEnd.Y - pointStart.Y);
        }

        /// <summary>
        /// 计算旋转角度
        /// </summary>
        /// <param name="centerPoint"></param>
        /// <param name="clipType"></param>
        /// <returns></returns>
        private double GetRotateAngle(Point centerPoint, ClipCategory clipType)
        {
            //获得与中中心点同一高度的另一个点
            Point otherPoint2D;
            if (clipType == ClipCategory.Flank)
            {
                otherPoint2D = GetOtherPointEqualY(centerPoint);
            }
            else
            {
                otherPoint2D = GetOtherPointEqualX(centerPoint);
            }
            //计算（centerPoint2D,otherPoint2D）与x轴的夹角------------------------------------------------
            var vectorLine1 = GetVector(centerPoint, otherPoint2D);
            var vectorLine2 = new Vector2(_measureModelSlice.SliceWidth - 1, 0);
            var cosA = MathTools2D.GetCosA2D(vectorLine1, vectorLine2);
            return MathToolsBasis.ChangeRadianToAngle(Math.Acos(cosA));
        }

        /// <summary>
        /// 获得图像旋转参数
        /// </summary>
        /// <param name="clipType"></param>
        /// <returns></returns>
        private RotateParameter GetRotateParameter(ClipCategory clipType)
        {
            try
            {
                //获得旋转中心
                var centerPoint2D = MathTools2D.GetCenterPoint(_imageEndPoint);
                //获得旋转角度
                var angle = GetRotateAngle(centerPoint2D, clipType);

                return GetRotateMatrix(centerPoint2D, angle, clipType);
            }
            catch (Exception e)
            {
                Logger.WriteLineError("ModelMeasureSlice GetRotateParameter have an error," + e.Message + "," + e.StackTrace);
                return new RotateParameter(null);
            }
        }

        /// <summary>
        /// 获得旋转矩阵
        /// </summary>
        /// <param name="center"></param>
        /// <param name="rotateAngle"></param>
        /// <returns></returns>
        private RotateParameter GetRotateMatrix(Point center, double rotateAngle, ClipCategory clipType)
        {
            try
            {
                var imageSrcSize = new Size(_measureModelSlice.SliceWidth, _measureModelSlice.SliceHeight);
                //图像旋转矩阵
                var mapMatrix = MathTools2D.GetImageMapMatrix(center, rotateAngle, imageSrcSize);
                //旋转前图像端点
                var endPointListSrc = MathTools2D.GetImageEndPoint(imageSrcSize);
                //旋转后图像端点
                var endPointList = MathTools2D.GetRotatedPoints(endPointListSrc, mapMatrix);

                var coordinateRange = MathTools2D.GetCoordinateRange(endPointList);
                //旋转后图像的大小
                var imageRotatedSize = new Size((int)Math.Round(Math.Abs(coordinateRange.MaxX - coordinateRange.MinX)) + 1,
                    (int)Math.Round(Math.Abs(coordinateRange.MaxY - coordinateRange.MinY)) + 1);

                //判断是否需要旋转180度-----------------------
                var add180 = false;
                var index = MathTools2D.GetTopPointIndex(endPointList);
                var pointMin = endPointListSrc[index];
                var pointMin3D = MathTools3D.Change2DTo3D(pointMin, _coordinateTransformation3D);
                var center3D = MathTools3D.Change2DTo3D(center, _coordinateTransformation3D);
                if (clipType == ClipCategory.Flank && pointMin3D.Y - center3D.Y > _precision)
                {
                    add180 = true;
                }
                if (clipType == ClipCategory.Soleplate && pointMin3D.X - center3D.X > _precision)
                {
                    add180 = true;
                }
                //计算最终旋转参数
                var rotateParameter = new RotateParameter
                {
                    Add180 = add180,
                    ImageRotatedSize = imageRotatedSize,
                    Angle = rotateAngle,
                    MapMatrix = mapMatrix
                };
                return rotateParameter;
            }
            catch (Exception e)
            {
                Logger.WriteLineError("ModelMeasureSlice GetRotateMatrix error," + e.Message + "," + e.StackTrace);
                return new RotateParameter(null);
            }
        }

    }
}
