﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.InteropServices;
using System.Threading.Tasks;
using Emgu.CV;
using Emgu.CV.CvEnum;
using Emgu.CV.Structure;
using System.Drawing;
using System.IO;
using WingAIDiagnosisService.Carotid.CarotidInterFacesData;
using WingAIDiagnosisService.Carotid.MathTools;
using WingServerCommon.Log;

namespace WingAIDiagnosisService.Carotid.Utilities
{
    /// <summary>
    /// 模型尺寸
    /// </summary>
    public struct ModelSize
    {
        public int ModelLengthX;
        public int ModelLengthY;
        public int ModelLengthZ;
        public int OneFacePixelsNum;
    }

    /// <summary>
    /// 模型切片类型
    /// </summary>
    public enum ModelSliceType
    {
        XEqual = 0,
        YEqual,
        ZEqual,
        Normal
    }

    /// <summary>
    /// 模型切面
    /// </summary>
    public struct ModelSlice
    {
        //切片像素点数据
        public byte[] SliceImageData;
        //图像的宽
        public int SliceWidth;
        //图像的高
        public int SliceHeight;
    }

    /// <summary>
    /// 获得切片图像的类
    /// </summary>
    public class SliceHelper : IDisposable
    {
        //模型的尺寸
        private ModelSize _modelSize;
        //模型边界检测
        private ModelEdgeChecker _modelEdgeChecker;
        //读取到的模型数据
        private byte[][] _modelData;
        //一般切片
        private ModelSlice _generalModelSlice;

        //是否已经设置模型源文件
        private bool _isSourceBeSet;
        //确定一个平面的三个不共线的点
        private One3DPlaneThreeBasicPoints _inputThreeBasic3DPoints;
        //切片的端点list，世界点坐标
        private List<Vector3D> _endPointList3D;
        //切片的端点list，图像点坐标
        private List<Vector2I> _endPointList2I;
        //三维坐标变换参数
        private CoordinateTransformation3D _coordinateTransformation3D;
        //切片图像的最大长度
        private int _sliceMaxLength;

        /// 切片图像坐标系的基础3D点
        private One2DPlaneThreeBasic3DPoints _planeBasic3DPoints;
        /// 切片图像坐标系的基础2D点
        private One2DPlaneThreeBasic2DPoints _planeBasic2DPoints;
        //多线程线程个数
        private readonly ParallelOptions _parallelOption = new ParallelOptions { MaxDegreeOfParallelism = 2 * Environment.ProcessorCount };

        private static SliceHelper _instance;

        public static SliceHelper Instance
        {
            get => _instance ?? (_instance = new SliceHelper());
        }

        public SliceHelper()
        {
            _generalModelSlice = new ModelSlice();
            _inputThreeBasic3DPoints = new One3DPlaneThreeBasicPoints();
            _endPointList3D = new List<Vector3D>();
            _endPointList2I = new List<Vector2I>();
        }

        /// <summary>
        /// Set source of the model
        /// </summary>
        /// <param name="mdl">mdl file</param>
        /// <param name="carotidType">leftCarotid or rightCarotid</param>
        /// <param name="width">model's width distance</param>
        /// <param name="height">model's height distance</param>
        /// <param name="depth">model's depth distance</param>
        /// <param name="scanDepth">scan depth</param>
        /// <returns>Whether set model source successfully</returns>
        public bool SetModelSource(string mdlFilePath, int width, int height, int depth)
        {
            try
            {
                if (depth < 5)
                {
                    Logger.WriteLineInfo("Server SliceHelper,Do SetModelSource,depth<5.");
                    return false;
                }
                if (_isSourceBeSet)
                {
                    Release();
                }
                _isSourceBeSet = false;
                _modelSize = new ModelSize
                {
                    ModelLengthX = width,
                    ModelLengthY = height,
                    ModelLengthZ = depth,
                    OneFacePixelsNum = width * height
                };

                _modelData = new byte[_modelSize.ModelLengthZ][];
                for (var i = 0; i < _modelSize.ModelLengthZ; ++i)
                {
                    _modelData[i] = new byte[_modelSize.OneFacePixelsNum];
                }

                using (var img = new Mat())
                {
                    using (var stream = new FileStream(mdlFilePath, FileMode.Open))
                    {
                        var reader = new Carotid3DStreamReader(stream);
                        var imgCount = reader.ReadInt();
                        if (imgCount != depth)
                        {
                            Logger.WriteLineError("Server slicehelp mdl file's image count not equal with model depth.");
                            return false;
                        }
                        for (var i = 0; i < imgCount; ++i)
                        {
                            var readBytes = reader.ReadBytes();
                            CvInvoke.Imdecode(readBytes, ImreadModes.Grayscale, img);
                            Marshal.Copy(img.DataPointer, _modelData[i], 0, _modelSize.OneFacePixelsNum);
                        }
                    }
                }

                _sliceMaxLength = GetMaxImageLength();
                _generalModelSlice.SliceImageData = new byte[_sliceMaxLength];
                _modelEdgeChecker = new ModelEdgeChecker(_modelSize);
                _inputThreeBasic3DPoints = new One3DPlaneThreeBasicPoints
                {
                    PointA = new Vector3D(0, 0, 0),
                    PointB = new Vector3D(0, 0, 0),
                    PointC = new Vector3D(0, 0, 0)
                };

                _isSourceBeSet = true;
                return true;
            }
            catch (Exception e)
            {
                Logger.WriteLineError("Server SliceHelper, SetModelSource error," + e.Message + "," + e.StackTrace);
                return false;
            }
        }

        public ModelSize GetModeSzie()
        {
            return _modelSize;
        }
        public byte[][] GetModelData()
        {
            return _modelData;
        }

        public bool SetSlicePoint(List<Vector3D> pointsList)
        {
            //模型是否已设置
            if (!_isSourceBeSet)
            {
                Logger.WriteLineError("Server SliceHelper,Model source not be set");
                return false;
            }
            //输入点个数是否小于3
            if (pointsList.Count < 3)
            {
                Logger.WriteLineError("Server SliceHelper,input point count is less than 3");
                return false;
            }
           
            //端点是否在模型内
            foreach (var point in pointsList)
            {
                if (_modelEdgeChecker.IsInModel(point)) continue;
                Logger.WriteLineError(
                    $" Server SliceHelper,Slice input point error: out of model x:{point.X}, y:{point.Y}, z:{point.Z},model size x:{_modelSize.ModelLengthX}, y:{_modelSize.ModelLengthY}, z:{_modelSize.ModelLengthZ}");
                return false;
            }

            var count = pointsList.Count;
            for (var i = 0; i < count; ++i)
            {
                for (var j = 0; j < count; ++j)
                {
                    for (var z = 0; z < count; ++z)
                    {
                        if (i == j || j == z || i == z) continue;
                        //计算切面的基础点
                        _inputThreeBasic3DPoints.PointA = pointsList[i];
                        _inputThreeBasic3DPoints.PointB = pointsList[j];
                        _inputThreeBasic3DPoints.PointC = pointsList[z];
                        if (MathTools3D.IsParallel(_inputThreeBasic3DPoints.PointA, _inputThreeBasic3DPoints.PointB, _inputThreeBasic3DPoints.PointC))
                        {
                            return false;
                        }
                    }
                }
            }
            return true;
        }

        /// <summary>
        /// 计算模型切片
        /// </summary>
        /// <param name="inputData"></param>
        /// <returns></returns>
        private bool GetModelSlice(List<Vector3D> pointsList)
        {
            try
            {
                if (!SetSlicePoint(pointsList))
                {
                    return false;
                }
                //获得切面类型
                var sliceType = MathTools3D.GetModelSliceType(_inputThreeBasic3DPoints);
                //计算切面与模型的交点
                _endPointList3D = MathTools3D.GetIntersectionOfPlaneAndModel(_inputThreeBasic3DPoints, _modelSize, sliceType);
                //计算旋转矩阵----------------------------------------------------------------------------------------             
                _coordinateTransformation3D = GetCoordinateTransformation();
                //获得三维点对应的二维点
                _endPointList2I = MathTools3D.GetRotatedImageEndPoints(_endPointList3D, _planeBasic3DPoints, sliceType);
                //顺时针排序端点
                MathTools3D.SortEndPointClockwise(_planeBasic2DPoints, _endPointList3D, _endPointList2I);
                //获得切片图像大小
                GetSlicePanelSize(sliceType);
                //设置切面图像像素点的值
                return SetModelSliceData(sliceType);
            }
            catch (Exception e)
            {
                Logger.WriteLineError("Server SliceHelper, GetModelSlice error." + e.Message + "," + e.StackTrace);
                return false;
            }
        }

        public Image<Gray, byte> GetImageFromThreePoint(List<Vector3D> points, bool isNeedFlip)
        {

            if (GetModelSlice(points))
            {
                using (var mat = new Mat(_generalModelSlice.SliceHeight, _generalModelSlice.SliceWidth, DepthType.Cv8U, 1))
                {
                    Marshal.Copy(_generalModelSlice.SliceImageData, 0, mat.DataPointer, _generalModelSlice.SliceWidth * _generalModelSlice.SliceHeight);
                    //判断是否需要垂直翻转，z坐标大的在上
                    var image = mat.ToImage<Gray, byte>();
                    if (isNeedFlip)
                    {
                        if (IsFlipVertically(image.Height))
                        {
                            CvInvoke.Flip(image, image, FlipType.Vertical);
                        }
                    }

                    return image;
                }
            }
            else
            {
                return null;
            }

        }

        public Vector2I GetPoint2D(Vector3D point)
        {
            return MathTools3D.Chang3DTo2D(point, _planeBasic3DPoints);
        }

        public bool IsFlipVertically(int imageHeight)
        {
            Point point1 = new Point(0, 0);
            var point13D = MathTools3D.Change2DTo3D(point1, _coordinateTransformation3D);
            Point point2 = new Point(0, imageHeight - 1);
            var point23D = MathTools3D.Change2DTo3D(point2, _coordinateTransformation3D);
            if (point13D.Z < point23D.Z)
            {
                return true;
            }
            return false;
        }

        public Image<Gray, byte> Get2DMeasureImage(List<Vector3D> pointList)
        {
            try
            {
                if (!GetModelSlice(pointList))
                {
                    Logger.WriteLineError("Server SliceHelper, Get2DMeasureImage GetModelSlice failed.");
                    return null;
                }
                using (var mat = new Mat(_generalModelSlice.SliceHeight, _generalModelSlice.SliceWidth, DepthType.Cv8U, 1))
                {
                    Marshal.Copy(_generalModelSlice.SliceImageData, 0, mat.DataPointer, _generalModelSlice.SliceWidth * _generalModelSlice.SliceHeight);
                    //判断是否需要垂直翻转，z坐标大的在上
                    using var image = mat.ToImage<Gray, byte>();
                    //先水平翻转
                    CvInvoke.Flip(image, image, FlipType.Horizontal);
                    var imageRotate = new Image<Gray, byte>(image.Height, image.Width);
                    CvInvoke.Rotate(image, imageRotate, RotateFlags.Rotate90Clockwise);
                    return imageRotate;
                }
            }
            catch (Exception e)
            {
                Logger.WriteLineError("Server SliceHelper, Get2DMeasureImage error," + e.Message + "," + e.StackTrace);
                return null;
            }
        }

        //Should be called when window closed.
        public void Release()
        {
            _generalModelSlice.SliceImageData = null;
            _modelData = null;
            //GC.Collect();
        }

        /// <summary>
        /// 获取模型最大的切面尺寸
        /// </summary>
        /// <returns></returns>
        private int GetMaxImageLength()
        {
            var lenList = new List<int> { _modelSize.ModelLengthX, _modelSize.ModelLengthY, _modelSize.ModelLengthZ };

            var maxHypotenuse1 = Math.Sqrt(lenList[0] * lenList[0] + lenList[1] * lenList[1]);
            var maxHypotenuse2 = Math.Sqrt(lenList[1] * lenList[1] + lenList[2] * lenList[2]);
            var maxHypotenuse3 = Math.Sqrt(lenList[0] * lenList[0] + lenList[2] * lenList[2]);

            var maxLength = Math.Max(maxHypotenuse1, maxHypotenuse2);
            maxLength = Math.Max(maxLength, maxHypotenuse3);
            return (int)Math.Round(maxLength * maxLength);
        }

        /// <summary>
        /// 计算旋转矩阵
        /// </summary>
        /// <param name="endPointList3D"></param>
        /// <returns></returns>
        private CoordinateTransformation3D GetCoordinateTransformation()
        {
            //旋转后的图像坐标系,基于的三个世界点
            _planeBasic3DPoints = MathTools3D.GetOne2DPlaneBasicPoints(_endPointList3D);
            //计算_plane2DBasic3DPoints对应的二维点
            _planeBasic2DPoints = MathTools3D.ChangePlaneBasic3DPointTo2D(_planeBasic3DPoints);
            //根据坐标系基础点的三维坐标和二维坐标计算旋转矩阵
            return MathTools3D.GetCoordinateTransformationParameter(_planeBasic3DPoints, _planeBasic2DPoints);
        }

        /// <summary>
        /// 设置切面图像像素点的值
        /// </summary>
        /// <param name="sliceType"></param>
        /// <returns></returns>
        private bool SetModelSliceData(ModelSliceType sliceType)
        {
            try
            {
                //获得切片像素值
                switch (sliceType)
                {
                    case ModelSliceType.XEqual:
                        return GetModeSliceDataX(MathToolsBasis.DoubleToInt(_inputThreeBasic3DPoints.PointA.X));
                    case ModelSliceType.YEqual:
                        return GetModeSliceDataY(MathToolsBasis.DoubleToInt(_inputThreeBasic3DPoints.PointA.Y));
                    case ModelSliceType.ZEqual:
                        return GetModeSliceDataZ(MathToolsBasis.DoubleToInt(_inputThreeBasic3DPoints.PointA.Z));
                    case ModelSliceType.Normal:
                        return GetModeSliceDataNormal();
                    default:
                        return false;
                }
            }
            catch (Exception e)
            {
                Logger.WriteLineError("Server SliceHelper, SetModelSliceData error," + e.Message + "," + e.StackTrace);
                return false;
            }
        }

        /// <summary>
        /// x坐标相等时获得切片图像数据的函数
        /// </summary>
        /// <param name="point1"></param>
        /// <returns></returns>
        private bool GetModeSliceDataX(double xPos)
        {
            var x = MathToolsBasis.DoubleToInt(xPos);
            x = Math.Min(_modelSize.ModelLengthX - 1, x);
            x = Math.Max(0, x);

            Parallel.For(0, _generalModelSlice.SliceHeight, _parallelOption, h =>
            {
                var dstIndex = h * _generalModelSlice.SliceWidth;
                var z = _modelSize.ModelLengthZ - 1 - h;
                var srcIndex = _modelSize.ModelLengthX * (_modelSize.ModelLengthY - 1) + x;
                for (var w = 0; w < _generalModelSlice.SliceWidth; ++w)
                {
                    _generalModelSlice.SliceImageData[dstIndex] = _modelData[z][srcIndex];
                    ++dstIndex;
                    srcIndex -= _modelSize.ModelLengthX;
                }
            });
            return true;
        }

        /// <summary>
        /// y坐标相等时获得切片图像数据的函数
        /// </summary>
        /// <param name="yPos"></param>
        /// <returns></returns>
        private bool GetModeSliceDataY(double yPos)
        {
            var y = MathToolsBasis.DoubleToInt(yPos);
            y = Math.Min(_modelSize.ModelLengthY - 1, y);
            y = Math.Max(0, y);
            var z = _modelSize.ModelLengthZ - 1;
            var srcIndex = (y + 1) * _modelSize.ModelLengthX - 1;

            Parallel.For(0, _generalModelSlice.SliceHeight, _parallelOption, h =>
            {
                var z1 = z - h;
                var dstIndex = h * _generalModelSlice.SliceWidth;
                for (var w = 0; w < _generalModelSlice.SliceWidth; ++w)
                {
                    _generalModelSlice.SliceImageData[dstIndex] = _modelData[z1][srcIndex - w];
                    ++dstIndex;
                }
            });
            return true;
        }

        /// <summary>
        /// z坐标相等时获得切片图像数据的函数
        /// </summary>
        /// <param name="zPos"></param>
        /// <returns></returns>
        private bool GetModeSliceDataZ(double zPos)
        {
            var z = MathToolsBasis.DoubleToInt(zPos);
            z = Math.Min(_modelSize.ModelLengthZ - 1, z);
            z = Math.Max(0, z);
            var dstIndex = 0;
            var srcIndex = _modelSize.ModelLengthX * _modelSize.ModelLengthY - 1;
            var count = _generalModelSlice.SliceHeight * _generalModelSlice.SliceWidth;
            for (var i = 0; i < count; ++i)
            {
                _generalModelSlice.SliceImageData[dstIndex] = _modelData[z][srcIndex];
                ++dstIndex;
                --srcIndex;
            }
            return true;
        }

        /// <summary>
        /// XYZ坐标都不相等时获得切片图像数据的函数
        /// </summary>
        /// <param name="point1"></param>
        /// <param name="point2"></param>
        /// <param name="point3"></param>
        /// <returns></returns>
        private bool GetModeSliceDataNormal()
        {
            try
            {
                //分别计算每一行的图像像素值
                Parallel.For(0, _generalModelSlice.SliceHeight, _parallelOption, GetSliceData);
                return true;
            }
            catch (Exception e)
            {
                Logger.WriteLineError("Server SliceHelper, GetModeSliceDataNormal have an error," + e.Message + "," + e.StackTrace);
                return false;
            }
        }

        /// <summary>
        /// 计算切面图像点像素值,每一行
        /// </summary>
        /// <param name="start"></param>
        /// <param name="end"></param>
        private void GetSliceData(int start)
        {
            try
            {
                // Origin formula
                // image3D[count2].X = point2D.X * mat0 + point2D.Y * mat1 + _translateT[0];
                // image3D[count2].Y = point2D.X * mat3 + point2D.Y * mat4 + _translateT[1];
                // image3D[count2].Z = point2D.X * mat6 + point2D.Y * mat7 + _translateT[2];

                var tran = _coordinateTransformation3D.TranslationMatrix;
                var rotationMatrix = _coordinateTransformation3D.RotationMatrix;
                var startIndex = start * _generalModelSlice.SliceWidth;
                var maxCount = startIndex + _generalModelSlice.SliceWidth;
                var modelLengthX = _modelSize.ModelLengthX;

                var x = tran[0] + rotationMatrix[1] * start;
                var y = tran[1] + rotationMatrix[4] * start;
                var z = tran[2] + rotationMatrix[7] * start;

                var negativeEdge = MathToolsBasis.NegPrecision;
                var xEdge = _modelSize.ModelLengthX - 1 + MathToolsBasis.Precision;
                var yEdge = _modelSize.ModelLengthY - 1 + MathToolsBasis.Precision;
                var zEdge = _modelSize.ModelLengthZ - 1 + MathToolsBasis.Precision;

                while (startIndex < maxCount)
                {
                    if (x > negativeEdge && x < xEdge && y > negativeEdge && y < yEdge && z > negativeEdge && z < zEdge)
                    {
                        _generalModelSlice.SliceImageData[startIndex] = _modelData[(int)(z + 0.5)][modelLengthX * (int)(y + 0.5) + (int)(x + 0.5)];
                    }
                    else
                    {
                        _generalModelSlice.SliceImageData[startIndex] = 0;
                    }
                    x += rotationMatrix[0];
                    y += rotationMatrix[3];
                    z += rotationMatrix[6];
                    ++startIndex;
                }
            }
            catch (Exception e)
            {
                Logger.WriteLineError($"Server SliceHelper, GetSliceData error {e}");
            }
        }

        /// <summary>
        /// 获得切片最小尺寸
        /// </summary>
        private void GetSlicePanelSize(ModelSliceType sliceType)
        {
            //获得切片像素值
            switch (sliceType)
            {
                case ModelSliceType.XEqual:
                    _generalModelSlice.SliceWidth = _modelSize.ModelLengthY;
                    _generalModelSlice.SliceHeight = _modelSize.ModelLengthZ;
                    break;
                case ModelSliceType.YEqual:
                    _generalModelSlice.SliceWidth = _modelSize.ModelLengthX;
                    _generalModelSlice.SliceHeight = _modelSize.ModelLengthZ;
                    break;
                case ModelSliceType.ZEqual:
                    _generalModelSlice.SliceWidth = _modelSize.ModelLengthX;
                    _generalModelSlice.SliceHeight = _modelSize.ModelLengthY;
                    break;
                default:
                    _generalModelSlice.SliceWidth = _endPointList2I.Max(m => m.X);
                    _generalModelSlice.SliceHeight = _endPointList2I.Max(m => m.Y);
                    break;
            }

            var imageLength = _generalModelSlice.SliceWidth * _generalModelSlice.SliceHeight;
            if (_sliceMaxLength < imageLength)
            {
                _generalModelSlice.SliceImageData = new byte[imageLength];
                _sliceMaxLength = imageLength;
            }
        }

        public void Dispose()
        {
            _modelData = null;
            _instance = null;
            _generalModelSlice.SliceImageData = null;
            _endPointList3D.Clear();
            _endPointList2I.Clear();
        }

    }
}